/*
 * Name:    d_epics_motor.c 
 *
 * Purpose: MX motor driver for the EPICS APS-XFD developed motor record.
 *          Note that this driver uses the EZCA channel access library
 *          which is packaged as an extension to EPICS and is not in
 *          the base release.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_motor.h"
#include "d_epics_motor.h"

#include "mx_clock.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

#include "mx_epics.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_epics_motor_record_function_list = {
	mxd_epics_motor_initialize_type,
	mxd_epics_motor_create_record_structures,
	mxd_epics_motor_finish_record_initialization,
	mxd_epics_motor_delete_record,
	mxd_epics_motor_print_structure,
	mxd_epics_motor_read_parms_from_hardware,
	mxd_epics_motor_write_parms_to_hardware,
	mxd_epics_motor_open,
	mxd_epics_motor_close
};

MX_MOTOR_FUNCTION_LIST mxd_epics_motor_motor_function_list = {
	mxd_epics_motor_motor_is_busy,
	mxd_epics_motor_move_absolute,
	mxd_epics_motor_get_position,
	mxd_epics_motor_set_position,
	mxd_epics_motor_soft_abort,
	mxd_epics_motor_immediate_abort,
	mxd_epics_motor_positive_limit_hit,
	mxd_epics_motor_negative_limit_hit,
	mxd_epics_motor_find_home_position,
	mxd_epics_motor_constant_velocity_move,
	mxd_epics_motor_get_parameter,
	mxd_epics_motor_set_parameter
};

/* EPICS motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_epics_motor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_EPICS_MOTOR_STANDARD_FIELDS
};

long mxd_epics_motor_num_record_fields
		= sizeof( mxd_epics_motor_record_field_defaults )
			/ sizeof( mxd_epics_motor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_epics_motor_rfield_def_ptr
			= &mxd_epics_motor_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_epics_motor_get_pointers( MX_MOTOR *motor,
			MX_EPICS_MOTOR **epics_motor,
			const char *calling_fname )
{
	const char fname[] = "mxd_epics_motor_get_pointers()";

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( epics_motor == (MX_EPICS_MOTOR **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( motor->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
    "The MX_RECORD pointer for the MX_MOTOR pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*epics_motor = (MX_EPICS_MOTOR *) motor->record->record_type_struct;

	if ( *epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_EPICS_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_epics_motor_initialize_type( long type )
{
		return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_motor_create_record_structures()";

	MX_MOTOR *motor;
	MX_EPICS_MOTOR *epics_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	epics_motor = (MX_EPICS_MOTOR *) malloc( sizeof(MX_EPICS_MOTOR) );

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_EPICS_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = epics_motor;
	record->class_specific_function_list
				= &mxd_epics_motor_motor_function_list;

	motor->record = record;

	/* An EPICS motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_finish_record_initialization( MX_RECORD *record )
{
	MX_MOTOR *motor;
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	motor = (MX_MOTOR *) record->record_class_struct;

	motor->acceleration_type = MXF_MTR_ACCEL_TIME;

	return status;
}

MX_EXPORT mx_status_type
mxd_epics_motor_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_epics_motor_print_structure()";

	MX_MOTOR *motor;
	MX_EPICS_MOTOR *epics_motor;
	double position, deadband;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type      = EPICS_MOTOR.\n\n");

	fprintf(file, "  name            = %s\n", record->name);
	fprintf(file, "  EPICS record    = %s\n",
					epics_motor->epics_record_name);

	mx_status = mx_motor_get_position( record, &position );

	if ( mx_status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	fprintf(file, "  position        = %g %s  (%g).\n",
		motor->position, motor->units, 
		motor->raw_position.analog );
	fprintf(file, "  scale           = %g %s per EPICS EGU.\n",
		motor->scale, motor->units );
	fprintf(file, "  offset          = %g %s.\n",
		motor->offset, motor->units );
	fprintf(file, "  backlash        = %g %s  (%g).\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog );
	fprintf(file, "  negative limit  = %g %s  (%g).\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog );
	fprintf(file, "  positive limit  = %g %s  (%g).\n",
	    motor->positive_limit, motor->units,
	    motor->raw_positive_limit.analog );

	deadband = motor->scale * motor->deadband.analog;

	fprintf(file, "  deadband        = %g %s  (%g).\n\n",
		deadband, motor->units,
		motor->deadband.analog );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_open( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_motor_open()";

	MX_MOTOR *motor;
	MX_EPICS_MOTOR *epics_motor;
#if 0
	char pvname[60];
	double double_value;
	double acceleration_parameters[MX_MOTOR_NUM_ACCELERATION_PARAMS];
#endif
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

#if 0
	/* The following is commented out because it takes too long.
	 * In one benchmark test, each call to EPICS was taking
	 * around 120 milliseconds.  At MRCAT, with 40 EPICS-controlled
	 * motors (as of Sept. 2001), this was adding around 20 extra
	 * seconds to the startup time of each MX client.
	 */

	/* Here we find out if this EPICS record really exists in an
	 * IOC's database by asking for the record's version number.
	 */

	sprintf( pvname, "%s.VERS", epics_motor->epics_record_name );

	mx_status = mx_caget( pvname, ezcaDouble, 1, &double_value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2, ("(%s) epics motor version number = '%g'",
		epics_motor->epics_record_name, double_value));

	epics_motor->epics_record_version = double_value;

	/* Update the values of the speed, base speed, and acceleration
	 * from the values in the EPICS server.
	 */

	mx_status = mx_motor_get_speed( record, &double_value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_motor_get_base_speed( record, &double_value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_motor_get_acceleration_parameters( record,
						acceleration_parameters );

	return mx_status;
#else
	return MX_SUCCESSFUL_RESULT;
#endif
}

MX_EXPORT mx_status_type
mxd_epics_motor_close( MX_RECORD *record )
{
	/* We need do nothing special here.  The connection to EPICS
	 * will go away automatically when the program exits.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_motor_is_busy()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	short done_moving_field;
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	motor->busy = FALSE;

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".DMOV" );

	ezca_status = ezcaGet( pvname, ezcaShort, 1, &done_moving_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to read motion status of motor '%s' failed. EZCA status = %d (%s )",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	switch ( done_moving_field ) {
	case 0:
		motor->busy = TRUE;
		break;
	case 1:
		motor->busy = FALSE;
		break;
	default:
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Done moving field (.DMOV) for motor %s had an unexpected value of %d",
			epics_motor->epics_record_name, done_moving_field );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_move_absolute()";

	MX_EPICS_MOTOR *epics_motor;
	double new_destination;
	char pvname[60];
	chid *motor_channel_id;
	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status, ca_status;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	new_destination = motor->raw_destination.analog;

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".VAL" );

	/* A ezcaPut() call here would attempt to synchronously wait for
	 * the end of the motor move, which is not what we want at all.
	 * We must dive down into raw Channel Access to be able to start
	 * a motor without waiting for the move to complete.
	 */

	ezca_status = ezcaPvToChid( pvname, &motor_channel_id );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to get EPICS channel id for '%s' failed. EZCA status = %d (%s )",
			pvname, ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Send the move command to Channel Access. */

	ca_status = ca_put( DBR_DOUBLE, *motor_channel_id, &new_destination );

	if ( ca_status != ECA_NORMAL ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Attempt to move motor '%s' via EPICS CA failed. CA status = %d",
			epics_motor->epics_record_name, ca_status );
	}

	/* Flush out all pending IO requests in Channel Access to ensure
	 * that the motor starts moving as soon as possible.
	 */

	ca_status = ca_flush_io();

	if ( ca_status != ECA_NORMAL ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Attempt to send buffered Channel Access commands for motor '%s' failed. "
"CA status = %d",
			epics_motor->epics_record_name, ca_status );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_get_position()";

	MX_EPICS_MOTOR *epics_motor;
	double present_value;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".RBV" );

	ezca_status = ezcaGet( pvname, ezcaDouble, 1, &present_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to get position of EPICS motor %s failed. EZCA status = %d (%s )",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	motor->raw_position.analog = present_value;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_set_position( MX_MOTOR *motor )
{
	/* Setting the motor position is a three step process:
	 * 
	 * 1.  .SET <-- 1
	 * 2.  .VAL <-- new value
	 * 3.  .SET <-- 0
	 */

	const char fname[] = "mxd_epics_motor_set_position()";

	MX_EPICS_MOTOR *epics_motor;
	double user_set_position;
	short set_flag;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	user_set_position = motor->raw_set_position.analog;

	/*** Change the motor record from 'Use' to 'Set' mode. ***/

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".SET" );

	set_flag = 1;

	ezca_status = ezcaPut( pvname, ezcaShort, 1, &set_flag );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to change EPICS motor '%s' to Set mode failed. EZCA status = %d (%s )",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/*** Change the user position in the EPICS database. ***/

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-6 );
	strcat( pvname, ".VAL" );

	ezca_status = ezcaPut(pvname, ezcaDouble, 1, &user_set_position);

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to set user position for EPICS motor '%s' failed.  "
"EZCA status = %d (%s )",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/*** Put the motor record back into 'Use' mode. ***/

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".SET" );

	set_flag = 0;

	ezca_status = ezcaPut( pvname, ezcaShort, 1, &set_flag );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to change EPICS motor '%s' to Use mode failed. EZCA status = %d (%s )",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

/* mxd_epics_motor_soft_abort() stops the motor in a way that
 * allows the motor to immediately be moved again without taking
 * any special effort.
 */

MX_EXPORT mx_status_type
mxd_epics_motor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_soft_abort()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short stop_field;
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".STOP" );

	/* Stop the motor. */

	stop_field = 1;		/* Stop */

	ezca_status = ezcaPut( pvname, ezcaShort, 1, &stop_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to soft abort EPICS motor '%s' failed. EZCA status = %d (%s )",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* The .STOP field is automatically reset to 0 by the IOC database,
	 * so the motor is immediately ready to move again.
	 */

	return MX_SUCCESSFUL_RESULT;
}

/* mxd_epics_motor_immediate_abort() stops the motor in a way that
 * requires manual intervention (for example, via an MEDM screen)
 * to reset the .SMPG field back to "Go".
 */

MX_EXPORT mx_status_type
mxd_epics_motor_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_immediate_abort()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short smpg_field;
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".SPMG" );

	/* Stop the motor. */

	smpg_field = 0;		/* Stop */

	ezca_status = ezcaPut( pvname, ezcaShort, 1, &smpg_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to immediately abort EPICS motor '%s' failed. EZCA status = %d (%s )",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_positive_limit_hit()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short hls_field;
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".HLS" );

	ezca_status = ezcaGet( pvname, ezcaShort, 1, &hls_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to check positive limit for EPICS motor '%s' failed. "
	"EZCA status = %d (%s )",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	if ( hls_field == 0 ) {
		motor->positive_limit_hit = FALSE;
	} else {
		motor->positive_limit_hit = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_negative_limit_hit()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short lls_field;
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".LLS" );

	ezca_status = ezcaGet( pvname, ezcaShort, 1, &lls_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to check negative limit for EPICS motor '%s' failed. "
	"EZCA status = %d (%s )",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	if ( lls_field == 0 ) {
		motor->negative_limit_hit = FALSE;
	} else {
		motor->negative_limit_hit = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_find_home_position()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short home_value;
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );

	if ( motor->home_search >= 0 ) {
		strcat( pvname, ".HOMF" );
	} else {
		strcat( pvname, ".HOMR" );
	}

	/* Send the home command without waiting for it to finish. */

	home_value = 1;

	ezca_status = ezcaPutOldCa( pvname, ezcaShort, 1, &home_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to perform a home search with EPICS motor '%s' failed.  "
	"EZCA status = %d (%s )",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_constant_velocity_move( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_constant_velocity_move()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short move_command;
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );

	if ( motor->constant_velocity_move >= 0 ) {
		strcat( pvname, ".JOGF" );
	} else {
		strcat( pvname, ".JOGR" );
	}

	/* Send the move command without waiting for it to finish. */

	move_command = 1;

	ezca_status = ezcaPutOldCa( pvname, ezcaShort, 1, &move_command );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to perform a constant velocity move search with EPICS motor '%s' "
"failed.  EZCA status = %d (%s )",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_get_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_get_parameter()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	float float_value;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for motor '%s' for parameter type '%s' (%d).",
		fname, motor->record->name,
		mx_get_field_label_string( motor->record,
			motor->parameter_type ),
		motor->parameter_type));

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		sprintf( pvname, "%s.VELO", epics_motor->epics_record_name );

		MX_DEBUG( 2,("%s: about to get '%s'", fname, pvname));

		mx_status = mx_caget( pvname, ezcaFloat, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG( 2,("%s: '%s' value = %g", fname,
						pvname, float_value));

		motor->raw_speed = float_value;
		break;

	case MXLV_MTR_BASE_SPEED:
		sprintf( pvname, "%s.VBAS", epics_motor->epics_record_name );

		MX_DEBUG( 2,("%s: about to get '%s'", fname, pvname));

		mx_status = mx_caget( pvname, ezcaFloat, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG( 2,("%s: '%s' value = %g", fname,
						pvname, float_value));

		motor->raw_base_speed = float_value;
		break;

	case MXLV_MTR_ACCELERATION_PARAMETERS:
		sprintf( pvname, "%s.ACCL", epics_motor->epics_record_name );

		MX_DEBUG( 2,("%s: about to get '%s'", fname, pvname));

		mx_status = mx_caget( pvname, ezcaFloat, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG( 2,("%s: '%s' value = %g", fname,
						pvname, float_value));

		motor->acceleration_parameters[0] = float_value;
		motor->acceleration_parameters[1] = 0.0;
		motor->acceleration_parameters[2] = 0.0;
		motor->acceleration_parameters[3] = 0.0;
		break;
	default:
		return mx_motor_default_get_parameter_handler( motor );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_set_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_set_parameter()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	float float_value;
	mx_status_type mx_status;

	mx_status = mxd_epics_motor_get_pointers( motor, &epics_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for motor '%s' for parameter type '%s' (%d).",
		fname, motor->record->name,
		mx_get_field_label_string( motor->record,
			motor->parameter_type ),
		motor->parameter_type));

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		sprintf( pvname, "%s.VELO", epics_motor->epics_record_name );

		float_value = motor->raw_speed;

		MX_DEBUG( 2,("%s: sending %g to '%s'.", fname,
					float_value, pvname));

		mx_status = mx_caput( pvname, ezcaFloat, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		break;
	case MXLV_MTR_BASE_SPEED:
		sprintf( pvname, "%s.VBAS", epics_motor->epics_record_name );

		float_value = motor->raw_base_speed;

		MX_DEBUG( 2,("%s: sending %g to '%s'.", fname,
					float_value, pvname));

		mx_status = mx_caput( pvname, ezcaFloat, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;
	case MXLV_MTR_ACCELERATION_PARAMETERS:
		sprintf( pvname, "%s.ACCL", epics_motor->epics_record_name );

		float_value = motor->acceleration_parameters[0];

		MX_DEBUG( 2,("%s: sending %g to '%s'.", fname,
					float_value, pvname));

		mx_status = mx_caput( pvname, ezcaFloat, 1, &float_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		break;
	default:
		return mx_motor_default_set_parameter_handler( motor );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */

