/*
 * Name:    d_mcs_mce.c
 *
 * Purpose: MX MCE driver to use pairs of MCS channels as incremental
 *          encoders.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2000-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_measurement.h"
#include "mx_mce.h"
#include "mx_mcs.h"
#include "d_mcs_mce.h"

/* Initialize the MCE driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_mcs_encoder_record_function_list = {
	mxd_mcs_encoder_initialize_type,
	mxd_mcs_encoder_create_record_structures,
	mxd_mcs_encoder_finish_record_initialization,
	mxd_mcs_encoder_delete_record,
	NULL,
	mxd_mcs_encoder_read_parms_from_hardware,
	mxd_mcs_encoder_write_parms_to_hardware,
	mxd_mcs_encoder_open,
	mxd_mcs_encoder_close
};

MX_MCE_FUNCTION_LIST mxd_mcs_encoder_mce_function_list = {
	mxd_mcs_encoder_get_overflow_status,
	mxd_mcs_encoder_reset_overflow_status,
	mxd_mcs_encoder_read,
	mxd_mcs_encoder_get_current_num_values
};

/* MCS encoder data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_mcs_encoder_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_MCE_STANDARD_FIELDS,
	MXD_MCS_ENCODER_STANDARD_FIELDS
};

long mxd_mcs_encoder_num_record_fields
		= sizeof( mxd_mcs_encoder_record_field_defaults )
		  / sizeof( mxd_mcs_encoder_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_mcs_encoder_rfield_def_ptr
			= &mxd_mcs_encoder_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_mcs_encoder_get_pointers( MX_MCE *mce,
			MX_MCS_ENCODER **mcs_encoder,
			MX_MCS **mcs,
			const char *calling_fname )
{
	const char fname[] = "mxd_mcs_encoder_get_pointers()";

	if ( mce == (MX_MCE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCE pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( mcs_encoder == (MX_MCS_ENCODER **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_MCS_ENCODER pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*mcs_encoder = (MX_MCS_ENCODER *) (mce->record->record_type_struct);

	if ( *mcs_encoder == (MX_MCS_ENCODER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCS_ENCODER pointer for record '%s' is NULL.",
			mce->record->name );
	}

	*mcs = (MX_MCS *) (*mcs_encoder)->mcs_record->record_class_struct;

	if ( *mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MCS pointer for MCS record '%s' is NULL.",
			(*mcs_encoder)->mcs_record->name );
	}
	if ( (*mcs)->data_array == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"data_array pointer for MCS record '%s' is NULL.",
			(*mcs_encoder)->mcs_record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxd_mcs_encoder_initialize_type( long record_type )
{
	long num_record_fields;
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	long maximum_num_values_varargs_cookie;
	mx_status_type mx_status;

	mx_status = mx_mce_initialize_type( record_type,
					&num_record_fields,
					&record_field_defaults,
					&maximum_num_values_varargs_cookie );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mcs_encoder_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_mcs_encoder_create_record_structures()";

	MX_MCE *mce;
	MX_MCS_ENCODER *mcs_encoder;

	/* Allocate memory for the necessary structures. */

	mce = (MX_MCE *) malloc( sizeof(MX_MCE) );

	if ( mce == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MCE structure." );
	}

	mcs_encoder = (MX_MCS_ENCODER *) malloc( sizeof(MX_MCS_ENCODER) );

	if ( mcs_encoder == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MCS_ENCODER structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = mce;
	record->record_type_struct = mcs_encoder;
	record->class_specific_function_list
			= &mxd_mcs_encoder_mce_function_list;

	mce->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_encoder_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_mcs_encoder_finish_record_initialization()";

	MX_MCE *mce;
	MX_MCS_ENCODER *mcs_encoder;
	MX_MCS *mcs;

	mce = (MX_MCE *) record->record_class_struct;

	if ( mce == (MX_MCE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCE pointer for record '%s' is NULL.",
			record->name );
	}

	mce->encoder_type = MXT_MCE_DELTA_ENCODER;

	mce->current_num_values = mce->maximum_num_values;

	/* Get a pointer to the MX_MCS structure used by this record. */

	mcs_encoder = (MX_MCS_ENCODER *) record->record_type_struct;

	if ( mcs_encoder == (MX_MCS_ENCODER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCS_ENCODER pointer for record '%s' is NULL.",
			record->name );
	}

	if ( mcs_encoder->mcs_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mcs_encoder->mcs_record pointer for record '%s' is NULL.",
			record->name );
	}

	mcs = (MX_MCS *) mcs_encoder->mcs_record->record_class_struct;

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCS pointer for multichannel scaler '%s' is NULL.",
			mcs_encoder->mcs_record->name );
	}

	/* Check to see if the scaler channel numbers listed are in
	 * the legal range.
	 */

	if ( ((long)mcs_encoder->up_channel - mcs->first_scaler_number) >=
			mcs->maximum_num_scalers )
	{
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The up channel %lu for MCS encoder '%s' is outside the allowed range %ld-%ld "
"for MCS '%s'.", mcs_encoder->up_channel, record->name,
		mcs->first_scaler_number,
		mcs->maximum_num_scalers + mcs->first_scaler_number - 1L,
		mcs_encoder->mcs_record->name );
	}

	if ( ((long)mcs_encoder->down_channel - mcs->first_scaler_number) >=
			mcs->maximum_num_scalers )
	{
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The down channel %lu for MCS encoder '%s' is outside the allowed range %ld-%ld "
"for MCS '%s'.", mcs_encoder->down_channel, record->name,
		mcs->first_scaler_number,
		mcs->maximum_num_scalers + mcs->first_scaler_number - 1L,
		mcs_encoder->mcs_record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_encoder_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_encoder_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_encoder_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_encoder_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_encoder_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_encoder_get_overflow_status( MX_MCE *mce )
{
	mce->overflow_set = FALSE;
	mce->underflow_set = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_encoder_reset_overflow_status( MX_MCE *mce )
{
	mce->overflow_set = FALSE;
	mce->underflow_set = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_encoder_read( MX_MCE *mce )
{
	const char fname[] = "mxd_mcs_encoder_read()";

	MX_MCS_ENCODER *mcs_encoder;
	MX_MCS *mcs;
	long i;
	long up_channel, down_channel;
	long up_index, down_index;
	long *up_value_array, *down_value_array;
	mx_status_type mx_status;

	mx_status = mxd_mcs_encoder_get_pointers( mce,
					&mcs_encoder, &mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for MCE '%s'",
			fname, mce->record->name));

	up_channel = (long) mcs_encoder->up_channel;
	down_channel = (long) mcs_encoder->down_channel;

	up_index = up_channel - mcs->first_scaler_number;
	down_index = down_channel - mcs->first_scaler_number;

	mx_status = mx_mcs_read_scaler( mcs_encoder->mcs_record,
						up_index, NULL, NULL );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_mcs_read_scaler( mcs_encoder->mcs_record,
						down_index, NULL, NULL );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	up_value_array = mcs->data_array[ up_index ];
	down_value_array = mcs->data_array[ down_index ];

	mce->current_num_values = (long) mcs->current_num_measurements;

	for ( i = 0; i < mce->current_num_values; i++ ) {

		mce->value_array[i] = (double)
				( up_value_array[i] - down_value_array[i] );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_encoder_get_current_num_values( MX_MCE *mce )
{
	const char fname[] = "mxd_mcs_encoder_get_current_num_values()";

	MX_MCS_ENCODER *mcs_encoder;
	MX_MCS *mcs;
	mx_status_type mx_status;

	mx_status = mxd_mcs_encoder_get_pointers( mce,
					&mcs_encoder, &mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for MCE '%s'", fname, mce->record->name));

	mce->current_num_values = (long) mcs->current_num_measurements;

	MX_DEBUG( 2,("%s: mce->current_num_values = %ld",
		fname, mce->current_num_values));

	return MX_SUCCESSFUL_RESULT;
}

