/*
 * Name:    d_network_mca.c
 *
 * Purpose: MX multichannel analyzer driver for network MCAs.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2000-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_mca.h"
#include "mx_net.h"
#include "d_network_mca.h"

/* Initialize the MCA driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_network_mca_record_function_list = {
	mxd_network_mca_initialize_type,
	mxd_network_mca_create_record_structures,
	mxd_network_mca_finish_record_initialization,
	mxd_network_mca_delete_record,
	mxd_network_mca_print_structure,
	mxd_network_mca_read_parms_from_hardware,
	mxd_network_mca_write_parms_to_hardware,
	mxd_network_mca_open,
	mxd_network_mca_close,
	NULL,
	mxd_network_mca_resynchronize
};

MX_MCA_FUNCTION_LIST mxd_network_mca_mca_function_list = {
	mxd_network_mca_start,
	mxd_network_mca_stop,
	mxd_network_mca_read,
	mxd_network_mca_clear,
	mxd_network_mca_busy,
	mxd_network_mca_get_parameter,
	mxd_network_mca_set_parameter,
};

MX_RECORD_FIELD_DEFAULTS mxd_network_mca_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_MCA_STANDARD_FIELDS,
	MXD_NETWORK_MCA_STANDARD_FIELDS
};

long mxd_network_mca_num_record_fields
		= sizeof( mxd_network_mca_record_field_defaults )
		  / sizeof( mxd_network_mca_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_network_mca_rfield_def_ptr
			= &mxd_network_mca_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_network_mca_get_pointers( MX_MCA *mca,
			MX_NETWORK_MCA **network_mca,
			const char *calling_fname )
{
	const char fname[] = "mxd_network_mca_get_pointers()";

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The mca pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( mca->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_RECORD pointer for mca pointer passed by '%s' is NULL.",
			calling_fname );
	}

	if ( mca->record->mx_type != MXT_MCA_NETWORK ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The mca '%s' passed by '%s' is not a network mca.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			mca->record->name, calling_fname,
			mca->record->mx_superclass,
			mca->record->mx_class,
			mca->record->mx_type );
	}

	if ( network_mca != (MX_NETWORK_MCA **) NULL ) {

		*network_mca = (MX_NETWORK_MCA *)
				(mca->record->record_type_struct);

		if ( *network_mca == (MX_NETWORK_MCA *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_NETWORK_MCA pointer for mca record '%s' passed by '%s' is NULL",
				mca->record->name, calling_fname );
		}
	}

	if ( (*network_mca)->server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX server record pointer for MX network mca '%s' is NULL.",
			mca->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_network_mca_initialize_type( long record_type )
{
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	long num_record_fields;
	long maximum_num_channels_varargs_cookie;
	long maximum_num_rois_varargs_cookie;
	long num_soft_rois_varargs_cookie;
	mx_status_type mx_status;

	mx_status = mx_mca_initialize_type( record_type,
				&num_record_fields,
				&record_field_defaults,
				&maximum_num_channels_varargs_cookie,
				&maximum_num_rois_varargs_cookie,
				&num_soft_rois_varargs_cookie );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_mca_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_network_mca_create_record_structures()";

	MX_MCA *mca;
	MX_NETWORK_MCA *network_mca;

	/* Allocate memory for the necessary structures. */

	mca = (MX_MCA *) malloc( sizeof(MX_MCA) );

	if ( mca == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MCA structure." );
	}

	network_mca = (MX_NETWORK_MCA *) malloc( sizeof(MX_NETWORK_MCA) );

	if ( network_mca == (MX_NETWORK_MCA *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_NETWORK_MCA structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = mca;
	record->record_type_struct = network_mca;
	record->class_specific_function_list
		= &mxd_network_mca_mca_function_list;

	mca->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mca_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_network_mca_finish_record_initialization()";

	MX_MCA *mca;
	mx_status_type mx_status;

	mca = (MX_MCA *) record->record_class_struct;

	mca->channel_array = ( unsigned long * )
		malloc( mca->maximum_num_channels * sizeof( unsigned long ) );

	if ( mca->channel_array == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory allocating an %lu channel data array.",
			mca->maximum_num_channels );
	}

	mx_status = mx_mca_finish_record_initialization( record );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_mca_delete_record( MX_RECORD *record )
{
	MX_MCA *mca;

	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {

		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	mca = (MX_MCA *) record->record_class_struct;

	if ( mca != NULL ) {

		if ( mca->channel_array != NULL ) {
			free( mca->channel_array );

			mca->channel_array = NULL;
		}
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mca_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_network_mca_print_structure()";

	MX_MCA *mca;
	MX_NETWORK_MCA *network_mca;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed is NULL.");
	}

	mca = (MX_MCA *) (record->record_class_struct);

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_MCA pointer for record '%s' is NULL.", record->name);
	}

	network_mca = (MX_NETWORK_MCA *) (record->record_type_struct);

	if ( network_mca == (MX_NETWORK_MCA *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_NETWORK_MCA pointer for record '%s' is NULL.", record->name);
	}

	fprintf(file, "MCA parameters for record '%s':\n", record->name);

	fprintf(file, "  MCA type              = NETWORK_MCA.\n\n");
	fprintf(file, "  server                = %s\n",
					network_mca->server_record->name);
	fprintf(file, "  remote record         = %s\n",
					network_mca->remote_record_name);
	fprintf(file, "  maximum # of channels = %lu\n",
					mca->maximum_num_channels);

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mca_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mca_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mca_open( MX_RECORD *record )
{
	const char fname[] = "mxd_network_mca_open()";

	MX_MCA *mca;
	MX_NETWORK_MCA *network_mca;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	unsigned long i;
	long remote_num_soft_rois;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	mca = (MX_MCA *) record->record_class_struct;

	mx_status = mxd_network_mca_get_pointers(mca, &network_mca, fname);

	if ( mx_status.code != MXE_SUCCESS ) {
		mca->busy = FALSE;

		return mx_status;
	}

	/* Set some reasonable defaults. */

	mca->current_num_channels = mca->maximum_num_channels;
	mca->current_num_rois = mca->maximum_num_rois;

	for ( i = 0; i < mca->maximum_num_rois; i++ ) {
		mca->roi_array[i][0] = 0;
		mca->roi_array[i][1] = 0;
	}

	mca->preset_type = MXF_MCA_PRESET_LIVE_TIME;
	mca->parameter_type = 0;

	mca->roi[0] = 0;
	mca->roi[1] = 0;
	mca->roi_integral = 0;

	mca->channel_number = 0;
	mca->channel_value = 0;

	mca->real_time = 0.0;
	mca->live_time = 0.0;

	mca->preset_real_time = 0.0;
	mca->preset_live_time = 0.0;

	/* Check to see if the remote value of 'num_soft_rois' is smaller
	 * than the local value.
	 */

	sprintf( rfname, "%s.num_soft_rois", network_mca->remote_record_name );

	mx_status = mx_get( network_mca->server_record,
				rfname, MXFT_LONG, &remote_num_soft_rois );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( remote_num_soft_rois < mca->num_soft_rois ) {

		/* If the remote record has fewer soft ROIs than the
		 * local record, unilaterally reduce the local value
		 * 'num_soft_rois'.  This wastes local memory, but
		 * is safer.
		 */

		mx_warning(
	"'num_soft_rois' (%ld) for record '%s' is larger than the value of "
	"remote record field '%s' (%ld) on server '%s'.  The local value of "
	"'num_soft_rois' will be reduced to %ld.",
			mca->num_soft_rois, record->name,
			rfname, remote_num_soft_rois,
			network_mca->server_record->name,
			remote_num_soft_rois );

		mca->num_soft_rois = remote_num_soft_rois;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mca_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mca_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxd_network_mca_resynchronize()";

	MX_MCA *mca;
	MX_NETWORK_MCA *network_mca;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int resynchronize;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	mca = (MX_MCA *) record->record_class_struct;

	mx_status = mxd_network_mca_get_pointers(mca, &network_mca, fname);

	if ( mx_status.code != MXE_SUCCESS ) {
		mca->busy = FALSE;

		return mx_status;
	}

	sprintf(rfname, "%s.resynchronize", network_mca->remote_record_name);

	resynchronize = 1;

	mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_INT, &resynchronize );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_mca_start( MX_MCA *mca )
{
	const char fname[] = "mxd_network_mca_start()";

	MX_NETWORK_MCA *network_mca;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long value;
	mx_status_type mx_status;

	mx_status = mxd_network_mca_get_pointers( mca, &network_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Set the preset type. */

	sprintf( rfname, "%s.preset_type", network_mca->remote_record_name );

	value = mca->preset_type;

	mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_INT, &value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Set the preset interval appropriately depending on the
	 * preset type.
	 */

	switch( mca->preset_type ) {
	case MXF_MCA_PRESET_NONE:
		/* Do nothing. */

		break;
	case MXF_MCA_PRESET_LIVE_TIME:
		sprintf( rfname, "%s.preset_live_time",
					network_mca->remote_record_name );

		mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_DOUBLE,
				&(mca->preset_live_time) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;
	case MXF_MCA_PRESET_REAL_TIME:
		sprintf( rfname, "%s.preset_real_time",
					network_mca->remote_record_name );

		mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_DOUBLE,
				&(mca->preset_real_time) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unsupported preset type %d for MCA '%s'.",
			mca->preset_type, mca->record->name );
		break;
	}

	/* Tell the counting to start. */

	value = 1;

	sprintf( rfname, "%s.start", network_mca->remote_record_name );

	mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_INT, &value );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_mca_stop( MX_MCA *mca )
{
	const char fname[] = "mxd_network_mca_stop()";

	MX_NETWORK_MCA *network_mca;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long value;
	mx_status_type mx_status;

	mx_status = mxd_network_mca_get_pointers( mca, &network_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	value = 1;

	sprintf( rfname, "%s.stop", network_mca->remote_record_name );

	mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_INT, &value );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_mca_read( MX_MCA *mca )
{
	const char fname[] = "mxd_network_mca_read()";

	MX_NETWORK_MCA *network_mca;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long dimension_array[1];
	unsigned long num_channels;
	mx_status_type mx_status;

	mx_status = mxd_network_mca_get_pointers( mca, &network_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_mca_get_num_channels( mca->record, &num_channels );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.channel_array", network_mca->remote_record_name );

	dimension_array[0] = (long) num_channels;

	mx_status = mx_get_array( network_mca->server_record,
			rfname, MXFT_ULONG, 1, dimension_array,
				mca->channel_array );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_mca_clear( MX_MCA *mca )
{
	const char fname[] = "mxd_network_mca_clear()";

	MX_NETWORK_MCA *network_mca;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long value;
	mx_status_type mx_status;

	mx_status = mxd_network_mca_get_pointers( mca, &network_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	value = 1;

	sprintf( rfname, "%s.clear", network_mca->remote_record_name );

	mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_INT, &value );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_mca_busy( MX_MCA *mca )
{
	const char fname[] = "mxd_network_mca_busy()";

	MX_NETWORK_MCA *network_mca;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	int busy;
	mx_status_type mx_status;

	mx_status = mxd_network_mca_get_pointers( mca, &network_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.busy", network_mca->remote_record_name );

	mx_status = mx_get( network_mca->server_record,
				rfname, MXFT_INT, &busy );

	mca->busy = busy;

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_mca_get_parameter( MX_MCA *mca )
{
	const char fname[] = "mxd_network_mca_get_parameter()";

	MX_NETWORK_MCA *network_mca;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long dimension_array[2];
	mx_status_type mx_status;

	mx_status = mxd_network_mca_get_pointers( mca, &network_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( mca->parameter_type == MXLV_MCA_CURRENT_NUM_CHANNELS ) {

		sprintf( rfname, "%s.current_num_channels",
				network_mca->remote_record_name );

		mx_status = mx_get( network_mca->server_record,
				rfname, MXFT_LONG,
				&(mca->current_num_channels) );

		if ( mca->current_num_channels > mca->maximum_num_channels ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"The MCA '%s' controlled by server '%s' is reported to have %lu channels, "
"but the record '%s' is only configured to support up to %lu channels.",
				network_mca->remote_record_name,
				network_mca->server_record->name,
				mca->current_num_channels,
				mca->record->name,
				mca->maximum_num_channels );
		}
	} else
	if ( mca->parameter_type == MXLV_MCA_ROI_ARRAY ) {

		sprintf( rfname, "%s.current_num_rois",
					network_mca->remote_record_name );

		mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_LONG, &(mca->current_num_rois) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.roi_array",
					network_mca->remote_record_name );

		dimension_array[0] = mca->current_num_rois;
		dimension_array[1] = 2;

		mx_status = mx_get_array( network_mca->server_record,
				rfname, MXFT_ULONG, 2, dimension_array,
				&(mca->roi_array) );
	} else
	if ( mca->parameter_type == MXLV_MCA_ROI_INTEGRAL_ARRAY ) {

		sprintf( rfname, "%s.current_num_rois",
					network_mca->remote_record_name );

		mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_LONG, &(mca->current_num_rois) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.roi_integral_array",
					network_mca->remote_record_name );

		dimension_array[0] = mca->current_num_rois;

		mx_status = mx_get_array( network_mca->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
				&(mca->roi_integral_array) );
	} else
	if ( mca->parameter_type == MXLV_MCA_ROI ) {

		sprintf( rfname, "%s.roi_number",
					network_mca->remote_record_name );

		mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_ULONG, &(mca->roi_number) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.roi", network_mca->remote_record_name );

		dimension_array[0] = 2;

		mx_status = mx_get_array( network_mca->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
				&(mca->roi) );

	} else
	if ( mca->parameter_type == MXLV_MCA_ROI_INTEGRAL ) {

		sprintf( rfname, "%s.roi_number",
					network_mca->remote_record_name );

		mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_ULONG, &(mca->roi_number) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.roi_integral",
				network_mca->remote_record_name );

		mx_status = mx_get( network_mca->server_record,
				rfname, MXFT_ULONG,
				&(mca->roi_integral) );

	} else
	if ( mca->parameter_type == MXLV_MCA_SOFT_ROI_ARRAY ) {

		sprintf( rfname, "%s.soft_roi_array",
					network_mca->remote_record_name );

		dimension_array[0] = mca->num_soft_rois;
		dimension_array[1] = 2;

		mx_status = mx_get_array( network_mca->server_record,
				rfname, MXFT_ULONG, 2, dimension_array,
				&(mca->soft_roi_array) );
	} else
	if ( mca->parameter_type == MXLV_MCA_SOFT_ROI_INTEGRAL_ARRAY ) {

		sprintf( rfname, "%s.soft_roi_integral_array",
					network_mca->remote_record_name );

		dimension_array[0] = mca->num_soft_rois;

		mx_status = mx_get_array( network_mca->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
				&(mca->soft_roi_integral_array) );
	} else
	if ( mca->parameter_type == MXLV_MCA_SOFT_ROI ) {

		sprintf( rfname, "%s.soft_roi_number",
					network_mca->remote_record_name );

		mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_ULONG, &(mca->soft_roi_number) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.soft_roi",
					network_mca->remote_record_name );

		dimension_array[0] = 2;

		mx_status = mx_get_array( network_mca->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
				&(mca->soft_roi) );

	} else
	if ( mca->parameter_type == MXLV_MCA_SOFT_ROI_INTEGRAL ) {

		sprintf( rfname, "%s.soft_roi_number",
					network_mca->remote_record_name );

		mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_ULONG, &(mca->soft_roi_number) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.soft_roi_integral",
				network_mca->remote_record_name );

		mx_status = mx_get( network_mca->server_record,
				rfname, MXFT_ULONG,
				&(mca->soft_roi_integral) );

	} else
	if ( mca->parameter_type == MXLV_MCA_CHANNEL_NUMBER ) {

		sprintf( rfname, "%s.channel_number",
				network_mca->remote_record_name );

		mx_status = mx_get( network_mca->server_record,
				rfname, MXFT_ULONG,
				&(mca->channel_number) );

	} else
	if ( mca->parameter_type == MXLV_MCA_CHANNEL_VALUE ) {

		sprintf( rfname, "%s.channel_value",
				network_mca->remote_record_name );

		mx_status = mx_get( network_mca->server_record,
				rfname, MXFT_ULONG,
				&(mca->channel_value) );

	} else
	if ( mca->parameter_type == MXLV_MCA_REAL_TIME ) {

		sprintf( rfname, "%s.real_time",
				network_mca->remote_record_name );

		mx_status = mx_get( network_mca->server_record,
				rfname, MXFT_DOUBLE,
				&(mca->real_time) );

	} else
	if ( mca->parameter_type == MXLV_MCA_LIVE_TIME ) {

		sprintf( rfname, "%s.live_time",
				network_mca->remote_record_name );

		mx_status = mx_get( network_mca->server_record,
				rfname, MXFT_DOUBLE,
				&(mca->live_time) );

	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			mca->parameter_type );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mca_set_parameter( MX_MCA *mca )
{
	const char fname[] = "mxd_network_mca_set_parameter()";

	MX_NETWORK_MCA *network_mca;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long dimension_array[2];
	mx_status_type mx_status;

	mx_status = mxd_network_mca_get_pointers( mca, &network_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( mca->parameter_type == MXLV_MCA_ROI_ARRAY ) {

		sprintf( rfname, "%s.current_num_rois",
					network_mca->remote_record_name );

		mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_LONG, &(mca->current_num_rois) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.roi_array",
					network_mca->remote_record_name );

		dimension_array[0] = mca->current_num_rois;
		dimension_array[1] = 2;

		mx_status = mx_put_array( network_mca->server_record,
				rfname, MXFT_ULONG, 2, dimension_array,
				&(mca->roi_array) );

	} else
	if ( mca->parameter_type == MXLV_MCA_ROI ) {

		sprintf( rfname, "%s.roi_number",
					network_mca->remote_record_name );

		mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_ULONG, &(mca->roi_number) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.roi", network_mca->remote_record_name );

		dimension_array[0] = 2;

		mx_status = mx_put_array( network_mca->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
				&(mca->roi) );

	} else
	if ( mca->parameter_type == MXLV_MCA_SOFT_ROI_ARRAY ) {

		sprintf( rfname, "%s.soft_roi_array",
					network_mca->remote_record_name );

		dimension_array[0] = mca->num_soft_rois;
		dimension_array[1] = 2;

		mx_status = mx_put_array( network_mca->server_record,
				rfname, MXFT_ULONG, 2, dimension_array,
				&(mca->soft_roi_array) );

	} else
	if ( mca->parameter_type == MXLV_MCA_SOFT_ROI ) {

		sprintf( rfname, "%s.soft_roi_number",
					network_mca->remote_record_name );

		mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_ULONG, &(mca->soft_roi_number) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.soft_roi",
					network_mca->remote_record_name );

		dimension_array[0] = 2;

		mx_status = mx_put_array( network_mca->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
				&(mca->soft_roi) );

	} else
	if ( mca->parameter_type == MXLV_MCA_CHANNEL_NUMBER ) {

		sprintf( rfname, "%s.channel_number",
				network_mca->remote_record_name );

		mx_status = mx_put( network_mca->server_record,
				rfname, MXFT_ULONG,
				&(mca->channel_number) );

	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			mca->parameter_type );
	}

	return MX_SUCCESSFUL_RESULT;
}

