/*
 * Name:    d_network_mcs.c 
 *
 * Purpose: MX multichannel scaler driver for MX network multichannel scalers.
 *
 * Author:  William Lavender
 *
 *----------------------------------------------------------------------------
 *
 * Copyright 2000-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mxconfig.h"

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_mcs.h"
#include "mx_net.h"
#include "d_network_mcs.h"

/* Initialize the mcs driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_network_mcs_record_function_list = {
	mxd_network_mcs_initialize_type,
	mxd_network_mcs_create_record_structures,
	mxd_network_mcs_finish_record_initialization,
	mxd_network_mcs_delete_record,
	mxd_network_mcs_print_structure,
	mxd_network_mcs_read_parms_from_hardware,
	mxd_network_mcs_write_parms_to_hardware,
	mxd_network_mcs_open,
	mxd_network_mcs_close
};

MX_MCS_FUNCTION_LIST mxd_network_mcs_mcs_function_list = {
	mxd_network_mcs_start,
	mxd_network_mcs_stop,
	mxd_network_mcs_clear,
	mxd_network_mcs_busy,
	mxd_network_mcs_read_all,
	mxd_network_mcs_read_scaler,
	mxd_network_mcs_read_measurement,
	mxd_network_mcs_read_timer,
	mxd_network_mcs_get_parameter,
	mxd_network_mcs_set_parameter
};

/* EPICS mcs data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_network_mcs_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_MCS_STANDARD_FIELDS,
	MXD_NETWORK_MCS_STANDARD_FIELDS
};

long mxd_network_mcs_num_record_fields
		= sizeof( mxd_network_mcs_record_field_defaults )
			/ sizeof( mxd_network_mcs_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_network_mcs_rfield_def_ptr
			= &mxd_network_mcs_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_network_mcs_get_pointers( MX_MCS *mcs,
			MX_NETWORK_MCS **network_mcs,
			const char *calling_fname )
{
	const char fname[] = "mxd_network_mcs_get_pointers()";

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCS pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( network_mcs == (MX_NETWORK_MCS **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_NETWORK_MCS pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( mcs->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
    "The MX_RECORD pointer for the MX_MCS pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*network_mcs = (MX_NETWORK_MCS *) mcs->record->record_type_struct;

	if ( *network_mcs == (MX_NETWORK_MCS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_NETWORK_MCS pointer for record '%s' is NULL.",
			mcs->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_network_mcs_initialize_type( long record_type )
{
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	long num_record_fields;
	long maximum_num_scalers_varargs_cookie;
	long maximum_num_measurements_varargs_cookie;
	mx_status_type status;

	status = mx_mcs_initialize_type( record_type,
				&num_record_fields,
				&record_field_defaults,
				&maximum_num_scalers_varargs_cookie,
				&maximum_num_measurements_varargs_cookie );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_mcs_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_network_mcs_create_record_structures()";

	MX_MCS *mcs;
	MX_NETWORK_MCS *network_mcs;

	/* Allocate memory for the necessary structures. */

	mcs = (MX_MCS *) malloc( sizeof(MX_MCS) );

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MCS structure." );
	}

	network_mcs = (MX_NETWORK_MCS *) malloc( sizeof(MX_NETWORK_MCS) );

	if ( network_mcs == (MX_NETWORK_MCS *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_NETWORK_MCS structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = mcs;
	record->record_type_struct = network_mcs;
	record->class_specific_function_list
				= &mxd_network_mcs_mcs_function_list;

	mcs->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_finish_record_initialization( MX_RECORD *record )
{
#if 0
	const char fname[] = "mxd_network_mcs_finish_record_initialization()";
#endif

	MX_MCS *mcs;

	mx_status_type mx_status;

	mx_status = mx_mcs_finish_record_initialization( record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mcs = (MX_MCS *) record->record_class_struct;

	mcs->first_scaler_number = 0;

#if 0
	MX_DEBUG(-2,("%s: record = %p", fname, record));
	MX_DEBUG(-2,("%s: record->name = '%s'", fname, record->name));
	MX_DEBUG(-2,("%s: mcs = %p", fname, mcs));
	MX_DEBUG(-2,("%s: mcs->data_array = %p", fname, mcs->data_array));
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_network_mcs_print_structure()";

	MX_MCS *mcs;
	MX_NETWORK_MCS *network_mcs;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed is NULL.");
	}

	mcs = (MX_MCS *) (record->record_class_struct);

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_MCS pointer for record '%s' is NULL.", record->name);
	}

	network_mcs = (MX_NETWORK_MCS *) (record->record_type_struct);

	if ( network_mcs == (MX_NETWORK_MCS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_NETWORK_MCS pointer for record '%s' is NULL.", record->name);
	}

	fprintf(file, "MCS parameters for record '%s':\n", record->name);

	fprintf(file, "  MCS type                  = NETWORK_MCS.\n\n");
	fprintf(file, "  server                    = %s\n",
					network_mcs->server_record->name);
	fprintf(file, "  remote record             = %s\n",
					network_mcs->remote_record_name);
	fprintf(file, "  maximum # of scalers      = %lu\n",
					mcs->maximum_num_scalers);
	fprintf(file, "  maximum # of measurements = %lu\n",
					mcs->maximum_num_scalers);

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxd_network_mcs_start( MX_MCS *mcs )
{
	const char fname[] = "mxd_network_mcs_start()";

	MX_NETWORK_MCS *network_mcs;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long start;
	mx_status_type mx_status;

	mx_status = mxd_network_mcs_get_pointers( mcs, &network_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.start", network_mcs->remote_record_name );

	MX_DEBUG( 2,("%s invoked for record field '%s'.", fname, rfname));

	start = 1;

	mx_status = mx_put( network_mcs->server_record,
				rfname, MXFT_LONG, &start );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_mcs_stop( MX_MCS *mcs )
{
	const char fname[] = "mxd_network_mcs_stop()";

	MX_NETWORK_MCS *network_mcs;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long stop;
	mx_status_type mx_status;

	mx_status = mxd_network_mcs_get_pointers( mcs, &network_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.stop", network_mcs->remote_record_name );

	MX_DEBUG( 2,("%s invoked for record field '%s'.", fname, rfname));

	stop = 1;

	mx_status = mx_put( network_mcs->server_record,
				rfname, MXFT_LONG, &stop );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_mcs_clear( MX_MCS *mcs )
{
	const char fname[] = "mxd_network_mcs_clear()";

	MX_NETWORK_MCS *network_mcs;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long clear;
	mx_status_type mx_status;

	mx_status = mxd_network_mcs_get_pointers( mcs, &network_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.clear", network_mcs->remote_record_name );

	MX_DEBUG( 2,("%s invoked for record field '%s'.", fname, rfname));

	clear = 1;

	mx_status = mx_put( network_mcs->server_record,
				rfname, MXFT_LONG, &clear );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_mcs_busy( MX_MCS *mcs )
{
	const char fname[] = "mxd_network_mcs_busy()";

	MX_NETWORK_MCS *network_mcs;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	int busy;
	mx_status_type mx_status;

	mx_status = mxd_network_mcs_get_pointers( mcs, &network_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.busy", network_mcs->remote_record_name );

	MX_DEBUG( 2,("%s invoked for record field '%s'.", fname, rfname));

	mx_status = mx_get( network_mcs->server_record,
				rfname, MXFT_INT, &busy );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( busy ) {
		mcs->busy = TRUE;
	} else {
		mcs->busy = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_read_all( MX_MCS *mcs )
{
	const char fname[] = "mxd_network_mcs_read_all()";

	long i;
	mx_status_type mx_status;

	if ( mcs == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MCS pointer passed was NULL." );
	}

	for ( i = 0; i < mcs->current_num_scalers; i++ ) {

		mcs->scaler_index = i;

		mx_status = mxd_network_mcs_read_scaler( mcs );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_read_scaler( MX_MCS *mcs )
{
	const char fname[] = "mxd_network_mcs_read_scaler()";

	MX_NETWORK_MCS *network_mcs;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long dimension_array[1];
	long *data_ptr;
	mx_status_type mx_status;

	mx_status = mxd_network_mcs_get_pointers( mcs, &network_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.scaler_index", network_mcs->remote_record_name );

	MX_DEBUG( 2,("%s is accessing record field '%s'.", fname, rfname));

	mx_status = mx_put( network_mcs->server_record,
				rfname, MXFT_LONG, &(mcs->scaler_index) );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.scaler_data", network_mcs->remote_record_name );

	MX_DEBUG( 2,("%s is accessing record field '%s'.", fname, rfname));

#if 0
	MX_DEBUG(-2,("%s: mcs->record = %p", fname, mcs->record));
	MX_DEBUG(-2,("%s: mcs->record->name = '%s'", fname, mcs->record->name));
	MX_DEBUG(-2,("%s: mcs = %p", fname, mcs));
	MX_DEBUG(-2,("%s: mcs->data_array = %p", fname, mcs->data_array));
	MX_DEBUG(-2,("%s: mcs->scaler_index = %ld", fname, mcs->scaler_index));
#endif

	data_ptr = mcs->data_array[ mcs->scaler_index ];

#if 0
	MX_DEBUG(-2,("%s: data_ptr = %p", fname, data_ptr));

	MX_DEBUG(-2,("%s: mcs->num_measurements = %ld",
			fname, mcs->current_num_measurements));
#endif

	MX_DEBUG( 1,("%s: reading '%s'.", fname, rfname));

	dimension_array[0] = (long) mcs->current_num_measurements;

	mx_status = mx_get_array( network_mcs->server_record,
				rfname, MXFT_LONG, 1, dimension_array,
				data_ptr );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 1,("%s: data received for '%s'.", fname, rfname));

#if 0
	{
		long i;

		for ( i = 0; i < mcs->current_num_measurements; i++ ) {
			fprintf(stderr,"%ld ", data_ptr[i]);
		}
		fprintf(stderr,"\n");
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_read_measurement( MX_MCS *mcs )
{
	const char fname[] = "mxd_network_mcs_read_measurement()";

	MX_NETWORK_MCS *network_mcs;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long dimension_array[1];
	mx_status_type mx_status;

	mx_status = mxd_network_mcs_get_pointers( mcs, &network_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.measurement_index",
				network_mcs->remote_record_name );

	mx_status = mx_put( network_mcs->server_record,
				rfname, MXFT_LONG, &(mcs->measurement_index) );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.measurement_data",
				network_mcs->remote_record_name );

	MX_DEBUG( 1,("%s: reading '%s'.", fname, rfname));

	dimension_array[0] = (long) mcs->current_num_scalers;

	mx_status = mx_get_array( network_mcs->server_record,
				rfname, MXFT_LONG, 1, dimension_array,
				mcs->measurement_data );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 1,("%s: data received for '%s'.", fname, rfname));

#if 0
	{
		long i;

		for ( i = 0; i < mcs->current_num_scalers; i++ ) {
			fprintf(stderr,"%ld ", mcs->measurement_data[i]);
		}
		fprintf(stderr,"\n");
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_read_timer( MX_MCS *mcs )
{
	const char fname[] = "mxd_network_mcs_read_timer()";

	MX_NETWORK_MCS *network_mcs;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long dimension_array[1];
	mx_status_type mx_status;

	mx_status = mxd_network_mcs_get_pointers( mcs, &network_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.timer_data", network_mcs->remote_record_name );

	MX_DEBUG( 1,("%s: reading '%s'.", fname, rfname));

	dimension_array[0] = (long) mcs->current_num_measurements;

	mx_status = mx_get_array( network_mcs->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				mcs->timer_data );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 1,("%s: data received for '%s'.", fname, rfname));

#if 0
	{
		long i;

		for ( i = 0; i < mcs->current_num_measurements; i++ ) {
			fprintf(stderr,"%g ", mcs->timer_data[i]);
		}
		fprintf(stderr,"\n");
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_get_parameter( MX_MCS *mcs )
{
	const char fname[] = "mxd_network_mcs_get_parameter()";

	MX_NETWORK_MCS *network_mcs;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long mode;
	unsigned long num_measurements, measurement_counts;
	double measurement_time;
	mx_status_type mx_status;

	mx_status = mxd_network_mcs_get_pointers( mcs, &network_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 1,("%s invoked for MCS '%s', type = %d",
		fname, mcs->record->name, mcs->parameter_type));

	if ( mcs->parameter_type == MXLV_MCS_MODE ) {

		sprintf( rfname, "%s.mode",
					network_mcs->remote_record_name );

		mx_status = mx_get( network_mcs->server_record,
				rfname, MXFT_LONG, &mode );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mcs->mode = mode;

	} else if ( mcs->parameter_type == MXLV_MCS_MEASUREMENT_TIME ) {

		sprintf( rfname, "%s.measurement_time",
					network_mcs->remote_record_name );

		mx_status = mx_get( network_mcs->server_record,
				rfname, MXFT_DOUBLE, &measurement_time );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mcs->measurement_time = measurement_time;

	} else if ( mcs->parameter_type == MXLV_MCS_MEASUREMENT_COUNTS ) {

		sprintf( rfname, "%s.measurement_counts",
					network_mcs->remote_record_name );

		mx_status = mx_get( network_mcs->server_record,
				rfname, MXFT_ULONG, &measurement_counts );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mcs->measurement_counts = measurement_counts;

	} else if ( mcs->parameter_type == MXLV_MCS_CURRENT_NUM_MEASUREMENTS ){

		sprintf( rfname, "%s.current_num_measurements",
					network_mcs->remote_record_name );

		mx_status = mx_get( network_mcs->server_record,
				rfname, MXFT_ULONG, &num_measurements );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mcs->current_num_measurements = num_measurements;

	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			mcs->parameter_type );
	}
	MX_DEBUG( 1,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mcs_set_parameter( MX_MCS *mcs )
{
	const char fname[] = "mxd_network_mcs_set_parameter()";

	MX_NETWORK_MCS *network_mcs;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long mode;
	unsigned long num_measurements, measurement_counts;
	double measurement_time;
	mx_status_type mx_status;

	mx_status = mxd_network_mcs_get_pointers( mcs, &network_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 1,("%s invoked for MCS '%s', type = %d",
		fname, mcs->record->name, mcs->parameter_type));

	if ( mcs->parameter_type == MXLV_MCS_MODE ) {

		sprintf( rfname, "%s.mode",
					network_mcs->remote_record_name );

		mode = mcs->mode;

		mx_status = mx_put( network_mcs->server_record,
				rfname, MXFT_LONG, &mode );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

	} else if ( mcs->parameter_type == MXLV_MCS_MEASUREMENT_TIME ) {

		sprintf( rfname, "%s.measurement_time",
					network_mcs->remote_record_name );

		measurement_time = mcs->measurement_time;

		MX_DEBUG( 1,("%s: sending %g to '%s'",
					fname, measurement_time, rfname));

		mx_status = mx_put( network_mcs->server_record,
				rfname, MXFT_DOUBLE, &measurement_time );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

	} else if ( mcs->parameter_type == MXLV_MCS_MEASUREMENT_COUNTS ) {

		sprintf( rfname, "%s.measurement_counts",
					network_mcs->remote_record_name );

		measurement_counts = mcs->measurement_counts;

		MX_DEBUG( 1,("%s: sending %ld to '%s'",
					fname, measurement_counts, rfname));

		mx_status = mx_put( network_mcs->server_record,
				rfname, MXFT_ULONG, &measurement_counts );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

	} else if ( mcs->parameter_type == MXLV_MCS_CURRENT_NUM_MEASUREMENTS ){

		sprintf( rfname, "%s.current_num_measurements",
					network_mcs->remote_record_name );

		num_measurements = mcs->current_num_measurements;

		MX_DEBUG( 1,("%s: sending %lu to '%s'",
					fname, num_measurements, rfname));

		mx_status = mx_put( network_mcs->server_record,
				rfname, MXFT_ULONG, &num_measurements );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			mcs->parameter_type );
	}
	MX_DEBUG( 1,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

