/*
 * Name:    i_pdi45.h
 *
 * Purpose: Header for MX driver for Prairie Digital, Inc. Model 45
 *          data acquisition and control module.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999, 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#ifndef __I_PDI45_H__
#define __I_PDI45_H__

#include "mx_record.h"
#include "mx_generic.h"

#define MXI_PDI45_STANDARD_FIELDS \
  {-1, -1, "rs232_record", MXFT_RECORD, NULL, 0, {0}, \
	MXF_REC_TYPE_STRUCT, offsetof(MX_PDI45, rs232_record), \
	{0}, NULL, (MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY)}

MX_API mx_status_type mxi_pdi45_initialize_type( long type );
MX_API mx_status_type mxi_pdi45_create_record_structures( MX_RECORD *record );
MX_API mx_status_type mxi_pdi45_finish_record_initialization(
							MX_RECORD *record );
MX_API mx_status_type mxi_pdi45_delete_record( MX_RECORD *record );
MX_API mx_status_type mxi_pdi45_read_parms_from_hardware( MX_RECORD *record );
MX_API mx_status_type mxi_pdi45_write_parms_to_hardware( MX_RECORD *record );
MX_API mx_status_type mxi_pdi45_open( MX_RECORD *record );
MX_API mx_status_type mxi_pdi45_close( MX_RECORD *record );

extern MX_RECORD_FUNCTION_LIST mxi_pdi45_record_function_list;
extern MX_GENERIC_FUNCTION_LIST mxi_pdi45_generic_function_list;

extern long mxi_pdi45_num_record_fields;
extern MX_RECORD_FIELD_DEFAULTS *mxi_pdi45_rfield_def_ptr;

/**************************************************************************/

/* Define the data structures used by a PDI Model 45 interface. */

#define PDI45_NUM_DIGITAL_CHANNELS	8

typedef struct {
	MX_RECORD *record;

	MX_RECORD *rs232_record;
	int version;
	int data_bits;
	int parity;
	int four_step_mode;
	int speed;
	int turn_around_delay;
	int communication_timer;
	int io_type[ PDI45_NUM_DIGITAL_CHANNELS ];
} MX_PDI45;

/* Error codes returned by the Model 45 (from page 15). */

#define PDI45_ERR_POWER_UP_CLEAR_EXPECTED		0x00
#define PDI45_ERR_UNDEFINED_COMMAND			0x01
#define PDI45_ERR_CHECKSUM_ERROR			0x02
#define PDI45_ERR_INPUT_BUFFER_OVERFLOW			0x03
#define PDI45_ERR_NON_PRINTABLE_CHARACTER_RECEIVED	0x04
#define PDI45_ERR_DATA_FIELD_ERROR			0x05
#define PDI45_ERR_COMMUNICATION_TIME_OUT_ERROR		0x06
#define PDI45_ERR_INVALID_SPECIFIED_VALUE		0x07
#define PDI45_ERR_MODULE_INTERFACE_ERROR		0x10

/* Communication configuration bit patterns (from page 18). */

#define PDI45_COM_7_DATA_BITS		0x1
#define PDI45_COM_PARITY_ENABLED	0x2
#define PDI45_COM_ODD_PARITY		0x4

#define PDI45_COM_300_BAUD		0x0
#define PDI45_COM_600_BAUD		0x1
#define PDI45_COM_1200_BAUD		0x2
#define PDI45_COM_2400_BAUD		0x3
#define PDI45_COM_4800_BAUD		0x4
#define PDI45_COM_9600_BAUD		0x5
#define PDI45_COM_19200_BAUD		0x6
#define PDI45_COM_38400_BAUD		0x7
#define PDI45_COM_57600_BAUD		0x8
#define PDI45_COM_115200_BAUD		0x9

/* Digital I/O type mask.  To see the details of the individual
 * types, see pages 21 and 22 of the manual.  The descriptions
 * of their distinctions are rather long to encapsulate in 
 * a #define statement.
 */

#define PDI45_OUTPUT_TYPE_MASK	0x80

/* Type definitions for pdi45_set_digital_io_type(). */

#define PDI45_INPUT_TYPE	0x00
#define PDI45_OUTPUT_TYPE	0x80

/* Miscellaneous definitions. */

#define PDI45_END_OF_STRING	'\r'

#define PDI45_COMMAND_LENGTH	50

/* PDI Model 45 interface functions. */

MX_API mx_status_type pdi45_open( MX_PDI45 *pdi45 );

MX_API mx_status_type pdi45_close( MX_PDI45 *pdi45 );

MX_API mx_status_type pdi45_read( MX_PDI45 *pdi45,
					char *buffer, int buffer_length );

MX_API mx_status_type pdi45_write( MX_PDI45 *pdi45, char *buffer );

MX_API mx_status_type pdi45_command( MX_PDI45 *pdi45,
					char *command, char *response );

MX_API mx_status_type pdi45_checksum( MX_PDI45 *pdi45,
					char *command, char *checksum );

MX_API mx_status_type pdi45_check_checksum( char *response );

MX_API mx_status_type pdi45_debug_mode( MX_PDI45 *pdi45, int mode );

MX_API mx_status_type pdi45_get_error_code( char *buffer, int *error_code );

/***** Commands allowed only in setup mode. *****/

/* --- Not all setup mode commands are supported yet. --- */

MX_API mx_status_type pdi45_set_digital_io_type( MX_PDI45 *pdi45,
					int dline, int io_type, int dtype );

/***** Commands allowed in both setup mode and normal mode. *****/

MX_API mx_status_type pdi45_read_output_timer( MX_PDI45 *pdi45,
				int dline, int *timer_on, int *hex_value );

MX_API mx_status_type pdi45_set_output_timer( MX_PDI45 *pdi45,
				int dline, int hex_value );

/***** Commands allowed only in normal mode. *****/

MX_API mx_status_type pdi45_read_communication_config( MX_PDI45 *pdi45 );

MX_API mx_status_type pdi45_power_up_clear( MX_PDI45 *pdi45 );

MX_API mx_status_type pdi45_reset( MX_PDI45 *pdi45 );

MX_API mx_status_type pdi45_read_digital_io_types( MX_PDI45 *pdi45 );

MX_API mx_status_type pdi45_read_digital_power_up_levels( MX_PDI45 *pdi45,
			int *io_directions, int *power_up_levels );

MX_API mx_status_type pdi45_set_digital_on( MX_PDI45 *pdi45, int dline );

MX_API mx_status_type pdi45_set_digital_off( MX_PDI45 *pdi45, int dline );

MX_API mx_status_type pdi45_read_all_digital_io_levels( MX_PDI45 *pdi45,
					int *value );

MX_API mx_status_type pdi45_read_digital_io_level( MX_PDI45 *pdi45,
					int dline, int *value );

MX_API mx_status_type pdi45_read_all_latches( MX_PDI45 *pdi45,
					int *value );

MX_API mx_status_type pdi45_read_latch( MX_PDI45 *pdi45,
					int dline, int *value );

MX_API mx_status_type pdi45_read_and_clear_all_latches( MX_PDI45 *pdi45,
					int *value );

MX_API mx_status_type pdi45_read_and_clear_latch( MX_PDI45 *pdi45,
					int dline, int *value );

MX_API mx_status_type pdi45_clear_all_latches( MX_PDI45 *pdi45 );

MX_API mx_status_type pdi45_clear_latch( MX_PDI45 *pdi45, int dline );

MX_API mx_status_type pdi45_enable_counter( MX_PDI45 *pdi45, int dline );

MX_API mx_status_type pdi45_disable_counter( MX_PDI45 *pdi45, int dline );

MX_API mx_status_type pdi45_read_counter( MX_PDI45 *pdi45,
					int dline, int *count );

MX_API mx_status_type pdi45_read_and_clear_counter( MX_PDI45 *pdi45,
					int dline, int *count );

MX_API mx_status_type pdi45_clear_counter( MX_PDI45 *pdi45, int dline );

MX_API mx_status_type pdi45_retrigger_time_delay( MX_PDI45 *pdi45, int dline );

MX_API mx_status_type pdi45_set_analog_output( MX_PDI45 *pdi45,
					int channel, double voltage );

MX_API mx_status_type pdi45_read_analog_output( MX_PDI45 *pdi45,
					int channel, double *voltage );

MX_API mx_status_type pdi45_read_analog_input( MX_PDI45 *pdi45,
					int channel, double *voltage );

MX_API mx_status_type pdi45_read_analog_output_power_up_levels(MX_PDI45 *pdi45,
			double *channel0_output, double *channel1_output);

#endif /* __I_PDI45_H__ */

