/*
 * Name:    i_pmac.c
 *
 * Purpose: MX driver for Delta Tau PMAC controllers.  Note that this is
 *          the driver for talking to the PMAC itself.  PMAC-controlled
 *          motors are interfaced via the driver in libmx/d_pmac.c.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999, 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"

#if HAVE_EPICS
#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"
#endif /* HAVE_EPICS */

#include "i_pmac.h"

MX_RECORD_FUNCTION_LIST mxi_pmac_record_function_list = {
	mxi_pmac_initialize_type,
	mxi_pmac_create_record_structures,
	mxi_pmac_finish_record_initialization,
	mxi_pmac_delete_record,
	NULL,
	mxi_pmac_read_parms_from_hardware,
	mxi_pmac_write_parms_to_hardware,
	mxi_pmac_open,
	mxi_pmac_close,
	NULL,
	mxi_pmac_resynchronize
};

MX_GENERIC_FUNCTION_LIST mxi_pmac_generic_function_list = {
	mxi_pmac_getchar,
	mxi_pmac_putchar,
	mxi_pmac_read,
	mxi_pmac_write,
	mxi_pmac_input_is_available,
	mxi_pmac_discard_unread_input,
	mxi_pmac_discard_unwritten_output
};

MX_RECORD_FIELD_DEFAULTS mxi_pmac_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_PMAC_STANDARD_FIELDS
};

long mxi_pmac_num_record_fields
		= sizeof( mxi_pmac_record_field_defaults )
			/ sizeof( mxi_pmac_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_pmac_rfield_def_ptr
			= &mxi_pmac_record_field_defaults[0];

#define MXI_PMAC_DEBUG			FALSE

#define MXI_PMAC_LOW_LEVEL_DEBUG	FALSE

/* A private function for the use of the driver. */

static mx_status_type
mxi_pmac_get_pointers( MX_GENERIC *generic,
			MX_PMAC **pmac,
			const char *calling_fname )
{
	const char fname[] = "mxi_pmac_get_pointers()";

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_GENERIC pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( pmac == (MX_PMAC **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_PMAC pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*pmac = (MX_PMAC *) generic->record->record_type_struct;

	if ( *pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_PMAC pointer for record '%s' is NULL.",
			generic->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*==========================*/

MX_EXPORT mx_status_type
mxi_pmac_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_pmac_create_record_structures()";

	MX_GENERIC *generic;
	MX_PMAC *pmac;

	/* Allocate memory for the necessary structures. */

	generic = (MX_GENERIC *) malloc( sizeof(MX_GENERIC) );

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_GENERIC structure." );
	}

	pmac = (MX_PMAC *) malloc( sizeof(MX_PMAC) );

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_PMAC structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = generic;
	record->record_type_struct = pmac;
	record->class_specific_function_list
				= &mxi_pmac_generic_function_list;

	generic->record = record;
	pmac->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxi_pmac_finish_record_initialization()";

	MX_PMAC *pmac;
	char *port_type_name;
	int i, length;

	pmac = (MX_PMAC *) record->record_type_struct;

	/* Identify the port_type of the record. */

	port_type_name = pmac->port_type_name;

	length = strlen( port_type_name );

	for ( i = 0; i < length; i++ ) {
		if ( isupper( (int) (port_type_name[i]) ) ) {
			port_type_name[i] = tolower( port_type_name[i] );
		}
	}

	if ( strcmp( port_type_name, "rs232" ) == 0 ) {
		pmac->port_type = MX_PMAC_PORT_TYPE_RS232;

	} else if ( strcmp( port_type_name, "epics_ect" ) == 0 ) {
		pmac->port_type = MX_PMAC_PORT_TYPE_EPICS_TC;

	} else {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The PMAC interface port_type '%s' is not a valid port_type.",
			port_type_name );
	}

	/* Perform port_type specific processing. */

	switch( pmac->port_type ) {
	case MX_PMAC_PORT_TYPE_RS232:
		pmac->port_record = mx_get_record( record->list_head,
							pmac->port_args );

		if ( pmac->port_record == NULL ) {
			return mx_error( MXE_NOT_FOUND, fname,
			"The RS-232 interface record '%s' does not exist.",
				pmac->port_args );
		}
		if ( pmac->port_record->mx_superclass != MXR_INTERFACE ) {
			return mx_error( MXE_TYPE_MISMATCH, fname,
				"'%s' is not an interface record.",
				pmac->port_record->name );
		}
		if ( pmac->port_record->mx_class != MXI_RS232 ) {
			return mx_error( MXE_TYPE_MISMATCH, fname,
				"'%s' is not an RS-232 interface.",
				pmac->port_record->name );
		}
		break;

	case MX_PMAC_PORT_TYPE_EPICS_TC:
		pmac->port_record = NULL;
		break;
	}

	/* Mark the PMAC device as being closed. */

	record->flags |= ( ~MXF_REC_OPEN );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_open( MX_RECORD *record )
{
#if HAVE_EPICS
	const char fname[] = "mxi_pmac_open()";

	MX_PMAC *pmac;

	MX_DEBUG( 2, ("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	pmac = (MX_PMAC *) (record->record_type_struct);

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for record '%s' is NULL.", record->name);
	}

	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	if ( pmac->port_type == MX_PMAC_PORT_TYPE_EPICS_TC ) {
		ezcaAutoErrorMessageOff();
	}
#endif /* HAVE_EPICS */

	/* Synchronize with the controller. */

	return mxi_pmac_resynchronize( record );
}

MX_EXPORT mx_status_type
mxi_pmac_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxi_pmac_resynchronize()";

	MX_GENERIC *generic;
	MX_PMAC *pmac;
#if 0
	MX_RS232 *rs232;
	int hw_flow_control;
	char command[20];
#endif
	char response[80];
	char pmac_type_name[20];
	int num_items;
	mx_status_type mx_status;

	MX_DEBUG( 2, ("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	generic = (MX_GENERIC *) record->record_class_struct;

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_GENERIC pointer for record '%s' is NULL.", record->name);
	}

	pmac = (MX_PMAC *) record->record_type_struct;

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for record '%s' is NULL.", record->name);
	}

#if 0  /* Attempting to program the I-variables here may be inappropriate. */

	/* Attempt to program the necessary PMAC I-variables in the blind.
	 *
	 * We must skip this step if we are talking to Tom Coleman's EPICS
	 * driver since that has its own expectations about I-variable values.
	 */

	if ( pmac->port_type != MX_PMAC_PORT_TYPE_EPICS_TC ) {

		/* First set I3=2 so that responses are terminated by
		 * either <ACK> or <BELL>.  Everything else depends
		 * on this being true.
		 */

		mx_status = mxi_pmac_send_command( pmac, "I3=2", MXI_PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mx_status = mxi_pmac_discard_unread_input( generic,
							MXI_PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Turn off checksums. */

		mx_status = mxi_pmac_send_command( pmac, "I4=0", MXI_PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mx_status = mxi_pmac_discard_unread_input( generic,
							MXI_PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Are we supposed to be using hardware flow control? */

		hw_flow_control = FALSE;

		if ( pmac->port_type == MX_PMAC_PORT_TYPE_RS232 ) {

			if ( pmac->port_record == (MX_RECORD *) NULL ) {
				return mx_error( MXE_CORRUPT_DATA_STRUCTURE,
					fname,
			"pmac->port_record pointer for PMAC '%s' is NULL.",
					record->name );
			}
			if ( pmac->port_record->mx_class != MXI_RS232 ) {
				return mx_error( MXE_TYPE_MISMATCH, fname,
		"The PMAC interface record '%s' is not an RS-232 record, "
		"although PMAC record '%s' says that it is supposed to be.",
					pmac->port_record->name,
					record->name );
			}

			rs232 = (MX_RS232 *)
				pmac->port_record->record_class_struct;

			if ( rs232 == (MX_RS232 *) NULL ) {
				return mx_error( MXE_CORRUPT_DATA_STRUCTURE,
					fname,
		"MX_RS232 pointer for PMAC interface record '%s' is NULL.",
					pmac->port_record->name );
			}

			switch( rs232->flow_control ) {
			case MXF_232_HARDWARE_FLOW_CONTROL:
				hw_flow_control = TRUE;
				break;
			case MXF_232_NO_FLOW_CONTROL:
				hw_flow_control = FALSE;
				break;
			default:
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"PMAC RS-232 interface '%s' must use only hardware flow control "
	"or no flow control.  Flow control type '%c' is not supported.",
					pmac->port_record->name,
					rs232->flow_control );
				break;
			}
		}

		/* Reprogram I1 for the number of cards and flow control
		 * type that we are supposed to have.
		 */

		if ( pmac->num_cards <= 1 ) {
			if ( hw_flow_control ) {
				strcpy( command, "I1=0" );
			} else {
				strcpy( command, "I1=1" );
			}
		} else {
			if ( hw_flow_control ) {
				strcpy( command, "I1=2" );
			} else {
				strcpy( command, "I1=3" );
			}
		}

		mx_status = mxi_pmac_send_command( pmac, command, MXI_PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mx_status = mxi_pmac_discard_unread_input( generic,
							MXI_PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		if ( pmac->num_cards > 1 ) {

			/* Select card 0 as the addressed card. */

			mx_status = mxi_pmac_send_command( pmac, "@0",
							MXI_PMAC_DEBUG );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			mx_status = mxi_pmac_discard_unread_input( generic,
							MXI_PMAC_DEBUG );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;
		}

		/* Set the PMACs to abbreviated decimal program listing form. */

		if ( pmac->num_cards > 1 ) {
			mx_status = mxi_pmac_send_command( pmac, "@@I9=0",
							MXI_PMAC_DEBUG );
		} else {
			mx_status = mxi_pmac_send_command( pmac, "I9=0",
							MXI_PMAC_DEBUG );
		}

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mx_status = mxi_pmac_discard_unread_input( generic,
							MXI_PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

#endif /* End of commented out initialization of I-variables. */

	/* Now that most of the necessary PMAC I-variables have been
	 * reprogrammed (hopefully), try to verify that the PMAC is
	 * actually there.
	 */

	/* Verify that the PMAC controller is active by asking it for
	 * its card type.
	 */

	mx_status = mxi_pmac_command( pmac, "TYPE",
				response, sizeof response, MXI_PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	num_items = sscanf( response, "%s,", pmac_type_name );

	if ( num_items != 1 ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
			"The device attached to this interface does not "
			"seem to be a PMAC controller."
			"Response to PMAC TYPE command = '%s'",
				response );
	}

	if ( strncmp( pmac_type_name, "PMAC1",5 ) == 0 ) {
		pmac->pmac_type = MX_PMAC_TYPE_PMAC1;

	} else
	if ( strncmp( pmac_type_name, "PMAC2",5 ) == 0 ) {
		pmac->pmac_type = MX_PMAC_TYPE_PMAC2;

	} else
	if ( strncmp( pmac_type_name, "PMACUL",6 ) == 0 ) {
		pmac->pmac_type = MX_PMAC_TYPE_PMACUL;

	} else
	if ( strncmp( pmac_type_name, "TURBO1",6 ) == 0 ) {
		pmac->pmac_type = MX_PMAC_TYPE_TURBO1;

	} else
	if ( strncmp( pmac_type_name, "TURBO2",6 ) == 0 ) {
		pmac->pmac_type = MX_PMAC_TYPE_TURBO2;

	} else
	if ( strncmp( pmac_type_name, "TURBOU",6 ) == 0 ) {
		pmac->pmac_type = MX_PMAC_TYPE_TURBOU;

	} else {
		if ( ( strncmp( pmac_type_name, "PMAC", 4 ) != 0 )
		  && ( strncmp( pmac_type_name, "TURBO", 5 ) != 0 ) )
		{
			return mx_error( MXE_INTERFACE_IO_ERROR, fname,
				"The device attached to this interface does "
				"not seem to be a PMAC controller."
				"Response to PMAC TYPE command = '%s'",
					response );
		} else {
			return mx_error( MXE_TYPE_MISMATCH, fname,
			"Unrecognized PMAC type '%s' for PMAC '%s'.",
				pmac_type_name, record->name );
		}
	}

	/* Ask for the style of error message reporting. */

	mx_status = mxi_pmac_command( pmac, "I6",
				response, sizeof response, MXI_PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	num_items = sscanf( response, "%d", &(pmac->i6_variable) );

	if ( num_items != 1 ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"Unable to parse response to I6 command.  Response = '%s'",
			response );
	}

	/* Now ask for PROM firmware version. */

	mx_status = mxi_pmac_command( pmac, "VERSION",
				response, sizeof response, MXI_PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	num_items = sscanf( response, "%d.%d",
			&(pmac->major_version), &(pmac->minor_version) );

	if ( num_items != 2 ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
			"Unable to parse response to VERSION command.  "
			"Response = '%s'", response );
	}

#if MXI_PMAC_DEBUG
	MX_DEBUG(-2, ("%s: PMAC version: major = %d, minor = %d",
		fname, pmac->major_version, pmac->minor_version));

	MX_DEBUG(-2,("%s: I6 = %d", fname, pmac->i6_variable));
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pmac_getchar( MX_GENERIC *generic, char *c, int flags )
{
	const char fname[] = "mxi_pmac_getchar()";

	MX_PMAC *pmac;
	mx_status_type mx_status;

	mx_status = mxi_pmac_get_pointers( generic, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( pmac->port_type ) {
	case MX_PMAC_PORT_TYPE_RS232:
		mx_status = mx_rs232_getchar( pmac->port_record, c, flags );
		break;

	case MX_PMAC_PORT_TYPE_EPICS_TC:
		*c = '\0';

		return mx_error( MXE_UNSUPPORTED, fname,
			"Single character PMAC I/O is not supported "
			"for Tom Coleman's EPICS interface." );
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface port_type = %d.",
			pmac->port_type );
		break;
	}
	return mx_status;
}

MX_EXPORT mx_status_type
mxi_pmac_putchar( MX_GENERIC *generic, char c, int flags )
{
	const char fname[] = "mxi_pmac_putchar()";

	MX_PMAC *pmac;
	mx_status_type mx_status;

	mx_status = mxi_pmac_get_pointers( generic, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( pmac->port_type ) {
	case MX_PMAC_PORT_TYPE_RS232:
		mx_status = mx_rs232_putchar( pmac->port_record, c, flags );
		break;

	case MX_PMAC_PORT_TYPE_EPICS_TC:
		return mx_error( MXE_UNSUPPORTED, fname,
			"Single character PMAC I/O is not supported "
			"for Tom Coleman's EPICS interface." );
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface port_type = %d.",
			pmac->port_type );
		break;
	}
	return mx_status;
}

MX_EXPORT mx_status_type
mxi_pmac_read( MX_GENERIC *generic, void *buffer, size_t count )
{
	const char fname[] = "mxi_pmac_read()";

	MX_PMAC *pmac;
	mx_status_type mx_status;

	mx_status = mxi_pmac_get_pointers( generic, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( pmac->port_type ) {
	case MX_PMAC_PORT_TYPE_RS232:
		mx_status = mx_rs232_read( pmac->port_record,
						buffer, count, NULL, 0 );
		break;

	case MX_PMAC_PORT_TYPE_EPICS_TC:
		return mx_error( MXE_UNSUPPORTED, fname,
			"The driver for Tom Coleman's PMAC EPICS interface "
			"does not support binary I/O." );
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface port_type = %d.",
			pmac->port_type );
		break;
	}
	return mx_status;
}

MX_EXPORT mx_status_type
mxi_pmac_write( MX_GENERIC *generic, void *buffer, size_t count )
{
	const char fname[] = "mxi_pmac_write()";

	MX_PMAC *pmac;
	mx_status_type mx_status;

	mx_status = mxi_pmac_get_pointers( generic, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( pmac->port_type ) {
	case MX_PMAC_PORT_TYPE_RS232:
		mx_status = mx_rs232_write( pmac->port_record,
						buffer, count, NULL, 0 );
		break;

	case MX_PMAC_PORT_TYPE_EPICS_TC:
		return mx_error( MXE_UNSUPPORTED, fname,
			"The driver for Tom Coleman's PMAC EPICS interface "
			"does not support binary I/O." );
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface port_type = %d.",
			pmac->port_type );
		break;
	}
	return mx_status;
}

MX_EXPORT mx_status_type
mxi_pmac_input_is_available( MX_GENERIC *generic, int *input_available )
{
	const char fname[] = "mxi_pmac_input_is_available()";

	MX_PMAC *pmac;
	MX_RECORD *rs232_record;
	mx_status_type mx_status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	pmac = (MX_PMAC *)(generic->record->record_type_struct);

	if ( pmac == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for MX_GENERIC structure is NULL." );
	}

	switch( pmac->port_type ) {
	case MX_PMAC_PORT_TYPE_RS232:
		rs232_record = pmac->port_record;

		if ( rs232_record == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_RECORD pointer for MX_PMAC structure is NULL." );
		}
		mx_status = mx_rs232_input_is_available(rs232_record,
							input_available);
		break;

	case MX_PMAC_PORT_TYPE_EPICS_TC:
		/* WML: I haven't decided how to implement this yet. */

		*input_available = TRUE;
		mx_status = MX_SUCCESSFUL_RESULT;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface port_type = %d.",
			pmac->port_type );
		break;
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mxi_pmac_discard_unread_input( MX_GENERIC *generic, int debug_flag )
{
	const char fname[] = "mxi_pmac_discard_unread_input()";

	MX_PMAC *pmac;
	mx_status_type mx_status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	pmac = (MX_PMAC *)(generic->record->record_type_struct);

	if ( pmac == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for MX_GENERIC structure is NULL." );
	}

	switch( pmac->port_type ) {
	case MX_PMAC_PORT_TYPE_RS232:
		mx_status = mx_rs232_discard_unread_input( pmac->port_record,
								debug_flag );
		break;

	case MX_PMAC_PORT_TYPE_EPICS_TC:
		/* WML: I haven't decided how to implement this yet. */

		mx_status = MX_SUCCESSFUL_RESULT;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface port_type = %d.",
			pmac->port_type );
		break;
	}
	return mx_status;
}

MX_EXPORT mx_status_type
mxi_pmac_discard_unwritten_output( MX_GENERIC *generic, int debug_flag )
{
	const char fname[] = "mxi_pmac_discard_unwritten_output()";

	MX_PMAC *pmac;
	mx_status_type mx_status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	pmac = (MX_PMAC *)(generic->record->record_type_struct);

	if ( pmac == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC pointer for MX_GENERIC structure is NULL." );
	}

	switch( pmac->port_type ) {
	case MX_PMAC_PORT_TYPE_RS232:
		mx_status = mx_rs232_discard_unwritten_output(
				pmac->port_record, debug_flag );
		break;

	case MX_PMAC_PORT_TYPE_EPICS_TC:
		/* As far as I know, this is not possible for EPICS,
		 * so we just say that it succeeded.
		 */

		mx_status = MX_SUCCESSFUL_RESULT;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface port_type = %d.",
			pmac->port_type );
		break;
	}
	return mx_status;
}

/* === Functions specific to this driver. === */

static mx_status_type
mxi_pmac_rs232_send_command( MX_PMAC *, char *, int );

static mx_status_type
mxi_pmac_rs232_receive_response( MX_PMAC *, char *, int, int );
				

#if HAVE_EPICS
static mx_status_type
mxi_pmac_epics_tc_command( MX_PMAC *, char *, char *, int, int );
#endif

static struct {
	int mx_error_code;
	char error_message[100];
} mxi_pmac_error_messages[] = {
	{ MXE_INTERFACE_ACTION_FAILED,				/* 0 */
		"Unknown error code 0" },
	{ MXE_CLIENT_REQUEST_DENIED,				/* 1 */
		"Command not allowed during program execution" },
	{ MXE_PERMISSION_DENIED,				/* 2 */
		"Password error" },
	{ MXE_INTERFACE_ACTION_FAILED,				/* 3 */
		"Data error or unrecognized command" },
	{ MXE_INTERFACE_IO_ERROR,				/* 4 */
		"Illegal character: bad value or serial parity/framing error" },
	{ MXE_CLIENT_REQUEST_DENIED,				/* 5 */
		"Command not allowed unless buffer is open" },
	{ MXE_CLIENT_REQUEST_DENIED,				/* 6 */
		"No room in buffer for command" },
	{ MXE_CLIENT_REQUEST_DENIED,				/* 7 */
		"Buffer already in use" },
	{ MXE_INTERFACE_IO_ERROR,				/* 8 */
		"MACRO auxiliary communications error" },
	{ MXE_ILLEGAL_ARGUMENT,					/* 9 */
		"Program structural error" },
	{ MXE_LIMIT_WAS_EXCEEDED,				/* 10 */
	"Both overtravel limits set for a motor in the coordinate system" },
	{ MXE_DEVICE_ACTION_FAILED,				/* 11 */
		"Previous move not completed" },
	{ MXE_DEVICE_ACTION_FAILED,				/* 12 */
		"A motor in the coordinate system is open-loop" },
	{ MXE_DEVICE_ACTION_FAILED,				/* 13 */
		"A motor in the coordinate system is not activated" },
	{ MXE_ILLEGAL_ARGUMENT,					/* 14 */
		"No motors in the coordinate system" },
	{ MXE_ILLEGAL_ARGUMENT,					/* 15 */
		"Not pointing to valid program buffer" },
	{ MXE_ILLEGAL_ARGUMENT,					/* 16 */
		"Running improperly structured program" },
	{ MXE_DEVICE_ACTION_FAILED,				/* 17 */
		"Trying to resume after H or Q with motors out of "
		"stopped position" },
	{ MXE_CLIENT_REQUEST_DENIED,				/* 18 */
		"Attempt to perform phase reference during move, "
		"or move during phase reference" },
	{ MXE_CLIENT_REQUEST_DENIED,				/* 19 */
		"Illegal position-change command while moves stored "
		"in CCBUFFER" },
};

static int mxi_pmac_num_error_messages =
	sizeof( mxi_pmac_error_messages )
	/ sizeof( mxi_pmac_error_messages[0] );

/*----*/

MX_EXPORT mx_status_type
mxi_pmac_command( MX_PMAC *pmac, char *command,
		char *response, int response_buffer_length,
		int debug_flag )
{
	const char fname[] = "mxi_pmac_command()";

	char alt_response_buffer[40];
	char *receive_response;
	int receive_response_length;
	int length, error_code, num_items;
	mx_status_type mx_status;

	/* WARNING: Please note that this routine _assumes_ that I3 == 2. */

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_PMAC pointer passed was NULL." );
	}

#if HAVE_EPICS
	/* Tom Coleman's EPICS driver is handled separately. */

	if ( pmac->port_type == MX_PMAC_PORT_TYPE_EPICS_TC ) {
		return mxi_pmac_epics_tc_command( pmac, command, 
					response, response_buffer_length,
					debug_flag );
	}
#endif

	/* Send the command string. */

	mx_status = mxi_pmac_send_command( pmac, command, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Get the response. */

	if ( response == NULL ) {
		receive_response = alt_response_buffer;
		receive_response_length = sizeof(alt_response_buffer);
	} else {
		receive_response = response;
		receive_response_length = response_buffer_length;
	}

	mx_status = mxi_pmac_receive_response( pmac,
					receive_response,
					receive_response_length,
					debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* If the command completed without an error, delete the <ACK>
	 * character at the end of the response and return the rest
	 * of the response to the caller.
	 */

	if ( receive_response[0] != MX_BELL ) {

		if ( response != NULL ) {
			length = strlen( response );

			if ( response[length - 1] == MX_ACK ) {
				response[length - 1] = '\0';
			}
		}
		return MX_SUCCESSFUL_RESULT;
	}

	/* Otherwise, we must handle the error. */

	error_code = -1;

	switch( pmac->i6_variable ) {
	case 0:
	case 2:
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
	"The PMAC interface '%s' reported an error for the command '%s'.  "
	"There is no information available about the reason.",
			pmac->record->name, command );
		break;
	case 3:
		receive_response++;	/* Skip over the <CR> character. */
	case 1:
		/* If present, delete the <CR> at the end of the
		 * error message.
		 */

		length = strlen( receive_response );

		if ( receive_response[ length - 1 ] == MX_CR ) {
			receive_response[ length - 1 ] = '\0';
		}

		num_items = sscanf( receive_response, "ERR%d", &error_code );

		if ( num_items != 1 ) {
			return mx_error( MXE_INTERFACE_IO_ERROR, fname,
	"The PMAC interface '%s' reported an error for the command '%s'.  "
	"An attempt to parse the error status failed.  "
	"PMAC error message = '%s'",
			pmac->record->name, command, receive_response );
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The PMAC interface '%s' reported an error for the command '%s', "
	"but the PMAC I6 variable has the unexpected value of %d.  "
	"This interferes with the parsing of error messages from the PMAC.  "
	"PMAC error message = '%s'",
			pmac->record->name, command,
			pmac->i6_variable, receive_response );
		break;
	}

	if ( error_code < 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The error code (%d) returned for PMAC command '%s' was less than 0.  "
	"This shouldn't be able to happen and is probably an MX driver bug."
	"PMAC error message = '%s'",
			error_code, command, receive_response );
	}

	if ( error_code >= mxi_pmac_num_error_messages ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
	"The PMAC interface '%s' reported an error for the command '%s'.  "
	"Error code = %d, error message = 'Unknown error code %d'",
			pmac->record->name, command,
			error_code, error_code );
	}

	return mx_error( mxi_pmac_error_messages[error_code].mx_error_code,
			fname,
	"The PMAC interface '%s' reported an error for the command '%s'.  "
	"Error code = %d, error message = '%s'",
			pmac->record->name, command, error_code, 
			mxi_pmac_error_messages[error_code].error_message );
}

MX_EXPORT mx_status_type
mxi_pmac_send_command( MX_PMAC *pmac,
			char *command,
			int debug_flag )
{
	const char fname[] = "mxi_pmac_send_command()";

	mx_status_type mx_status;

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_PMAC pointer passed was NULL." );
	}
	if ( command == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"command buffer pointer passed was NULL." );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: sent '%s' to '%s'",
			fname, command, pmac->record->name));
	}

	switch( pmac->port_type ) {
	case MX_PMAC_PORT_TYPE_RS232:
		mx_status = mxi_pmac_rs232_send_command( pmac,
						command, debug_flag );
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface port_type = %d.",
			pmac->port_type );
		break;
	}
	return mx_status;
}

MX_EXPORT mx_status_type
mxi_pmac_receive_response( MX_PMAC *pmac,
			char *response,
			int response_buffer_length,
			int debug_flag )
{
	const char fname[] = "mxi_pmac_receive_response()";

	size_t length;
	mx_status_type mx_status;

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_PMAC pointer passed was NULL." );
	}
	if ( response == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"response buffer pointer passed was NULL." );
	}

	switch( pmac->port_type ) {
	case MX_PMAC_PORT_TYPE_RS232:
		mx_status = mxi_pmac_rs232_receive_response( pmac,
					response, response_buffer_length,
					debug_flag );
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown PMAC interface port_type = %d.",
			pmac->port_type );
		break;
	}
	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Is there a trailing CR?  If so, then overwrite it with
	 * a null character.
	 */

	length = strlen( response );

	if ( length > 0 ) {
		if ( response[length-1] == MX_CR ) {
			response[length-1] = '\0';
		}
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: received '%s' from '%s'",
			fname, response, pmac->record->name));
	}
	return MX_SUCCESSFUL_RESULT;
}

/*----*/

static mx_status_type
mxi_pmac_rs232_send_command( MX_PMAC *pmac,
			char *command,
			int debug_flag )
{
#if MXI_PMAC_LOW_LEVEL_DEBUG
	const char fname[] = "mxi_pmac_rs232_send_command()";
#endif

	char c;
	char *ptr;
	mx_status_type mx_status;

	/* First, send a ctrl-Z to make the serial port the active interface. */

	c = MX_CTRL_Z;

	mx_status = mx_rs232_putchar( pmac->port_record, c, MXF_232_WAIT );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if MXI_PMAC_LOW_LEVEL_DEBUG
	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: sent 0x%x '%c'", fname, c, c));
	}
#endif

	/* Now send the main string. */

	ptr = &command[0];

	while ( *ptr != '\0' ) {
		c = *ptr;

		mx_status = mx_rs232_putchar( pmac->port_record, c, MXF_232_WAIT );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

#if MXI_PMAC_LOW_LEVEL_DEBUG
		if ( debug_flag ) {
			MX_DEBUG(-2, ("%s: sent 0x%x '%c'", fname, c, c));
		}
#endif
		ptr++;
	}

	/* Send a CR after the end of the command. */

	c = MX_CR;

	mx_status = mx_rs232_putchar( pmac->port_record, c, MXF_232_WAIT );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if MXI_PMAC_LOW_LEVEL_DEBUG
	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: (eos) sent 0x%x '%c'", fname, c, c));
	}
#endif
	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxi_pmac_rs232_receive_response( MX_PMAC *pmac,
			char *response,
			int response_buffer_length,
			int debug_flag )
{
	const char fname[] = "mxi_pmac_rs232_receive_response()";

	int i, bell_received;
	char c;
	mx_status_type mx_status;

	/* This function implicitly assumes I3 == 2.  This means that
	 * all PMAC responses will be terminated by either an <ACK>
	 * or a <BELL> character.  If I6 == 1, a <BELL> character
	 * will be followed by an error code terminated by a <CR>.
	 *
	 * Note that we must leave space at the end of the response
	 * buffer for the terminating null byte.
	 */

	bell_received = FALSE;

	for ( i = 0; i < (response_buffer_length - 1); i++ ) {

		response[i] = '\0';	/* Just in case we get an error. */

		mx_status = mx_rs232_getchar( pmac->port_record,
							&c, MXF_232_WAIT );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

#if MXI_PMAC_LOW_LEVEL_DEBUG
		if ( debug_flag ) {
			MX_DEBUG(-2,("%s: received 0x%x '%c'", fname, c, c));
		}
#endif

		response[i] = c;

		if ( c == MX_ACK ) {

			response[i] = '\0';
			break;			/* Exit the while() loop. */
		} else
		if ( c == MX_BELL ) {

			bell_received = TRUE;

			if ( pmac->i6_variable != 1 ) {

				response[i] = '\0';
				break;		/* Exit the while() loop. */
			}
		} else
		if ( c == MX_CR ) {
			if ( bell_received ) {

				response[i] = '\0';
				break;		/* Exit the while() loop. */
			}
		}
	}

	/* Was the response too long to fit in the response buffer? */

	if ( i >= (response_buffer_length - 1) ) {

		/* Null terminate the response buffer. */

		response[ response_buffer_length - 1 ] = '\0';

		/* Clear out any remaining unread characters. */

		mx_status = mx_rs232_discard_unread_input( pmac->port_record,
								debug_flag );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"The PMAC command response was longer than the buffer length of %d.",
			response_buffer_length );
	}

	return MX_SUCCESSFUL_RESULT;
}

#if HAVE_EPICS

static mx_status_type
mxi_pmac_epics_tc_command( MX_PMAC *pmac, char *command,
			char *response, int response_buffer_length,
			int debug_flag )
{
	const char fname[] = "mxi_pmac_epics_tc_command()";

	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;

	if ( pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_PMAC pointer passed was NULL." );
	}
	if ( command == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"command buffer pointer passed was NULL." );
	}

	/* Send the command. */

	sprintf( pvname, "%sStrCmd", pmac->port_args );

	ezca_status = ezcaPut( pvname, ezcaString, 1, command );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error sending string '%s' to EPICS variable '%s'.  EZCA status = %d (%s)",
			command, pvname, ezca_status, ezca_error_message );

		return mx_error( MXE_DEVICE_IO_ERROR, fname,
			mx_error_message );
	}

	/* Get the response, if any. */

	if ( response != NULL ) {
		sprintf( pvname, "%sStrRsp", pmac->port_args );

		ezca_status = ezcaGet( pvname, ezcaString, 1, response );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
"Error receiving string from EPICS variable '%s'.  EZCA status = %d (%s)",
			pvname, ezca_status, ezca_error_message );

			return mx_error( MXE_DEVICE_IO_ERROR, fname,
				mx_error_message );
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

#endif
