/*
 * Name:    i_xia_xerxes.c
 *
 * Purpose: MX driver for the X-Ray Instrumentation Associates 
 *          XerXes library used by the DXP-2X multichannel analyzer.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"

#if HAVE_XIA_XERXES

#include <stdlib.h>
#include <errno.h>

#include <xerxes_structures.h>
#include <xerxes_errors.h>
#include <xerxes.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_types.h"
#include "mx_alloc.h"
#include "mx_mca.h"
#include "i_xia_xerxes.h"
#include "d_xia_dxp.h"

MX_RECORD_FUNCTION_LIST mxi_xia_xerxes_record_function_list = {
	mxi_xia_xerxes_initialize_type,
	mxi_xia_xerxes_create_record_structures,
	mxi_xia_xerxes_finish_record_initialization,
	mxi_xia_xerxes_delete_record,
	NULL,
	mxi_xia_xerxes_read_parms_from_hardware,
	mxi_xia_xerxes_write_parms_to_hardware,
	mxi_xia_xerxes_open,
	mxi_xia_xerxes_close,
	NULL,
	mxi_xia_xerxes_resynchronize
};

MX_RECORD_FIELD_DEFAULTS mxi_xia_xerxes_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_XIA_XERXES_STANDARD_FIELDS
};

long mxi_xia_xerxes_num_record_fields
		= sizeof( mxi_xia_xerxes_record_field_defaults )
			/ sizeof( mxi_xia_xerxes_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_xia_xerxes_rfield_def_ptr
			= &mxi_xia_xerxes_record_field_defaults[0];

#define MXI_XIA_XERXES_DEBUG			TRUE

static mx_status_type mxi_xia_xerxes_read_dspsymbol_array(
					MX_XIA_DXP_MCA *xia_dxp_mca,
					char *dspsymbol_namestem,
					int num_array_elements,
					unsigned short *dspsymbol_array );

extern const char *mxi_xia_xerxes_strerror( int xia_status );

static mx_status_type
mxi_xia_xerxes_get_pointers( MX_MCA *mca,
			MX_XIA_DXP_MCA **xia_dxp_mca,
			MX_XIA_XERXES **xia_xerxes,
			const char *calling_fname )
{
	const char fname[] = "mxi_xia_xerxes_get_pointers()";

	MX_RECORD *xia_dxp_record;

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( xia_dxp_mca == (MX_XIA_DXP_MCA **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_XIA_DXP_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( xia_xerxes == (MX_XIA_XERXES **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_XIA_XERXES pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( mca->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_RECORD pointer for the MX_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*xia_dxp_mca = (MX_XIA_DXP_MCA *) mca->record->record_type_struct;

	xia_dxp_record = (*xia_dxp_mca)->xia_dxp_record;

	if ( xia_dxp_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The xia_dxp_record pointer for MCA record '%s' is NULL.",
			mca->record->name );
	}

	*xia_xerxes = (MX_XIA_XERXES *) xia_dxp_record->record_type_struct;

	if ( (*xia_xerxes) == (MX_XIA_XERXES *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_XERXES pointer for XIA DXP record '%s' is NULL.",
			xia_dxp_record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*==========================*/

MX_EXPORT mx_status_type
mxi_xia_xerxes_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_xerxes_create_record_structures()";

	MX_XIA_XERXES *xia_xerxes;

	/* Allocate memory for the necessary structures. */

	xia_xerxes = (MX_XIA_XERXES *) malloc( sizeof(MX_XIA_XERXES) );

	if ( xia_xerxes == (MX_XIA_XERXES *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Can't allocate memory for MX_XIA_XERXES structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = NULL;;
	record->record_type_struct = xia_xerxes;

	record->class_specific_function_list = NULL;

	xia_xerxes->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_xerxes_finish_record_initialization()";

	MX_XIA_XERXES *xia_xerxes;

	xia_xerxes = (MX_XIA_XERXES *) record->record_type_struct;

	if ( xia_xerxes == (MX_XIA_XERXES *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_XERXES pointer for record '%s' is NULL.",
			record->name );
	}

	/* We cannot allocate the mca_record_array structures until
	 * mxi_xia_xerxes_open() is invoked, so for now we initialize
	 * these data structures to all zeros.
	 */

	xia_xerxes->num_mcas = 0;
	xia_xerxes->mca_record_array = NULL;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

/* mxi_xia_xerxes_read_config() generates a temporary xiasystems.cfg style
 * configuration file and then tells XerXes to load it.
 */

static mx_status_type
mxi_xia_xerxes_read_config( MX_XIA_XERXES *xia_xerxes )
{
	const char fname[] = "mxi_xia_xerxes_read_config()";

	FILE *file;
	char modules_filename[ MXU_FILENAME_LENGTH + 1 ];
	int xia_status, fileio_status, saved_errno;

	/* Open a temporary xiasystems.cfg style file. */

	file = mx_open_temporary_file( modules_filename, MXU_FILENAME_LENGTH );

	if ( file == NULL ) {
		saved_errno = errno;

		return mx_error( MXE_FILE_IO_ERROR, fname,
		"Unable to open temporary XerXes configuration file '%s'.  "
		"Error code = %d, error status = '%s'.",
			modules_filename, saved_errno, strerror(saved_errno) );
	}

	MX_DEBUG(-2,("%s: Temporary XerXes module config file = '%s'.",
		fname, modules_filename));

	/* Write out the contents of the file. */

#if 0
	fprintf( file, "preamp    %s\n", xia_xerxes->preamp_filename );
	fprintf( file, "modules   %s\n", xia_xerxes->config_filename );
	fprintf( file, "%-9s NULL\n", xia_xerxes->module_type );
#else
	fprintf( file, "preamp    NULL\n" );
	fprintf( file, "modules   NULL\n" );
	fprintf( file, "%-9s NULL\n", xia_xerxes->module_type );
#endif

	fileio_status = fclose( file );

	if ( fileio_status != 0 ) {
		saved_errno = errno;

		return mx_error( MXE_FILE_IO_ERROR, fname,
		"An error occurred while closing temporary "
		"XerXes configuration file '%s'.  "
		"Error code = %d, error status = '%s'.",
			modules_filename, saved_errno, strerror(saved_errno) );
	}

	/* Tell XerXes to read the temporary config file. */

	xia_status = dxp_read_config( modules_filename );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to read the modules configuration from "
		"the XerXes configuration file '%s'.  "
		"Error code = %d, '%s'", modules_filename,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* Delete the temporary config file. */

#if 0
	fileio_status = remove( modules_filename );

	if ( fileio_status != 0 ) {
		saved_errno = errno;

		return mx_error( MXE_FILE_IO_ERROR, fname, 
		"The attempt to delete the temporary modules configuration "
		"file '%s' failed.  Error code = %d, error status = '%s'.",
			modules_filename, saved_errno, strerror(saved_errno) );
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

/*
 * mxi_xia_xerxes_restore_config() encapsulates all of the operations that
 * must happen when dxp_restore_config() is called.
 */

static mx_status_type
mxi_xia_xerxes_restore_config( MX_XIA_XERXES *xia_xerxes )
{
	const char fname[] = "mxi_xia_xerxes_restore_config()";

	int xia_status;
	int config_file_lun, config_file_mode;
	unsigned short gate, resume_flag;

	/* Open the detector configuration file. */

	MX_DEBUG(-2,("%s: Loading XerXes configuration '%s'.",
		fname, xia_xerxes->config_filename));

	config_file_mode = 2;

	xia_status = dxp_open_file( &config_file_lun,
					xia_xerxes->config_filename,
					&config_file_mode );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_FILE_IO_ERROR, fname,
		"Unable to open the XerXes detector configuration file '%s'.  "
		"Error code = %d, '%s'",
			xia_xerxes->config_filename,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* Restore the detector configuration. */

	xia_status = dxp_restore_config( &config_file_lun );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to restore the detector configuration from "
		"the XerXes configuration file '%s'.  "
		"Error code = %d, '%s'",
			xia_xerxes->config_filename,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* Close the detector configuration file. */

	xia_status = dxp_close_file( &config_file_lun );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_FILE_IO_ERROR, fname,
		"Unable to close the XerXes detector configuration file '%s'.  "
		"Error code = %d, '%s'",
			xia_xerxes->config_filename,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/*
	 * Patrick Franz of XIA recommends that a short DXP run be executed
	 * after a restore configuration is done.
	 */

	gate = 1;
	resume_flag = 0;

	xia_status = dxp_start_run( &gate, &resume_flag );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Unable to start a run for XIA DXP interface '%s'.  "
		"Error code = %d, '%s'", xia_xerxes->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* Wait for 1 second. */

	mx_sleep(1);

	/* Send a stop run command to make sure that all DXP channels
	 * are stopped.
	 */

	xia_status = dxp_stop_run();

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_FILE_IO_ERROR, fname,
		"Unable to stop runs on all DXP channels.  "
		"Error code = %d, '%s'",
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_open( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_xerxes_open()";

	MX_XIA_XERXES *xia_xerxes;
	int xia_status;
	int i, num_mcas;
	Xia_Util_Functions md_util_struct;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	xia_xerxes = (MX_XIA_XERXES *) (record->record_type_struct);

	if ( xia_xerxes == (MX_XIA_XERXES *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_XIA_XERXES pointer for record '%s' is NULL.", record->name);
	}

	/* Get pointers to the XerXes MD utility functions we will use. */

	xia_status = dxp_md_init_util( &md_util_struct, "Unused");

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to get pointers to the XerXes MD utility functions.  "
		"Error code = %d, '%s'",
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	xia_xerxes->dxp_md_error = md_util_struct.dxp_md_error;
	xia_xerxes->dxp_md_alloc = md_util_struct.dxp_md_alloc;
	xia_xerxes->dxp_md_free  = md_util_struct.dxp_md_free;
	xia_xerxes->dxp_md_puts  = md_util_struct.dxp_md_puts;
	xia_xerxes->dxp_md_wait  = md_util_struct.dxp_md_wait;

	/* Initialize XerXes. */

	xia_status = dxp_init_library();

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to initialize XerXes.  "
		"Error code = %d, '%s'",
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* Read the modules configuration. */

	mx_status = mxi_xia_xerxes_read_config( xia_xerxes );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Restore the detector configuration. */

	mx_status = mxi_xia_xerxes_restore_config( xia_xerxes );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Find out how many detector channels (MCAs) there are. */

	xia_status = dxp_ndxpchan( &num_mcas );

	MX_DEBUG(-2,("%s: num_mcas = %d", fname, num_mcas));

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to find out how many detector channels are in "
		"the current configuration '%s'.  "
		"Error code = %d, '%s'",
			xia_xerxes->config_filename,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* Allocate and initialize an array to store pointers to all of
	 * the MCA records used by this interface.
	 */

	xia_xerxes->mca_record_array = ( MX_RECORD ** ) malloc(
					num_mcas * sizeof( MX_RECORD * ) );

	if ( xia_xerxes->mca_record_array == ( MX_RECORD ** ) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %d element array "
		"of MX_RECORD pointers for the mca_record_array data "
		"structure of record '%s'.", num_mcas, record->name );
	}

	for ( i = 0; i < num_mcas; i++ ) {
		xia_xerxes->mca_record_array[i] = NULL;
	}

	xia_xerxes->num_mcas = num_mcas;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_xerxes_resynchronize()";

	MX_XIA_XERXES *xia_xerxes;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	xia_xerxes = (MX_XIA_XERXES *) (record->record_type_struct);

	if ( xia_xerxes == (MX_XIA_XERXES *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_XIA_XERXES pointer for record '%s' is NULL.", record->name);
	}

	mx_status = mxi_xia_xerxes_restore_config( xia_xerxes );

	return mx_status;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxi_xia_xerxes_is_busy( MX_MCA *mca,
			int *busy_flag,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_read_status()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	int xia_status, active;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	xia_status = dxp_isrunning( &(xia_dxp_mca->system_channel), &active );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot get DXP running status for MCA '%s'.  "
		"Error code = %d, '%s'",
			mca->record->name, xia_status,
			mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* We are only interested in the two lowest order bits.
	 *
	 * Bit 0 tells you whether or not the board things a run
	 * is active.  Bit 1 tells you whether or not XerXes
	 * thinks a run is active.
	 */

	active &= 0x3;

	MX_DEBUG(-2,("%s: active = %#x", fname, active));

	switch( active ) {
	case 0x0:
		*busy_flag = FALSE;
		break;
	case 0x1:
		(void) mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Inconsistent active state flags %#x: "
		"The DXP board for MCA '%s' says that it is active, "
		"but XerXes says that it is not.",
			active, mca->record->name );
		*busy_flag = TRUE;
		break;
	case 0x2:
		(void) mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Inconsistent active state flags %#x: "
		"XerXes says that MCA '%s' is active, but the DXP board "
		"says that it is not.",
			active, mca->record->name );
		*busy_flag = FALSE;
		break;
	case 0x3:
		*busy_flag = TRUE;
		break;
	default:
		return mx_error( MXE_FUNCTION_FAILED, fname,
		"It should not be possible to get to this message.  "
		"If you do, then something peculiar has happened "
		"for MCA '%s' and this should be reported to the programmer.",
			mca->record->name );
		break;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_read_parameter( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type *value_ptr,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_read_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	unsigned short short_value;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Record '%s', channel %d, parameter '%s'.",
			fname, mca->record->name,
			xia_dxp_mca->system_channel, parameter_name));
	}

	xia_status = dxp_get_one_dspsymbol( &(xia_dxp_mca->system_channel),
						parameter_name, &short_value );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot read DXP parameter '%s' for system channel %d "
		"of XIA DXP interface '%s'.  "
		"Error code = %d, '%s'",
			parameter_name, xia_dxp_mca->system_channel,
			mca->record->name, xia_status,
			mxi_xia_xerxes_strerror( xia_status ) );
	}

	*value_ptr = (mx_uint32_type) short_value;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: value = %lu", fname, *value_ptr));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_write_parameter( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type value,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_write_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	unsigned short short_value;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Record '%s', channel %d, parameter '%s', value = %lu.",
			fname, mca->record->name, xia_dxp_mca->system_channel,
			parameter_name, (unsigned long) value));
	}

	short_value = (unsigned short) value;

	xia_status = dxp_set_one_dspsymbol( &(xia_dxp_mca->system_channel),
						parameter_name, &short_value );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot write the value %hu to DXP parameter '%s' "
		"for system channel %d of XIA DXP interface '%s'.  "
		"Error code = %d, '%s'", short_value,
			parameter_name, xia_dxp_mca->system_channel,
			mca->record->name, xia_status,
			mxi_xia_xerxes_strerror( xia_status ) );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Write to parameter '%s' succeeded.",
			fname, parameter_name));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_write_parameter_to_all_channels( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type value,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_write_parameter_to_all_channels()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	unsigned short short_value;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: Record '%s', parameter '%s', value = %lu.",
			fname, mca->record->name,
			parameter_name, (unsigned long) value));
	}

	short_value = (unsigned short) value;

#if 0
	/* The canonical method which does not work. */

	xia_status = dxp_set_dspsymbol( parameter_name, &short_value );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot write the value %hu to DXP parameter '%s' "
		"for all channels of XIA DXP interface '%s'.  ",
		"Error code = %d, '%s'",
			short_value, parameter_name, mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}
#else
	/* A temporary kludge which does work. */

	{
		MX_RECORD *mca_record;
		MX_XIA_DXP_MCA *local_dxp_mca;
		unsigned long i;

		for ( i = 0; i < xia_xerxes->num_mcas; i++ ) {

			mca_record = xia_xerxes->mca_record_array[i];

			if ( mca_record == (MX_RECORD *) NULL ) {

				continue;    /* Skip this one. */
			}

			local_dxp_mca = (MX_XIA_DXP_MCA *)
					mca_record->record_type_struct;

			MX_DEBUG(-2,("%s: writing to system channel %d.",
				fname, local_dxp_mca->system_channel));

			xia_status = dxp_set_one_dspsymbol(
					&(local_dxp_mca->system_channel),
					parameter_name, &short_value );

			if ( xia_status != DXP_SUCCESS ) {
				return mx_error( MXE_DEVICE_ACTION_FAILED,fname,
			"Cannot write the value %hu to DXP parameter '%s' "
			"for system channel %d of XIA DXP interface '%s'.  "
			"Error code = %d, '%s'", short_value,
				parameter_name, local_dxp_mca->system_channel,
				mca->record->name, xia_status,
				mxi_xia_xerxes_strerror( xia_status ) );
			}
		}
	}
#endif
	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Write parameter '%s' to all channels succeeded.",
			fname, parameter_name));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_start_run( MX_MCA *mca,
			int clear_flag,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_start_run()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	unsigned short gate, resume_flag;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: Record '%s', clear_flag = %d.",
				fname, mca->record->name, clear_flag));
	}
#if 0
	/* Kludge: explicitly set the values of the GAINDAC settings
	 * since Xerxes does not seem to be doing it for us.
	 */

	{
		MX_RECORD *mca_record;
		MX_XIA_DXP_MCA *local_dxp_mca;
		unsigned short gaindac;
		int i;

		for ( i = 0; i < xia_xerxes->num_mcas; i++ ) {

			mca_record = xia_xerxes->mca_record_array[i];

			if ( mca_record == (MX_RECORD *) NULL ) {

				continue;	/* Skip this one. */
			}

			local_dxp_mca = (MX_XIA_DXP_MCA *)
					mca_record->record_type_struct;

			gaindac = (unsigned short)
					xia_xerxes->gaindac_kludge_array[i];

			MX_DEBUG(-2,
				("%s: writing GAINDAC ( %#hx ) to MCA '%s'.",
				fname, gaindac, mca_record->name ));

			xia_status = dxp_set_one_dspsymbol(
					&(local_dxp_mca->system_channel),
					"GAINDAC", &gaindac );

			if ( xia_status != DXP_SUCCESS ) {
				return mx_error( MXE_DEVICE_ACTION_FAILED,fname,
			"Cannot write the GAINDAC value %#hx to MCA '%s'.  "
			"Error code = %d, '%s'",
				gaindac, mca_record->name, xia_status,
				mxi_xia_xerxes_strerror( xia_status ) );
			}
		}
	}
#endif

	gate = 1;

	if ( clear_flag ) {
		resume_flag = 0;
	} else {
		resume_flag = 1;
	}

	xia_status = dxp_start_run( &gate, &resume_flag );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot start data acquisition for XIA DXP interface '%s'.  "
		"Error code = %d, '%s'", mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	return MX_SUCCESSFUL_RESULT;
}


MX_EXPORT mx_status_type
mxi_xia_xerxes_stop_run( MX_MCA *mca,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_stop_run()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: Record '%s'.", fname, mca->record->name));
	}

	xia_status = dxp_stop_run();

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot stop data acquisition for XIA DXP interface '%s'.  "
		"Error code = %d, '%s'", mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_read_spectrum( MX_MCA *mca,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_read_spectrum()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	int xia_status;
	unsigned long *array_ptr;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( xia_dxp_mca->use_mca_channel_array ) {
		array_ptr = mca->channel_array;
	} else {
		array_ptr = xia_dxp_mca->spectrum_array;
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: reading out %ld channels from MCA '%s'.",
			fname, mca->current_num_channels, mca->record->name));
	}

	xia_status = dxp_readout_detector_run(
				&(xia_dxp_mca->system_channel),
				xia_dxp_mca->parameter_array,
				xia_dxp_mca->baseline_array,
				array_ptr );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot read the spectrum for XIA MCA '%s'.  "
		"Error code = %d, '%s'", mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: readout from MCA '%s' complete.",
			fname, mca->record->name));
	}

	return MX_SUCCESSFUL_RESULT;
}

#define TWO_TO_THE_32ND_POWER	4294967296.0

MX_EXPORT mx_status_type
mxi_xia_xerxes_read_statistics( MX_MCA *mca,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_read_statistics()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	unsigned long nevent;
	int xia_status;
	unsigned short parameter_array[3];
	double real_time;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	xia_status = dxp_get_statistics( &(xia_dxp_mca->system_channel),
					&(mca->live_time),
					&(xia_dxp_mca->input_count_rate),
					&(xia_dxp_mca->output_count_rate),
					&nevent );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot read statistics for XIA DXP interface '%s'.  "
		"Error code = %d, '%s'", mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,(
"%s: Record '%s', channel %d, live_time = %g, icr = %g, ocr = %g, nevent = %lu",
			fname, mca->record->name, xia_dxp_mca->system_channel,
			mca->live_time, xia_dxp_mca->input_count_rate,
			xia_dxp_mca->output_count_rate, nevent ));
	}

	xia_dxp_mca->num_events = (mx_uint32_type) nevent;

	xia_dxp_mca->num_fast_peaks = (mx_uint32_type) mx_round(
				xia_dxp_mca->input_count_rate * mca->live_time);

	xia_dxp_mca->num_underflows = 0;
	xia_dxp_mca->num_overflows = 0;

	mx_status = mxi_xia_xerxes_read_dspsymbol_array( xia_dxp_mca,
					"REALTIME", 3, parameter_array );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	real_time = (double) parameter_array[1] + 65536.0 * parameter_array[0]
				+ TWO_TO_THE_32ND_POWER * parameter_array[2];

	real_time *= ( 400.0e-9 );

	mca->real_time = real_time;

#if 1
	xia_dxp_mca->num_underflows = 0;
	xia_dxp_mca->num_overflows = 0;
#else
	mx_status = mxi_xia_xerxes_read_dspsymbol_array( xia_dxp_mca,
					"UNDERFLOWS", 2, parameter_array );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	xia_dxp_mca->num_underflows = (mx_uint32_type) parameter_array[0]
				+ 65536L * (mx_uint32_type) parameter_array[1];

	mx_status = mxi_xia_xerxes_read_dspsymbol_array( xia_dxp_mca,
					"OVERFLOWS", 2, parameter_array );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	xia_dxp_mca->num_overflows = (mx_uint32_type) parameter_array[0]
				+ 65536L * (mx_uint32_type) parameter_array[1];
#endif

	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

static mx_status_type
mxi_xia_xerxes_read_dspsymbol_array( MX_XIA_DXP_MCA *xia_dxp_mca,
				char *dspsymbol_namestem,
				int num_array_elements,
				unsigned short *dspsymbol_array )
{
	const char fname[] = "mxi_xia_xerxes_read_dspsymbol_array()";

	char dspsymbol_name[20];
	int i, xia_status;

	for ( i = 0; i < num_array_elements; i++ ) {

		sprintf( dspsymbol_name, "%s%d", dspsymbol_namestem, i );

		xia_status = dxp_get_one_dspsymbol(
					&(xia_dxp_mca->system_channel),
					dspsymbol_name,
					&(dspsymbol_array[i]) );

		if ( xia_status != DXP_SUCCESS ) {
			return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
			"Cannot read DXP parameter '%s' for system channel %d "
			"of XIA DXP MCA '%s'.  "
			"Error code = %d, '%s'",
				dspsymbol_name, xia_dxp_mca->system_channel,
				xia_dxp_mca->record->name, xia_status,
				mxi_xia_xerxes_strerror( xia_status ) );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

const char *
mxi_xia_xerxes_strerror( int xerxes_status )
{
	static struct {
		int xerxes_error_code;
		const char xerxes_error_string[30];
	} error_code_table[] = {
		{   0, "DXP_SUCCESS" },
		{   1, "DXP_MDOPEN" },
		{   2, "DXP_MDIO" },
		{   3, "DXP_MDINITIALIZE" },
		{   4, "DXP_MDLOCK" },
		{ 101, "DXP_WRITE_TSAR" },
		{ 102, "DXP_WRITE_CSR" },
		{ 103, "DXP_WRITE_WORD" },
		{ 104, "DXP_READ_WORD" },
		{ 105, "DXP_WRITE_BLOCK" },
		{ 106, "DXP_READ_BLOCK" },
		{ 107, "DXP_DISABLE_LAM" },
		{ 108, "DXP_CLR_LAM" },
		{ 109, "DXP_TEST_LAM" },
		{ 110, "DXP_READ_CSR" },
		{ 111, "DXP_WRITE_FIPPI" },
		{ 112, "DXP_WRITE_DSP" },
		{ 113, "DXP_WRITE_DATA" },
		{ 114, "DXP_READ_DATA" },
		{ 115, "DXP_ENABLE_LAM" },
		{ 116, "DXP_READ_GSR" },
		{ 117, "DXP_WRITE_GCR" },
		{ 118, "DXP_WRITE_WCR" },
		{ 119, "DXP_READ_WCR" },
		{ 201, "DXP_MEMERROR" },
		{ 202, "DXP_DSPRUNERROR" },
		{ 203, "DXP_FIPDOWNLOAD" },
		{ 204, "DXP_DSPDOWNLOAD" },
		{ 205, "DXP_INTERNAL_GAIN" },
		{ 206, "DXP_RESET_WARNING" },
		{ 207, "DXP_DETECTOR_GAIN" },
		{ 208, "DXP_NOSYMBOL" },
		{ 209, "DXP_SPECTRUM" },
		{ 210, "DXP_DSPLOAD" },
		{ 211, "DXP_DSPPARAMS" },
		{ 212, "DXP_DSPACCESS" },
		{ 213, "DXP_DSPPARAMBOUNDS" },
		{ 214, "DXP_ADC_RUNACTIVE" },
		{ 215, "DXP_ADC_READ" },
		{ 216, "DXP_ADC_TIMEOUT" },
		{ 217, "DXP_ALLOCMEM" },
		{ 218, "DXP_NOCONTROLTYPE" },
		{ 219, "DXP_NOFIPPI" },
		{ 301, "DXP_BAD_PARAM" },
		{ 302, "DXP_NODECIMATION" },
		{ 303, "DXP_OPEN_FILE" },
		{ 304, "DXP_NORUNTASKS" },
		{ 305, "DXP_OUTPUT_UNDEFINED" },
		{ 306, "DXP_INPUT_UNDEFINED" },
		{ 307, "DXP_ARRAY_TOO_SMALL" },
		{ 308, "DXP_NOCHANNELS" },
		{ 309, "DXP_NODETCHAN" },
		{ 310, "DXP_NOIOCHAN" },
		{ 311, "DXP_NOMODCHAN" },
		{ 312, "DXP_NEGBLOCKSIZE" },
		{ 313, "DXP_FILENOTFOUND" },
		{ 314, "DXP_NOFILETABLE" },
		{ 315, "DXP_INITIALIZE" },
		{ 316, "DXP_UNKNOWN_BTYPE" },
		{ 317, "DXP_NOMATCH" },
		{ 318, "DXP_BADCHANNEL" },
		{ 319, "DXP_DSPTIMEOUT" },
		{ 320, "DXP_INITORDER" },
		{ 401, "DXP_NOMEM" },
		{ 403, "DXP_CLOSE_FILE" },
		{ 404, "DXP_INDEXOOB" },
		{ 405, "DXP_RUNACTIVE" },
		{ 406, "DXP_MEMINUSE" },
		{1001, "DXP_DEBUG" }
	};

	static const char unrecognized_error_string[]
		= "Unrecognized XerXes error code";

	static int num_error_codes = sizeof( error_code_table )
				/ sizeof( error_code_table[0] );

	unsigned int i;

	for ( i = 0; i < num_error_codes; i++ ) {

		if ( xerxes_status == error_code_table[i].xerxes_error_code )
			break;		/* Exit the for() loop. */
	}

	if ( i >= num_error_codes ) {
		return &unrecognized_error_string[0];
	}

	return &( error_code_table[i].xerxes_error_string[0] );
}

#endif /* HAVE_XIA_XERXES */
