/*
 * Name:    mx_scan.c
 *
 * Purpose: Scan record support for the MX library.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 1999-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_key.h"
#include "mx_driver.h"
#include "mx_record.h"
#include "mx_scan.h"
#include "mx_scan_linear.h"
#include "mx_relay.h"
#include "mx_analog_input.h"
#include "mx_analog_output.h"
#include "mx_digital_input.h"
#include "mx_digital_output.h"
#include "mx_scaler.h"
#include "mx_amplifier.h"
#include "mx_mca.h"
#include "mx_variable.h"
#include "mx_alloc.h"

/* --------------- */

static mx_status_type ( *mx_scan_pause_request_handler ) ( MX_SCAN * )
				= mx_default_scan_pause_request_handler;

static mx_status_type mx_setup_scan_shutter( MX_SCAN * );

static mx_status_type mx_setup_measurement_permit_and_fault_handlers(
								MX_SCAN * );

static mx_status_type mx_scan_free_measurement_permit_and_fault_handlers(
								MX_SCAN * );

/* --------------- */

/* mx_scan_finish_record_initialization() performs data structure
 * initialization steps that must be done for all scan records.
 */

MX_EXPORT mx_status_type
mx_scan_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mx_scan_finish_record_initialization()";

	MX_SCAN *scan;
	MX_MEASUREMENT_TYPE_ENTRY *measurement_type_entry;
	MX_DATAFILE_TYPE_ENTRY *datafile_type_entry;
	MX_PLOT_TYPE_ENTRY *plot_type_entry;
	mx_status_type status;

	scan = ( MX_SCAN * ) record->record_superclass_struct;

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SCAN pointer for scan record '%s' is NULL.",
			record->name );
	}

	/* Find the driver for the specified measurement type. */

	status = mx_get_measurement_type_by_name( mx_measurement_type_list,
			scan->measurement_type, &measurement_type_entry );

	if ( status.code != MXE_SUCCESS )
		return status;

	scan->measurement.type              = measurement_type_entry->type;
	scan->measurement.typename          = scan->measurement_type;
	scan->measurement.measurement_type_struct = NULL;
	scan->measurement.measurement_function_list
			= measurement_type_entry->measurement_function_list;
	scan->measurement.measurement_arguments = scan->measurement_arguments;

	/* Find the driver for the specified datafile type. */

	status = mx_get_datafile_type_by_name( mx_datafile_type_list,
			scan->datafile_type, &datafile_type_entry );

	if ( status.code != MXE_SUCCESS )
		return status;

	scan->datafile.type                 = datafile_type_entry->type;
	scan->datafile.typename             = scan->datafile_type;
	scan->datafile.datafile_type_struct = NULL;
	scan->datafile.datafile_function_list
				= datafile_type_entry->datafile_function_list;
	scan->datafile.filename             = scan->datafile_name;

	/* Find the driver for the specified plot type. */

	status = mx_get_plot_type_by_name( mx_plot_type_list,
			scan->plot_type, &plot_type_entry );

	if ( status.code != MXE_SUCCESS )
		return status;

	scan->plot.type               = plot_type_entry->type;
	scan->plot.typename           = scan->plot_type;
	scan->plot.plot_type_struct   = NULL;
	scan->plot.plot_function_list = plot_type_entry->plot_function_list;
	scan->plot.plot_arguments     = scan->plot_arguments;

	scan->shutter_record = NULL;

	scan->num_measurement_permit_handlers = 0;
	scan->num_measurement_fault_handlers = 0;

	scan->measurement_permit_handler_array = NULL;
	scan->measurement_fault_handler_array = NULL;

	return MX_SUCCESSFUL_RESULT;
}

/* --------------- */

MX_EXPORT mx_status_type
mx_scan_print_scan_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mx_scan_print_scan_structure()";

	MX_SCAN *scan;
	MX_DRIVER *driver;
	int num_items;
	long i, measurement_counts;
	double measurement_time;
	size_t length;
	char format[40];
	char device_name[MXU_RECORD_NAME_LENGTH+1];
	char scan_type[MXU_DRIVER_NAME_LENGTH+1];

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	scan = (MX_SCAN *) record->record_superclass_struct;

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SCAN pointer for record '%s' is NULL.", record->name );
	}

	driver = mx_get_driver_by_type( mx_type_list, record->mx_type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record type %ld not found in mx_type_list.",
			record->mx_type );
	}

	mx_strncpy( scan_type, driver->name, MXU_DRIVER_NAME_LENGTH );

	length = strlen(scan_type);

	for ( i = 0; i < length; i++ ) {
		if ( islower( (int) (scan_type[i]) ) ) {
			scan_type[i] = toupper( scan_type[i] );
		}
	}

	fprintf( file, "  Scan type            = %s\n\n", scan_type );

	fprintf( file, "  Name                 = %s\n", record->name );
	fprintf( file, "  Number of scans      = %ld\n", scan->num_scans );
	fprintf( file, "\n" );

	if ( record->mx_type != MXS_XAF_STANDARD ) {
		if ( scan->num_motors == 1 ) {
			fprintf( file, "  1 motor:         " );
		} else {
			fprintf( file, "  %ld motors:        ",
					scan->num_motors );
		}
		for ( i = 0; i < scan->num_motors; i++ ) {
			if ( i > 0 ) {
				fprintf( file, ", " );
			}
			fprintf( file, "%s",
				scan->motor_record_array[i]->name );
		}
		fprintf( file, "\n" );
	}

	if ( scan->num_input_devices == 1 ) {
		fprintf( file, "  1 input device:  " );
	} else {
		fprintf( file, "  %ld input devices: ",
			scan->num_input_devices );
	}
	for ( i = 0; i < scan->num_input_devices; i++ ) {
		if ( i > 0 ) {
			fprintf( file, ", " );
		}
		fprintf( file, "%s", scan->input_device_array[i]->name );
	}
	fprintf( file, "\n\n" );

	fprintf( file, "  Settling time        = %g seconds\n\n",
					scan->settling_time );

	if ( record->mx_type != MXS_XAF_STANDARD ) {
		fprintf( file, "  Measurement type     = %s\n",
						scan->measurement_type );
	
		switch( scan->measurement.type ) {
		case MXM_PRESET_TIME:
			sprintf(format, "%%lg %%%ds", MXU_RECORD_NAME_LENGTH);
	
			num_items = sscanf(scan->measurement_arguments, format,
					&measurement_time, device_name );
	
			if ( num_items != 2 ) {
				return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"Cannot parse measurement arguments '%s' for scan '%s'.",
					scan->measurement_arguments,
					record->name );
			}
			fprintf( file,
			"  Measurement time     = %g seconds ( Timer '%s' )\n",
					measurement_time, device_name );
			break;
	
		case MXM_PRESET_COUNT:
			sprintf(format, "%%ld %%%ds", MXU_RECORD_NAME_LENGTH);
	
			num_items = sscanf(scan->measurement_arguments, format,
					&measurement_counts, device_name );
	
			if ( num_items != 2 ) {
				return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"Cannot parse measurement arguments '%s' for scan '%s'.",
					scan->measurement_arguments,
					record->name );
			}
			fprintf( file,
			"  Measurement preset   = %ld counts ( Scaler '%s' )\n",
					measurement_counts, device_name );
			break;
	
		default:
			fprintf( file, "Measurement arguments  = '%s'\n",
				scan->measurement_arguments );
			break;
		}
		fprintf( file, "\n" );
	}

	fprintf( file, "  Datafile type        = %s\n", scan->datafile_type );
	fprintf( file, "  Datafile name        = %s\n", scan->datafile_name );
	fprintf( file, "  Plot type            = %s\n", scan->plot_type );
	fprintf( file, "  Plot arguments       = %s\n", scan->plot_arguments );
	fprintf( file, "\n" );

	return MX_SUCCESSFUL_RESULT;
}

/* --------------- */

MX_EXPORT mx_status_type
mx_perform_scan( MX_RECORD *scan_record )
{
	const char fname[] = "mx_perform_scan()";

	MX_SCAN *scan;
	MX_SCAN_FUNCTION_LIST *scan_function_list;
	mx_status_type ( *prepare_for_scan_start_fn ) ( MX_SCAN * );
	mx_status_type ( *execute_scan_body_fn ) ( MX_SCAN * );
	mx_status_type ( *cleanup_after_scan_end_fn ) ( MX_SCAN * );
	MX_LIST_HEAD *list_head;
	long i;
	mx_status_type status;

	MX_DEBUG( 2, ( "%s invoked.", fname ) );

	/*** Find and check a lot of pointer variables that we will need. ***/

	if ( scan_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_RECORD pointer passed was NULL." );
	}

	scan = (MX_SCAN *) scan_record->record_superclass_struct;

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SCAN pointer for scan record '%s' is NULL.",
				scan_record->name );
	}

	scan_function_list = (MX_SCAN_FUNCTION_LIST *)
			scan_record->superclass_specific_function_list;

	if ( scan_function_list == (MX_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"Scan function list pointer for scan '%s' is NULL.",
			scan_record->name );
	}

	prepare_for_scan_start_fn = scan_function_list->prepare_for_scan_start;

	if ( prepare_for_scan_start_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"prepare_for_scan_start function pointer for scan '%s' is NULL.",
			scan_record->name );
	}

	execute_scan_body_fn = scan_function_list->execute_scan_body;

	if ( execute_scan_body_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"execute_scan_body function pointer for scan '%s' is NULL.",
			scan_record->name );
	}

	cleanup_after_scan_end_fn = scan_function_list->cleanup_after_scan_end;

	if ( cleanup_after_scan_end_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"cleanup_after_scan_end function pointer for scan '%s' is NULL.",
			scan_record->name );
	}

	list_head = mx_get_record_list_head_struct( scan_record );

	if ( list_head == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_LIST_HEAD pointer for scan record '%s' is NULL.",
			scan_record->name );
	}

	/*** Record that the scan has started. ***/

	if ( list_head->log_handler != NULL ) {
		mx_log_scan_start( list_head, scan );
	}

	/*** Configure the measurement type. ***/

	status = mx_configure_measurement_type( &(scan->measurement) );

	if ( status.code != MXE_SUCCESS ) {
		mx_log_scan_end( list_head, scan, status );
		return status;
	}

	/*** Loop through the iterations of the scan. ***/

	for ( i = 0; i < scan->num_scans; i++ ) {

		scan->current_scan_number = i;

		if ( scan->num_scans > 1 ) {
			mx_info( "Starting scan %ld of %ld.",
					i+1, scan->num_scans );
		}

		status = prepare_for_scan_start_fn( scan );

		if ( status.code != MXE_SUCCESS ) {
			mx_handle_abnormal_scan_termination(
					list_head, scan, status );
			return status;
		}

		/* Perform any measurement type specific processing required
		 * before the start of the scan.
		 */

		status = mx_prescan_measurement_processing(
						&(scan->measurement) );

		if ( status.code != MXE_SUCCESS ) {
			mx_handle_abnormal_scan_termination(
					list_head, scan, status );
			return status;
		}

		/*** Open the shutter if requested. ***/

		if ( scan->shutter_policy == MXF_SHUTTER_OPEN_FOR_SCAN ) {
			status = mx_relay_command( scan->shutter_record,
						MXF_OPEN_RELAY );

			if ( status.code != MXE_SUCCESS ) {
				mx_handle_abnormal_scan_termination(
						list_head, scan, status );
				return status;
			}
		}

		/*** Perform the scan. ***/

		scan->execute_scan_body_status = execute_scan_body_fn( scan );

		/* Close the shutter now if requested. */

		if ( scan->shutter_policy != MXF_SHUTTER_IGNORE ) {
			(void) mx_relay_command( scan->shutter_record,
						MXF_CLOSE_RELAY );
		}

		/* Perform any measurement type specific processing required
		 * after the end of the scan.
		 */

		status = mx_postscan_measurement_processing(
						&(scan->measurement) );

		if ( status.code != MXE_SUCCESS ) {
			mx_handle_abnormal_scan_termination(
					list_head, scan, status );
			return status;
		}

		/* Check the status of the scan that just completed. */

		switch( scan->execute_scan_body_status.code ) {
		case MXE_SUCCESS:
			mx_info( "Scan '%s' complete.", scan_record->name );
			break;
		case MXE_INTERRUPTED:
			mx_info( "Scan '%s' aborted.", scan_record->name );
			break;
		default:
			mx_info( "Scan '%s' terminated with error code %ld.",
				scan_record->name,
				scan->execute_scan_body_status.code );
			break;
		}

		status = cleanup_after_scan_end_fn( scan );

		if ( status.code != MXE_SUCCESS ) {
			mx_handle_abnormal_scan_termination(
					list_head, scan, status );
			return status;
		}
	}

	/* Deconfigure the measurement type. */

	(void) mx_deconfigure_measurement_type( &(scan->measurement) );

	/* Record that the scan has ended. */

	if ( list_head->log_handler != NULL ) {
		mx_log_scan_end( list_head, scan,
					scan->execute_scan_body_status );
	}
	return MX_SUCCESSFUL_RESULT;
}

/* --------------- */

MX_EXPORT mx_status_type
mx_standard_prepare_for_scan_start( MX_SCAN *scan )
{
	const char fname[] = "mx_standard_prepare_for_scan_start()";

	mx_status_type status;

	MX_DEBUG( 2, ( "%s invoked.", fname ) );

	scan->measurement_number = 0L;

	/* ==== Setup a scan shutter if requested. ==== */

	status = mx_setup_scan_shutter( scan );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* ==== Setup scan enable/disable support if requested. ==== */

	status = mx_setup_measurement_permit_and_fault_handlers( scan );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* ==== Try to open the datafile. ==== */

	status = mx_datafile_open( &(scan->datafile) );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* ==== Initialize graphical plotting of scan results. ==== */

	if ( mx_plotting_is_enabled( scan->record ) ) {

		status = mx_plot_open( &(scan->plot) );

		if ( status.code != MXE_SUCCESS )
			return status;

		scan->plot.section_number = 0;
	}

	/* ==== Write out the main scan header. ==== */

	status = mx_datafile_write_main_header( &(scan->datafile) );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* ==== Put all MX servers into fast mode. ==== */

	status = mx_set_fast_mode( scan->record, TRUE );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

/* --------------- */

MX_EXPORT mx_status_type
mx_standard_cleanup_after_scan_end( MX_SCAN *scan )
{
	const char fname[] = "mx_standard_cleanup_after_scan_end()";

	MX_LIST_HEAD *list_head;
	int enable_flag, prompt_for_plot_close;
	mx_status_type datafile_close_status;

	MX_DEBUG( 2, ( "%s invoked.", fname ) );

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"Scan routine was passed a NULL pointer.");
	}

	list_head = mx_get_record_list_head_struct( scan->record );

	if ( list_head == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_LIST_HEAD pointer for scan record '%s' is NULL.",
			scan->record->name );
	}

	/* Make sure the datafile is closed. */

	datafile_close_status = mx_datafile_close( &(scan->datafile) );

	/* Turn off fast mode in the MX servers. */

	(void) mx_set_fast_mode( scan->record, FALSE );

	/* Free unneeded handlers. */

	mx_scan_free_measurement_permit_and_fault_handlers( scan );

	/* Handle the user's final interactions with the plot. */

	enable_flag = mx_plotting_is_enabled( scan->record );

	prompt_for_plot_close = TRUE;

	switch ( enable_flag ) {
	case MXPF_PLOT_OFF:
	case MXPF_PLOT_NOWAIT:
		prompt_for_plot_close = FALSE;

		break;
	default:
		if (scan->plot.type == MXP_CHILD
		 || scan->plot.type == MXP_NONE ) {

			prompt_for_plot_close = FALSE;

		} else if ( scan->execute_scan_body_status.code != MXE_SUCCESS )
		{
			prompt_for_plot_close = TRUE;

		} else if ( scan->num_scans - scan->current_scan_number > 1 ) {

			prompt_for_plot_close = FALSE;
		} else {
			prompt_for_plot_close = TRUE;
		}
		break;
	}

	if ( prompt_for_plot_close ) {

			mx_info_dialog(
	"\aHit any key to close the plot...",
	"Scan Complete.  Click the OK button to close the plot", "OK");

	} else {
		/* Give the plotting program a little time to finish
		 * plotting the last data point.
		 */

		mx_msleep(500);
	}

	if ( enable_flag != MXPF_PLOT_OFF ) {
		(void) mx_plot_close( &(scan->plot) );
	}

	if ( scan->execute_scan_body_status.code != MXE_SUCCESS ) {
		return scan->execute_scan_body_status;
	}
	if ( datafile_close_status.code != MXE_SUCCESS ) {
		return datafile_close_status;
	}

	return MX_SUCCESSFUL_RESULT;
}

/* --------------- */

/* This function restores motor speeds and synchronous motion mode flags. */

MX_EXPORT mx_status_type
mx_scan_restore_speeds( MX_SCAN *scan )
{
	const char fname[] = "mx_scan_restore_speeds()";

	MX_RECORD *motor_record;
	long i;

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_SCAN pointer passed was NULL." );
	}

	for ( i = 0; i < scan->num_motors; i++ ) {

		motor_record = (scan->motor_record_array)[i];

		/* Errors in restoring the motor speed are ignored. */

		(void) mx_motor_restore_speed( motor_record );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* --------------- */

static mx_status_type
mx_setup_scan_shutter( MX_SCAN *scan )
{
	const char fname[] = "mx_setup_scan_shutter()";

	MX_RECORD *shutter_policy_record;
	char shutter_record_name[MXU_RECORD_NAME_LENGTH + 1];

	char format_buffer[20];
	char buffer[80];
	int num_items;
	mx_status_type status;

	MX_DEBUG( 2, ( "%s invoked.", fname ) );

	/* ==== Setup a scan shutter if requested. ==== */

	shutter_policy_record = mx_get_record( scan->record,
					MX_SCAN_SHUTTER_POLICY_RECORD_NAME );

	if ( shutter_policy_record == NULL ) {
		scan->shutter_policy = MXF_SHUTTER_IGNORE;
		scan->shutter_record = NULL;
	} else {
		status = mx_get_string_variable( shutter_policy_record,
						buffer, sizeof(buffer) );

		if ( status.code != MXE_SUCCESS )
			return status;

		sprintf( format_buffer, "%%ld %%%ds", MXU_RECORD_NAME_LENGTH );

		num_items = sscanf( buffer, format_buffer,
					&(scan->shutter_policy),
					shutter_record_name );

		if ( num_items != 2 ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The shutter policy control record '%s' is not correctly formatted.",
				MX_SCAN_SHUTTER_POLICY_RECORD_NAME );
		}

		/* Check that the shutter policy has a valid value. */

		switch( scan->shutter_policy ) {
		case MXF_SHUTTER_IGNORE:
		case MXF_SHUTTER_OPEN_FOR_SCAN:
		case MXF_SHUTTER_OPEN_FOR_DATAPOINT:
			break;
		default:
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The '%s' record is set to an illegal value of %ld.  "
"Allowed values are 0=ignore, 1=open for scan, 2=open for datapoint",
				MX_SCAN_SHUTTER_POLICY_RECORD_NAME,
				scan->shutter_policy );
			break;
		}

		/* Find the shutter record. */

		scan->shutter_record = mx_get_record( scan->record,
						shutter_record_name );

		if ( scan->shutter_record == NULL ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The shutter '%s' specified in the shutter policy control record '%s' "
"does not exist.", shutter_record_name, MX_SCAN_SHUTTER_POLICY_RECORD_NAME );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/* --------------- */

static mx_status_type
mx_get_scan_handler_name_array( MX_SCAN *scan, char *list_record_name,
			long *num_handlers, char ***handler_string_array )
{
	const char fname[] = "mx_get_scan_handler_name_array()";

	MX_RECORD *list_record;
	long num_dimensions, field_type;
	long *dimension_array;
	void *pointer_to_value;
	mx_status_type mx_status;

	*num_handlers = 0;

	list_record = mx_get_record( scan->record, list_record_name );

	if ( list_record == (MX_RECORD *) NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	mx_status = mx_get_variable_parameters( list_record,
					&num_dimensions,
					&dimension_array,
					&field_type,
					&pointer_to_value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( field_type != MXFT_STRING ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
"The measurement handler list '%s' is not a 2 dimensional string variable.",
			list_record_name );
	}
	if ( num_dimensions != 2 ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
"The measurement handler list '%s' is not a 2 dimensional string variable.  "
"Instead it has %ld dimensions.", list_record_name, num_dimensions );
	}
	if ( dimension_array[0] <= 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The measurement handler list '%s' has an illegal first dimension size of %ld."
"  This size must be greater than zero.", list_record_name, dimension_array[0]);
	}

	*num_handlers = dimension_array[0];

	*handler_string_array = (char **) pointer_to_value;

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mx_setup_measurement_permit_and_fault_handlers( MX_SCAN *scan )
{
	const char fname[] = "mx_setup_measurement_permit_and_fault_handlers()";

	MX_MEASUREMENT_PERMIT *permit_handler;
	MX_MEASUREMENT_FAULT *fault_handler;
	char **handler_string_array;
	long i, num_permit_entries, num_fault_entries;
	mx_status_type mx_status;

	scan->num_measurement_permit_handlers = 0;
	scan->num_measurement_fault_handlers = 0;

	/* Setup the permit handlers. */

	mx_status = mx_get_scan_handler_name_array( scan,
				MX_SCAN_PERMIT_HANDLER_LIST,
				&num_permit_entries,
				&handler_string_array );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( num_permit_entries > 0 ) {

		scan->measurement_permit_handler_array =
			(MX_MEASUREMENT_PERMIT **) malloc( num_permit_entries
					* sizeof( MX_MEASUREMENT_PERMIT * ) );

		if ( scan->measurement_permit_handler_array == NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Could not allocate %ld MX_MEASUREMENT_PERMIT handlers for scan '%s'.",
			num_permit_entries, scan->record->name );
		}
	}

	for ( i = 0; i < num_permit_entries; i++ ) {

		mx_status = mx_measurement_permit_create_handler(
				&permit_handler, scan,
				handler_string_array[i] );

		if ( mx_status.code != MXE_SUCCESS ) {
		    mx_scan_free_measurement_permit_and_fault_handlers( scan );

		    return mx_status;
		}

		/* If mx_measurement_permit_create_handler() returns a NULL
		 * pointer, that means that this scan does not need to use
		 * this permit handler.  Otherwise, add the returned handler
		 * to the array.
		 */

		if ( permit_handler != (MX_MEASUREMENT_PERMIT *) NULL ) {

			scan->measurement_permit_handler_array[ 
				scan->num_measurement_permit_handlers ] 
							= permit_handler;

			(scan->num_measurement_permit_handlers)++;
		}
	}

	/* Setup the fault handlers. */

	mx_status = mx_get_scan_handler_name_array( scan,
				MX_SCAN_FAULT_HANDLER_LIST,
				&num_fault_entries,
				&handler_string_array );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_scan_free_measurement_permit_and_fault_handlers( scan );

		return mx_status;
	}

	if ( num_fault_entries > 0 ) {

		scan->measurement_fault_handler_array =
			(MX_MEASUREMENT_FAULT **) malloc( num_fault_entries
					* sizeof( MX_MEASUREMENT_FAULT * ) );

		if ( scan->measurement_fault_handler_array == NULL ) {
		    mx_scan_free_measurement_permit_and_fault_handlers( scan );

		    return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Could not allocate %ld MX_MEASUREMENT_FAULT handlers for scan '%s'.",
			num_fault_entries, scan->record->name );
		}
	}

	for ( i = 0; i < num_fault_entries; i++ ) {

		mx_status = mx_measurement_fault_create_handler(
				&fault_handler, scan,
				handler_string_array[i] );

		if ( mx_status.code != MXE_SUCCESS ) {
		    mx_scan_free_measurement_permit_and_fault_handlers( scan );

		    return mx_status;
		}

		/* If mx_measurement_fault_create_handler() returns a NULL
		 * pointer, that means that this scan does not need to use
		 * this fault handler.  Otherwise, add the returned handler
		 * to the array.
		 */

		if ( fault_handler != (MX_MEASUREMENT_FAULT *) NULL ) {

			scan->measurement_fault_handler_array[ 
				scan->num_measurement_fault_handlers ] 
							= fault_handler;

			(scan->num_measurement_fault_handlers)++;
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mx_scan_free_measurement_permit_and_fault_handlers( MX_SCAN *scan )
{
	long i;

	if ( scan->measurement_permit_handler_array != NULL ) {

		for ( i = 0; i < scan->num_measurement_permit_handlers; i++ ) {

			(void) mx_measurement_permit_destroy_handler(
				scan->measurement_permit_handler_array[i] );

			scan->measurement_permit_handler_array[i] = NULL;
		}

		free( scan->measurement_permit_handler_array );

		scan->measurement_permit_handler_array = NULL;
	}

	if ( scan->measurement_fault_handler_array != NULL ) {

		for ( i = 0; i < scan->num_measurement_fault_handlers; i++ ) {

			(void) mx_measurement_fault_destroy_handler(
				scan->measurement_fault_handler_array[i] );

			scan->measurement_fault_handler_array[i] = NULL;
		}

		free( scan->measurement_fault_handler_array );

		scan->measurement_fault_handler_array = NULL;
	}

	return MX_SUCCESSFUL_RESULT;
}

/* --------------- */

MX_EXPORT mx_status_type
mx_scan_handle_data_measurement( MX_SCAN *scan )
{
	const char fname[] = "mx_scan_handle_data_measurement()";

	int fault_occurred, fault_status;
	long i;
	mx_status_type mx_status, measure_data_status;

	measure_data_status = MX_SUCCESSFUL_RESULT;

	/* Reset any faults that have occurred during the interval since
	 * the last measurement.
	 */

	for ( i = 0; i < scan->num_measurement_fault_handlers; i++ ) {

		mx_status = mx_measurement_fault_reset(
				scan->measurement_fault_handler_array[i],
				MXMF_PREPARE_FOR_FIRST_MEASUREMENT_ATTEMPT );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	/*** Begin the measurement fault handling loop ***/

	fault_occurred = TRUE;

	while ( fault_occurred == TRUE ) {

		MX_DEBUG( 2,("%s: beginning of fault_occurred loop.", fname ));

		if ( mx_user_requested_interrupt() ) {
			return mx_error( MXE_INTERRUPTED, fname,
			"The measurement retry loop was interrupted." );
		}

		/* Wait until all of the measurement permit handlers say
		 * that it is OK to take the measurement.
		 */

		for ( i = 0; i < scan->num_measurement_permit_handlers; i++ ) {

			mx_status = mx_measurement_permit_wait_for_permission(
				scan->measurement_permit_handler_array[i] );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			/* Put in a tiny sleep between tests. */

			mx_msleep(1);
		}

		/* Perform the measurement. */

		measure_data_status = mx_measure_data( &(scan->measurement) );

		if ( measure_data_status.code != MXE_SUCCESS )
			return measure_data_status;

		/* Did something bad happen during the measurement? */

		fault_occurred = FALSE;

		for ( i = 0; i < scan->num_measurement_fault_handlers; i++ ) {

			mx_status = mx_measurement_fault_check_for_fault(
				scan->measurement_fault_handler_array[i],
				&fault_status );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			if ( fault_status == TRUE ) {

				fault_occurred = TRUE;

				/* If a measurement fault occurred, reset it.*/

				mx_status = mx_measurement_fault_reset(
				    scan->measurement_fault_handler_array[i],
					MXMF_NONE );

				if ( mx_status.code != MXE_SUCCESS )
					return mx_status;
			}

			/* Put in a tiny sleep between tests. */

			mx_msleep(1);
		}

		MX_DEBUG( 2,("%s: end of fault_occurred loop.", fname ));

		if ( fault_occurred ) {
			mx_info(
"The measurement just performed was bad.  Retrying the measurement." );
		}

	}    /* End of measurement fault handling loop ***/

	return measure_data_status;
}

/* --------------- */

MX_EXPORT mx_status_type
mx_scan_handle_pause_request( MX_SCAN *scan )
{
	const char fname[] = "mx_scan_handle_pause_request()";

	mx_status_type status;

	if ( mx_scan_pause_request_handler != NULL ) {

		status = ( *mx_scan_pause_request_handler )( scan );
	} else {
		return mx_error( MXE_NULL_ARGUMENT, fname,
"The mx_scan_pause_request_handler pointer is NULL.  This shouldn't happen." );
	}

	return status;
}

MX_EXPORT void
mx_set_scan_pause_request_handler(
	mx_status_type (*pause_request_handler)( MX_SCAN * ) )
{
	if ( pause_request_handler == NULL ) {
		mx_scan_pause_request_handler
				= mx_default_scan_pause_request_handler;
	} else {
		mx_scan_pause_request_handler = pause_request_handler;
	}
	return;
}

MX_EXPORT mx_status_type
mx_default_scan_pause_request_handler( MX_SCAN *scan )
{
	const char fname[] = "mx_default_scan_pause_request_handler()";

	MX_DEBUG(-2,("%s invoked.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/* --------------- */

MX_EXPORT mx_status_type
mx_scan_display_scan_progress( MX_SCAN *scan )
{
	const char fname[] = "mx_scan_display_scan_progress()";

	MX_RECORD **input_device_array;
	MX_RECORD *input_device;
	MX_LINEAR_SCAN *linear_scan;
	double *motor_position;
	long *motor_is_independent_variable;
	char buffer[2500];
	char little_buffer[50];
	int i, num_motors, num_input_devices;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.", fname));

	num_motors = scan->num_motors;
	motor_position = scan->motor_position;
	motor_is_independent_variable = scan->motor_is_independent_variable;

	num_input_devices = scan->num_input_devices;
	input_device_array = scan->input_device_array;

	strcpy(buffer, "");

	if ( num_motors <= 0 ) {
		if ( scan->record->mx_class == MXS_LINEAR_SCAN ) {
			linear_scan = (MX_LINEAR_SCAN *)
					(scan->record->record_class_struct);
			sprintf( little_buffer, " %ld",
					(linear_scan->step_number)[0] );
			strcat(buffer, little_buffer);
		}
	} else {
		for ( i = 0; i < num_motors; i++ ) {
			if ( motor_is_independent_variable[i] ) {
				sprintf( little_buffer, " %-.3f",
							motor_position[i] );
				strcat(buffer, little_buffer);
			}
		}
	}
	strcat(buffer, " - ");

	for ( i = 0; i < num_input_devices; i++ ) {

		input_device = input_device_array[i];

		status = mx_convert_device_value_to_string( input_device,
				little_buffer, sizeof(little_buffer) - 1 );

		if ( status.code != MXE_SUCCESS )
			return status;

		strcat(buffer, little_buffer);
		strcat(buffer, " ");
	} 

	/* Send the formatted progress string to the user. */

	mx_info(buffer);

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_clear_scan_input_devices( MX_SCAN *scan )
{
	const char fname[] = "mx_clear_scan_input_devices()";

	MX_RECORD *input_device;
	long i;
	mx_status_type status;

	/* ===== Clear all the input devices that need it. ===== */

	/* At present, only scalers need to be cleared. */

	if ( scan->num_input_devices > 0 ) {
		if ( scan->input_device_array == (MX_RECORD **) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
				"Scaler array pointer is NULL.");
		}
	}

	for ( i = 0; i < scan->num_input_devices; i++ ) {

		input_device = (scan->input_device_array)[i];

		if ( input_device == (MX_RECORD *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"Input device %ld pointer is NULL", i );
		}

		switch( input_device->mx_superclass ) {
		case MXR_VARIABLE:
		case MXR_SCAN:
			/* No clear operation needed. */
			break;

		case MXR_DEVICE:
			switch( input_device->mx_class ) {
			case MXC_SCALER:
				status = mx_scaler_clear( input_device );

				if ( status.code != MXE_SUCCESS ) {
					return status;
				}
				break;

			case MXC_TIMER:
				status = mx_timer_clear( input_device );

				if ( status.code != MXE_SUCCESS ) {
					return status;
				}
				break;

			case MXC_MULTICHANNEL_ANALYZER:
				status = mx_mca_clear( input_device );

				if ( status.code != MXE_SUCCESS ) {
					return status;
				}
				break;

			default:
				/* All other input devices do not need
				 * to be cleared.
				 */
				break;
			}
			break;

		default:
			/* Interface and other record superclasses
			 * cannot be scan input devices.
			 */

			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record superclass %ld cannot be a scan input.",
				input_device->mx_superclass );
			break;
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT void
mx_log_scan_start( MX_LIST_HEAD *list_head, MX_SCAN *scan )
{
	char buffer[200];

	/* Do nothing if no log handler has been installed. */

	if ( list_head->log_handler == NULL )
		return;

	/* If this scan is a child of another scan,
	 * do not record a message.
	 */

	if ( scan->datafile.type == MXDF_CHILD )
		return;

	/* Record the fact that the scan has started. */

	sprintf( buffer, "Scan %s started", scan->record->name );

	mx_log_message( list_head->log_handler, buffer );

	return;
}

MX_EXPORT void
mx_log_scan_end( MX_LIST_HEAD *list_head, MX_SCAN *scan,
				mx_status_type status )
{
	char buffer1[300];
	char buffer2[200];

	/* Do nothing if no log handler has been installed. */

	if ( list_head->log_handler == NULL )
		return;

	/* If this scan is a child of another scan,
	 * do not record a message.
	 */

	if ( scan->datafile.type == MXDF_CHILD )
		return;

	/* Record the status of the scan. */

	if ( status.code == MXE_SUCCESS ) {
		sprintf( buffer1, "Scan %s ended", scan->record->name );
	} else {
		sprintf( buffer1, "Scan %s ended    %s   %s",
			scan->record->name,
			mx_strerror(status.code, buffer2, sizeof buffer2),
			status.message );
	}

	mx_log_message( list_head->log_handler, buffer1 );

	return;
}

MX_EXPORT void
mx_handle_abnormal_scan_termination( MX_LIST_HEAD *list_head,
				MX_SCAN *scan,
				mx_status_type status )
{
	(void) mx_deconfigure_measurement_type( &(scan->measurement) );

	(void) mx_scan_free_measurement_permit_and_fault_handlers( scan );

	if ( list_head->log_handler != NULL ) {
		mx_log_scan_end( list_head, scan, status );
	}
}

MX_EXPORT double
mx_scan_get_measurement_time( MX_SCAN *scan )
{
	const char fname[] = "mx_scan_get_measurement_time()";

	double measurement_time;
	int num_items;

	if ( scan == (MX_SCAN *) NULL ) {
		(void) mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_SCAN pointer passed was NULL." );

		return -1.0;
	}

	if ( scan->measurement.type != MXM_PRESET_TIME ) {
		(void) mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"The scan '%s' is not a preset time scan.",
			scan->record->name );

		return -1.0;
	}

	num_items = sscanf( scan->measurement_arguments, "%lg",
					&measurement_time );

	if ( num_items != 1 ) {
		(void) mx_error( MXE_UNPARSEABLE_STRING, fname,
			"The measurement time cannot be found in the "
			"measurement_arguments field '%s' for scan '%s'.",
			scan->measurement_arguments,
			scan->record->name );

		return -1.0;
	}

	return measurement_time;
}


MX_EXPORT mx_status_type
mx_scan_fixup_varargs_record_field_defaults(
		MX_RECORD_FIELD_DEFAULTS *record_field_defaults_array,
		long num_record_fields,
		long *num_independent_variables_varargs_cookie,
		long *num_motors_varargs_cookie,
		long *num_input_devices_varargs_cookie )
{
	const char fname[] = "mx_scan_fixup_varargs_record_field_defaults()";

	MX_RECORD_FIELD_DEFAULTS *field;
	long referenced_field_index;
	mx_status_type status;

	MX_DEBUG( 8,("%s invoked.", fname));

	/****
	 **** Put the 'num_motors' value in the appropriate places.
	 ****/

	status = mx_find_record_field_defaults_index(
			record_field_defaults_array, num_record_fields,
			"num_motors", &referenced_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie(
			referenced_field_index, 0, num_motors_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 8,("%s: num_motors varargs cookie = %ld",
			fname, *num_motors_varargs_cookie));

	/*---*/

	status = mx_find_record_field_defaults(
			record_field_defaults_array, num_record_fields,
			"motor_position", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	field->dimension[0] = *num_motors_varargs_cookie;

	/*---*/

	status = mx_find_record_field_defaults(
			record_field_defaults_array, num_record_fields,
			"motor_array", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	field->dimension[0] = *num_motors_varargs_cookie;

	/*---*/

	status = mx_find_record_field_defaults(
			record_field_defaults_array, num_record_fields,
			"motor_is_independent_variable", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	field->dimension[0] = *num_motors_varargs_cookie;

	/****
	 **** Put the 'num_input_devices' value in the
	 **** 'input_device_array' field.
	 ****/

	status = mx_find_record_field_defaults_index(
			record_field_defaults_array, num_record_fields,
			"num_input_devices", &referenced_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie(
			referenced_field_index, 0,
			num_input_devices_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 8,("%s: num_input_devices varargs cookie = %ld",
			fname, *num_input_devices_varargs_cookie));

	/*---*/

	status = mx_find_record_field_defaults(
			record_field_defaults_array, num_record_fields,
			"input_device_array", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	field->dimension[0] = *num_input_devices_varargs_cookie;

	/****
	 **** Construct the 'num_independent_variables' varargs cookie
	 **** which is used by the individual device classes.
	 ****/

	status = mx_find_record_field_defaults_index(
			record_field_defaults_array, num_record_fields,
			"num_independent_variables", &referenced_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie(
			referenced_field_index, 0,
			num_independent_variables_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 8,("%s: num_independent_variables varargs_cookie = %ld",
			fname, *num_independent_variables_varargs_cookie));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_convert_device_value_to_string( MX_RECORD *input_device,
				char *buffer, size_t buffer_length )
{
	const char fname[] = "mx_convert_device_value_to_string()";

	MX_RECORD_FIELD *record_field;
	mx_status_type (*token_constructor) ( void *, char *, size_t,
					MX_RECORD *, MX_RECORD_FIELD * );
	MX_ANALOG_INPUT *analog_input;
	MX_ANALOG_OUTPUT *analog_output;
	MX_DIGITAL_INPUT *digital_input;
	MX_DIGITAL_OUTPUT *digital_output;
	MX_MOTOR *motor;
	MX_SCALER *scaler;
	MX_TIMER *timer;
	MX_AMPLIFIER *amplifier;
	void *field_value_ptr;
	double position, raw_position;
	mx_status_type status;

	if ( buffer_length <= 0 ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"Buffer length %ld is too short.", (long) buffer_length );
	}

	switch( input_device->mx_superclass ) {
	case MXR_SCAN:
		strcpy(buffer, "");
		break;
	case MXR_DEVICE:
		if ( buffer_length <= 10 ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"Buffer length %ld <= 10 is too short.", (long)buffer_length );
		}
		switch( input_device->mx_class ) {
		case MXC_ANALOG_INPUT:
			analog_input = (MX_ANALOG_INPUT *)
				(input_device->record_class_struct);
			sprintf(buffer, "%-10.*g", input_device->precision,
					analog_input->value);
			break;
		case MXC_ANALOG_OUTPUT:
			analog_output = (MX_ANALOG_OUTPUT *)
				(input_device->record_class_struct);
			sprintf(buffer, "%-10.*g", input_device->precision,
					analog_output->value);
			break;
		case MXC_DIGITAL_INPUT:
			digital_input = (MX_DIGITAL_INPUT *)
				(input_device->record_class_struct);
			sprintf(buffer, "%10ld", digital_input->value);
			break;
		case MXC_DIGITAL_OUTPUT:
			digital_output = (MX_DIGITAL_OUTPUT *)
				(input_device->record_class_struct);
			sprintf(buffer, "%10ld", digital_output->value);
			break;
		case MXC_MOTOR:
			motor = (MX_MOTOR *)
				(input_device->record_class_struct);
			switch( motor->subclass ) {
			case MXC_MTR_ANALOG:
				raw_position = motor->raw_position.analog;
				break;
			case MXC_MTR_STEPPER:
				raw_position
				    = (double)(motor->raw_position.stepper);
				break;
			default:
				return mx_error(MXE_NOT_YET_IMPLEMENTED, fname,
				"Motor subclass %d not yet implemented.",
					motor->subclass );
				break;
			}
			position = motor->offset + motor->scale * raw_position;

			sprintf(buffer, "%-10.*g", input_device->precision,
					position);
			break;
		case MXC_SCALER:
			scaler = (MX_SCALER *)
				(input_device->record_class_struct);
			sprintf(buffer, "%10ld", scaler->value);
			break;
		case MXC_TIMER:
			timer = (MX_TIMER *)
				(input_device->record_class_struct);
			sprintf(buffer, "%10g", timer->value);
			break;
		case MXC_AMPLIFIER:
			amplifier = (MX_AMPLIFIER *)
				(input_device->record_class_struct);
			sprintf(buffer, "%10g", amplifier->gain);
			break;
		case MXC_MULTICHANNEL_ANALYZER:

			/* We currently don't print out the values for MCAs. */

			sprintf(buffer, "%10s", input_device->name );
			break;
		default:
			return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
			"Record class %ld not yet implemented.",
				input_device->mx_class );
			break;
		}
		break;
	case MXR_VARIABLE:
		status = mx_find_record_field( input_device, "value",
						&record_field );

		if ( status.code != MXE_SUCCESS )
			return status;

		field_value_ptr = mx_get_field_value_pointer( record_field );

		if ( field_value_ptr == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Field value pointer for record field '%s' in record '%s' is NULL.",
				record_field->name, input_device->name );
		}

		status = mx_get_token_constructor( record_field->datatype,
						&token_constructor );

		if ( status.code != MXE_SUCCESS )
			return status;

		strcpy( buffer, "" );

		status = mx_create_array_description(
				field_value_ptr,
				(record_field->num_dimensions - 1),
				buffer, buffer_length,
				input_device, record_field,
				token_constructor );

		if ( status.code != MXE_SUCCESS )
			return status;

		break;
	default:
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
			"Record superclass %ld not yet implemented.",
				input_device->mx_superclass );
		break;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_scan_find_parent_scan( MX_SCAN *child_scan, MX_SCAN **parent_scan )
{
	const char fname[] = "mx_scan_find_parent_scan()";

	MX_RECORD *parent_scan_record;

	if ( child_scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The child_scan pointer passed is NULL." );
	}
	if ( parent_scan == (MX_SCAN **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The parent_scan pointer passed is NULL." );
	}

	if ( child_scan->datafile.type != MXDF_CHILD ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Scan '%s' is not a child scan and does not have a parent.",
			child_scan->record->name );
	}

	/* For a child scan, the "filename" field contains the
	 * name of the parent scan that opened the datafile.
	 */

	parent_scan_record = mx_get_record( child_scan->record,
					child_scan->datafile.filename );

	if ( parent_scan_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"The parent of child scan '%s' is supposed to be named '%s', "
		"but a record of that name does not exist.",
			child_scan->record->name,
			child_scan->datafile.filename );
	}
	if ( parent_scan_record->mx_superclass != MXR_SCAN ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"The parent of child scan '%s' is supposed to be named '%s', "
		"but the record by that name is not a scan record.",
			child_scan->record->name,
			child_scan->datafile.filename );
	}

	*parent_scan = (MX_SCAN *) parent_scan_record->record_superclass_struct;

	if ( (*parent_scan) == (MX_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_SCAN pointer for parent scan '%s' used by child scan '%s' is NULL.",
			parent_scan_record->name, child_scan->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mx_scan_get_pointer_to_measurement_number( MX_SCAN *scan, long **ptr )
{
	const char fname[] = "mx_scan_get_pointer_to_measurement_number()";

	MX_SCAN *parent_scan;
	mx_status_type mx_status;

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_SCAN pointer passed was NULL." );
	}
	if ( ptr == (long **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Measurement number pointer passed was NULL." );
	}

	if ( scan->datafile.type != MXDF_CHILD ) {
		*ptr = &(scan->measurement_number);
	} else {
		mx_status = mx_scan_find_parent_scan( scan, &parent_scan );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		*ptr = &(parent_scan->measurement_number);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_scan_increment_measurement_number( MX_SCAN *scan )
{
	long *number_ptr;
	mx_status_type mx_status;

	mx_status = mx_scan_get_pointer_to_measurement_number(
						scan, &number_ptr );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	(*number_ptr)++;

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mx_scan_get_pointer_to_datafile_filename( MX_SCAN *scan, char **ptr )
{
	const char fname[] = "mx_scan_get_pointer_to_datafile_filename()";

	MX_SCAN *parent_scan;
	mx_status_type mx_status;

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_SCAN pointer passed was NULL." );
	}
	if ( ptr == (char **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Measurement number pointer passed was NULL." );
	}

	if ( scan->datafile.type != MXDF_CHILD ) {
		*ptr = scan->datafile.filename;
	} else {
		mx_status = mx_scan_find_parent_scan( scan, &parent_scan );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		*ptr = parent_scan->datafile.filename;
	}

	return MX_SUCCESSFUL_RESULT;
}

#define NUMBER_STRING_LENGTH	40

MX_EXPORT mx_status_type
mx_scan_save_mca_measurements( MX_SCAN *scan, long num_mcas )
{
	const char fname[] = "mx_scan_save_mca_measurements()";

	FILE *savefile;
	MX_MCA **mca_array;
	MX_MCA *mca;
	MX_RECORD *input_device;
	int os_status, saved_errno, exit_i_loop;
	long *number_ptr;
	long i, j, measurement_number, mca_number;
	unsigned long mca_data_value, max_current_num_channels;
	char *datafile_filename, *extension_ptr;
	char number_string[NUMBER_STRING_LENGTH + 1];
	char mca_filename[MXU_FILENAME_LENGTH + NUMBER_STRING_LENGTH + 1];
	ptrdiff_t basename_length;
	mx_status_type mx_status;

	if ( num_mcas <= 0 ) {
		return MX_SUCCESSFUL_RESULT;
	}

	mx_status = mx_scan_get_pointer_to_measurement_number(
						scan, &number_ptr );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	measurement_number = *number_ptr;

	sprintf( number_string, "%03ld", measurement_number );

	/* mx_mca_read() will already have been invoked during the
	 * measurement process for each MCA, so all we have to do
	 * here is get pointers to the MCA structures and use the
	 * values that are already there.
	 */

	mca_array = (MX_MCA **) malloc( num_mcas * sizeof(MX_MCA *) );

	if ( mca_array == (MX_MCA **) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannot allocate a %ld element array of MX_MCA pointers.",
			num_mcas );
	}

	mca_number = 0;

	for ( i = 0; i < scan->num_input_devices; i++ ) {

		input_device = (scan->input_device_array)[i];

		if ( input_device->mx_class == MXC_MULTICHANNEL_ANALYZER ) {

			mca_array[ mca_number ] = (MX_MCA *)
					input_device->record_class_struct;

			if ( mca_array[ mca_number ] == (MX_MCA *) NULL ) {
				mx_free( mca_array );

				return mx_error(
					MXE_CORRUPT_DATA_STRUCTURE, fname,
					"MX_MCA pointer for MCA record '%s' "
					"is NULL.", input_device->name );
			}
			mca_number++;
		}
	}

	/* Construct the name of the MCA datafile based on the scan
	 * datafile name.
	 */

	mx_status = mx_scan_get_pointer_to_datafile_filename(
						scan, &datafile_filename );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_free( mca_array );
		return mx_status;
	}

	extension_ptr = strrchr( datafile_filename, '.' );

	if ( extension_ptr == NULL ) {

		/* The scan datafile name has no extension so we just
		 * append the measurement number.
		 */

		mx_strncpy( mca_filename,
			datafile_filename, sizeof(mca_filename) );

		mx_strappend( mca_filename, ".", sizeof(mca_filename) );

		mx_strappend( mca_filename,
			number_string, sizeof(mca_filename) );
	} else {
		basename_length = extension_ptr - datafile_filename;

		if ( basename_length > sizeof(mca_filename) ) {
			mx_free( mca_array );

			return mx_error( MXE_FUNCTION_FAILED, fname,
"basename_length '%ld' is greater than sizeof(mca_filename) '%ld'.  "
"This shouldn't be possible, so this is a bug if you see this message.",
				(long) basename_length,
				(long) sizeof(mca_filename) );
		}

		mx_strncpy( mca_filename,
			datafile_filename, basename_length );

		mx_strappend( mca_filename, "_", sizeof(mca_filename) );

		extension_ptr++;

		mx_strappend( mca_filename,
			extension_ptr, sizeof(mca_filename) );

		mx_strappend( mca_filename, ".", sizeof(mca_filename) );

		mx_strappend( mca_filename,
			number_string, sizeof(mca_filename) );
	}

	/* Open the MCA datafile. */

	savefile = fopen( mca_filename, "w" );

	if ( savefile == NULL ) {
		saved_errno = errno;

		mx_free( mca_array );

		return mx_error( MXE_FILE_IO_ERROR, fname,
		"Could not open MCA datafile '%s'.  Reason = '%s'",
			mca_filename, strerror(saved_errno) );
	}

	/* Find the largest value of 'mca->current_num_channels' over
	 * all of the MCAs.
	 */

	max_current_num_channels = 0;

	for ( i = 0; i < num_mcas; i++ ) {
		mca = mca_array[i];

		if ( mca->current_num_channels > max_current_num_channels )
			max_current_num_channels = mca->current_num_channels;
	}

	/* Loop until all of the MCAs have run out of channels. */

	exit_i_loop = FALSE;

	for ( i = 0; i < max_current_num_channels; i++ ) {

		for ( j = 0; j < num_mcas; j++ ) {

			mca = mca_array[j];

			if ( mca->current_num_channels <= i ) {
				mca_data_value = 0;
			} else {
				mca_data_value = mca->channel_array[i];
			}

			fprintf( savefile, "%10lu  ", mca_data_value );

			if ( feof(savefile) || ferror(savefile) ) {
				saved_errno = errno;

				(void) mx_error( MXE_FILE_IO_ERROR, fname,
				"Error writing to MCA datafile '%s' for "
				"channel %ld of MCA '%s'.  Reason = '%s'",
					mca_filename, i, mca->record->name,
					strerror(saved_errno));

				/* Force the for( i ) loop to exit. */

				exit_i_loop = TRUE;

				break;	/* Exit the for( j ) loop. */
			}
		}
		if ( exit_i_loop ) {
			break;	/* Exit the for( i ) loop. */
		}

		fprintf( savefile, "\n" );

		if ( feof(savefile) || ferror(savefile) ) {
			saved_errno = errno;

			(void) mx_error( MXE_FILE_IO_ERROR, fname,
			"Error writing line terminator to MCA datafile '%s' "
			"for channel %ld'.  Reason = '%s'",
				mca_filename, i,
				strerror(saved_errno));

			break;	/* Exit the for( i ) loop. */
		}
	}

	mx_free( mca_array );

	os_status = fclose( savefile );

	if ( os_status != 0 ) {
		saved_errno = errno;

		return mx_error( MXE_FILE_IO_ERROR, fname,
			"Unable to close MCA datafile '%s'.  Reason = '%s'",
				mca_filename, strerror(saved_errno) );
	}

	return MX_SUCCESSFUL_RESULT;
}

