/*
 * Name:    sq_aps_id.c
 *
 * Purpose: Driver for quick scans that use an MX MCS record to collect
 *          the data and simultaneously synchronously scan an Advanced
 *          Photon Source insertion device.  This scan type is a variant
 *          of the MCS quick scan and uses most of the same code.
 * 
 * Author:  William Lavender
 *
 * Requirements:  This scan type requires the presence of three MX database
 *          records with the following names:
 *
 *          id_ev         - An 'aps_gap' type MX record setup to control the
 *                          APS id gap in units of X-ray energy (eV).
 *
 *          id_ev_enabled - '1' means coupled insertion device moves are
 *                          enabled.  '0' means they are disabled.
 *
 *          id_ev_params  - A two element array of doubles whose elements are
 *
 *              id_ev_params[0] = undulator harmonic number.  The APS MEDM
 *                                controls have an alternate and better way
 *                                of setting the undulator harmonic, so this
 *                                value should be left as '1'.
 *
 *              id_ev_params[1] = offset between the commanded insertion device
 *                                position and the commanded monochromator
 *                                position in eV.
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2000-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mxconfig.h"

#include "mx_record.h"
#include "mx_driver.h"
#include "mx_variable.h"
#include "mx_scan.h"
#include "mx_scan_quick.h"
#include "sq_mcs.h"
#include "sq_aps_id.h"
#include "d_aps_gap.h"

#if HAVE_EPICS

#include "tsDefs.h"
#include "cadef.h"

#include "mx_epics.h"

MX_RECORD_FUNCTION_LIST mxs_apsid_quick_scan_record_function_list = {
	mxs_mcs_quick_scan_initialize_type,
	mxs_mcs_quick_scan_create_record_structures,
	mxs_mcs_quick_scan_finish_record_initialization,
	mxs_mcs_quick_scan_delete_record,
	mx_quick_scan_print_scan_structure,
	mxs_mcs_quick_scan_dummy_function,
	mxs_mcs_quick_scan_dummy_function,
	mxs_mcs_quick_scan_dummy_function,
	mxs_mcs_quick_scan_dummy_function
};

MX_SCAN_FUNCTION_LIST mxs_apsid_quick_scan_scan_function_list = {
	mxs_apsid_quick_scan_prepare_for_scan_start,
	mxs_apsid_quick_scan_execute_scan_body,
	mxs_apsid_quick_scan_cleanup_after_scan_end
};

MX_RECORD_FIELD_DEFAULTS mxs_apsid_quick_scan_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_SCAN_STANDARD_FIELDS,
	MX_QUICK_SCAN_STANDARD_FIELDS
};

long mxs_apsid_quick_scan_num_record_fields
			= sizeof( mxs_apsid_quick_scan_defaults )
			/ sizeof( mxs_apsid_quick_scan_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxs_apsid_quick_scan_def_ptr
			= &mxs_apsid_quick_scan_defaults[0];

MX_EXPORT mx_status_type
mxs_apsid_quick_scan_prepare_for_scan_start( MX_SCAN *scan )
{
	const char fname[] = "mxs_apsid_quick_scan_prepare_for_scan_start()";

	MX_QUICK_SCAN *quick_scan;
	MX_MCS_QUICK_SCAN *mcs_quick_scan;
	MX_APSID_QUICK_SCAN_EXTENSION *apsid_quick_scan_extension;
	MX_RECORD *id_ev_record, *energy_record;
	MX_MOTOR *id_ev_motor;
	double id_ev_start_position, energy_start_position;
	long id_ev_enabled, num_parameters;
	double *double_array;
	double id_ev_harmonic, id_ev_offset;
	int busy;
	void *pointer_to_value;
	mx_status_type status;

	status = mxs_mcs_quick_scan_get_pointers( scan,
			&quick_scan, &mcs_quick_scan, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Are insertion device coupled moves currently enabled in the
	 * MX database?
	 */

	status = mx_get_long_variable_by_name( scan->record->list_head,
					"id_ev_enabled", &id_ev_enabled );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( id_ev_enabled != 1 ) {

		/* If coupled moves are not enabled, then run this scan
		 * as an ordinary MCS quick scan.
		 */

		if ( mcs_quick_scan->extension_ptr != NULL ) {
			free( mcs_quick_scan->extension_ptr );

			mcs_quick_scan->extension_ptr = NULL;
		}

		/* Prepare for scan start. */

		status = mxs_mcs_quick_scan_prepare_for_scan_start( scan );

		return status;
	}

	/* If we get here, coupled insertion device moves _are_ enabled,
	 * so we need to perform extra preparation steps for this scan.
	 */

	/* Allocate some memory for the quick scan extension structure
	 * if it has not already been done.
	 */

	if ( mcs_quick_scan->extension_ptr != NULL ) {
		apsid_quick_scan_extension = (MX_APSID_QUICK_SCAN_EXTENSION *)
			mcs_quick_scan->extension_ptr; 
	} else {
		apsid_quick_scan_extension = (MX_APSID_QUICK_SCAN_EXTENSION *)
			malloc( sizeof(MX_APSID_QUICK_SCAN_EXTENSION) );

		if ( apsid_quick_scan_extension == NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
"Ran out of memory trying to allocate an MX_APSID_QUICK_SCAN_EXTENSION "
"structure for scan record '%s'.", scan->record->name );
		}

		mcs_quick_scan->extension_ptr = apsid_quick_scan_extension;
	}

	/* Find the insertion device control record. */

	id_ev_record = mx_get_record( scan->record, "id_ev" );

	if ( id_ev_record == NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The insertion device record 'id_ev' was not found, so this APS "
"insertion device quick scan '%s' cannot be run.  Scan aborted.",
			scan->record->name );
	}

	id_ev_motor = (MX_MOTOR *) id_ev_record->record_class_struct;

	if ( id_ev_motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer for insertion device '%s' is NULL.",
			id_ev_record->name );
	}

	apsid_quick_scan_extension->id_ev_record = id_ev_record;

	/* Get the insertion device harmonic number and energy offset. */

	status = mx_get_1d_array_by_name( scan->record->list_head,
			"id_ev_params", MXFT_DOUBLE,
			&num_parameters, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	double_array = (double *) pointer_to_value;

	id_ev_harmonic = double_array[0];
	id_ev_offset = double_array[1];

	apsid_quick_scan_extension->id_ev_harmonic = id_ev_harmonic;
	apsid_quick_scan_extension->id_ev_offset = id_ev_offset;

	/* The monochromator energy should be the first motor. */

	energy_record = scan->motor_record_array[0];

	/* Compute the start position of the insertion device. */

	energy_start_position = quick_scan->start_position[0];

	id_ev_start_position = mx_divide_safely(
				energy_start_position + id_ev_offset,
						id_ev_harmonic );

	/* Start the insertion device moving towards the start position. */

	mx_info("Moving the insertion device '%s' to %g %s.",
					id_ev_record->name,
					id_ev_start_position,
					id_ev_motor->units );

	status = mx_motor_move_absolute( id_ev_record,
					id_ev_start_position, MXF_MTR_NOWAIT );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Now we are ready to perform the rest of the setup
	 * for the quick scan.
	 */

	status = mxs_mcs_quick_scan_prepare_for_scan_start( scan );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* If the insertion device is still moving to the start position,
	 * wait for it to be done.
	 */

	status = mx_motor_is_busy( id_ev_record, &busy );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( busy ) {
		mx_info(
	"Waiting for the insertion device '%s' to get to the start position.",
			id_ev_record->name );

		status = mx_wait_for_motor_stop( id_ev_record, 0 );

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	mx_info( "The insertion device '%s' is at the start position.",
			id_ev_record->name );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_apsid_quick_scan_execute_scan_body( MX_SCAN *scan )
{
	const char fname[] = "mxs_apsid_quick_scan_execute_scan_body()";

	MX_QUICK_SCAN *quick_scan;
	MX_MCS_QUICK_SCAN *mcs_quick_scan;
	MX_APSID_QUICK_SCAN_EXTENSION *apsid_quick_scan_extension;
	MX_RECORD *id_ev_record;
	MX_APS_GAP *aps_gap_struct;
	int sector_number;
	char pvname[80];
	double starting_gap, ending_gap, gap_scan_time;
	long stop, start;
	char status_byte;
	mx_status_type status;

	status = mxs_mcs_quick_scan_get_pointers( scan,
			&quick_scan, &mcs_quick_scan, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* If the insertion device coupling is turned off, the pointer
	 * mcs_quick_scan->extension_ptr will be NULL.  In that case,
	 * just run this as a normal MCS quick scan.
	 */

	if ( mcs_quick_scan->extension_ptr == NULL ) {
		status = mxs_mcs_quick_scan_execute_scan_body( scan );

		return status;
	}

	/* Get some pointers that we will need. */

	apsid_quick_scan_extension = (MX_APSID_QUICK_SCAN_EXTENSION *)
					mcs_quick_scan->extension_ptr;

	id_ev_record = apsid_quick_scan_extension->id_ev_record;

	aps_gap_struct = (MX_APS_GAP *) id_ev_record->record_type_struct;

	/* Which APS sector are we at? */

	sector_number = aps_gap_struct->sector_number;

	/**** Configure the insertion device for synchronous scanning. ****/

	/* Set the gap to end at. */

	sprintf( pvname, "ID%02d:SSEndGap", sector_number );

	status = mx_caput( pvname, MX_CA_DOUBLE, 1, &ending_gap );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Set the gap to start at. */

	sprintf( pvname, "ID%02d:SSStartGap", sector_number );

	status = mx_caput( pvname, MX_CA_DOUBLE, 1, &starting_gap );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Set the scanning time. */

	sprintf( pvname, "ID%02d:SSTime", sector_number );

	status = mx_caput( pvname, MX_CA_DOUBLE, 1, &gap_scan_time );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Toggle the synchronous scan mode off and then back on. */

	sprintf( pvname, "ID%02d:SyncScanMode.VAL", sector_number );

	status = mx_caput( pvname, MX_CA_LONG, 1, &stop );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_caput( pvname, MX_CA_LONG, 1, &start );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Start the undulator gap moving. */

	sprintf( pvname, "ID%02d:SSStart.VAL", sector_number );

	status = mx_caput( pvname, MX_CA_LONG, 1, &start );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Wait for the premove to finish. */

	sprintf( pvname, "ID%02d:SSState.VAL", sector_number );

	do {
		status = mx_caget( pvname, MX_CA_BYTE, 1, &status_byte );

		if ( status.code != MXE_SUCCESS )
			return status;

		mx_msleep(1);

	} while ( status_byte != 2 );

	/* The status byte changes to 2 at 2.5 seconds before the insertion
	 * device reaches the start position.  Thus, we must wait for an
	 * additional 2.5 seconds before starting the other motors.
	 */

	mx_msleep(2500);

	/**** Start the rest of the motors moving. ****/

	status = mxs_mcs_quick_scan_execute_scan_body( scan );

	return status;
}

MX_EXPORT mx_status_type
mxs_apsid_quick_scan_cleanup_after_scan_end( MX_SCAN *scan )
{
	const char fname[] = "mxs_apsid_quick_scan_cleanup_after_scan_end()";

	MX_QUICK_SCAN *quick_scan;
	MX_MCS_QUICK_SCAN *mcs_quick_scan;
	mx_status_type status;

	status = mxs_mcs_quick_scan_get_pointers( scan,
			&quick_scan, &mcs_quick_scan, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxs_mcs_quick_scan_cleanup_after_scan_end( scan );

	return status;
}

#endif /* HAVE_EPICS */
