/*
 * Name:    ms_main.c
 *
 * Purpose: main() routine for the MX server.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>

#include "mx_unistd.h"

#include "mx_osdef.h"
#include "mx_util.h"
#include "mx_version.h"
#include "mx_types.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_socket.h"
#include "mx_select.h"
#include "mx_net.h"
#include "mx_clock.h"
#include "mx_log.h"
#include "mx_syslog.h"

#include "ms_server.h"
#include "ms_mxserver.h"
#include "ms_security.h"

#if USE_GETOPT
#  ifdef OS_CYGWIN
#    include <getopt.h>
#  else
     extern char *optarg;
     extern int optind;
#  endif
#endif

MXSRV_EVENT_HANDLER mxsrv_event_handler_list[] = {
	{ MXF_SRV_MX_SERVER_EVENT_TYPE,
	  "MX server",
	  &mxsrv_mx_server_socket_struct,
	  mxsrv_mx_server_socket_init,
	  mxsrv_mx_server_socket_process_event,
	  NULL },

	{ MXF_SRV_MX_CLIENT_EVENT_TYPE,
	  "MX client",
	  NULL,
	  mxsrv_mx_client_socket_init,
	  mxsrv_mx_client_socket_process_event,
	  mxsrv_mx_client_socket_proc_queued_event },

};

int mxsrv_num_event_handlers = sizeof( mxsrv_event_handler_list )
				/ sizeof( mxsrv_event_handler_list[0] );

static void
mxsrv_exit_handler( void )
{
	mx_info( "*** MX server process exiting. ***" );
}

static void
mxsrv_sigterm_handler( int signal_number )
{
	mx_info( "Received a request to shutdown via a SIGTERM signal." );

	exit(0);
}

static void
mxsrv_install_signal_and_exit_handlers( void )
{
	atexit( mxsrv_exit_handler );

	signal( SIGTERM, mxsrv_sigterm_handler );

#ifdef SIGILL
	signal( SIGILL, mx_standard_signal_error_handler );
#endif
#ifdef SIGTRAP
	signal( SIGTRAP, mx_standard_signal_error_handler );
#endif
#ifdef SIGIOT
	signal( SIGIOT, mx_standard_signal_error_handler );
#endif
#ifdef SIGBUS
	signal( SIGBUS, mx_standard_signal_error_handler );
#endif
#ifdef SIGFPE
	signal( SIGFPE, mx_standard_signal_error_handler );
#endif
#ifdef SIGSEGV
	signal( SIGSEGV, mx_standard_signal_error_handler );
#endif

#ifdef SIGPIPE
	/* We want to handle death of child events by detecting read
	 * or write errors to the associated pipe rather than being
	 * sent a SIGPIPE signal.  Thus, we ignore SIGPIPE and depend
	 * on the value of 'errno' to tell us what happened.
	 */

	signal( SIGPIPE, SIG_IGN );
#endif

	return;
}

int
main( int argc, char *argv[] )
{
	const char fname[] = "main()";

	MX_RECORD *mx_record_list, *current_record;
	MX_LIST_HEAD *list_head_struct;
	MX_CLOCK_TICK current_clock_tick;
	char mx_database_filename[MXU_FILENAME_LENGTH+1];
	char mx_connection_acl_filename[MXU_FILENAME_LENGTH+1];
	char ident_string[80];
	int i, debug_level, saved_errno;
	int num_fds, max_fd, current_socket;
	int server_port, default_display_precision, init_hw_flags;
	int install_syslog_handler, syslog_number, syslog_options;
	int display_stack_traceback;

	MXSRV_EVENT_HANDLER *event_handler;
	int max_sockets, handler_array_size;
	MXSRV_SOCKET_HANDLER_LIST socket_handler_list;
	fd_set readfds;
	struct timeval timeout;

	mx_status_type ( *init_fn ) (MX_RECORD *,
					MXSRV_SOCKET_HANDLER_LIST *,
					MXSRV_EVENT_HANDLER * );
	mx_status_type ( *process_event_fn ) ( MX_RECORD *,
					MXSRV_SOCKET_HANDLER *,
					MXSRV_SOCKET_HANDLER_LIST *,
					MXSRV_EVENT_HANDLER * );
	mx_status_type status;

#if USE_GETOPT
	int c, error_flag;
#endif

	strcpy( mx_database_filename, "mxserver.dat" );
	strcpy( mx_connection_acl_filename, "" );

	debug_level = 0;

	default_display_precision = 8;

	server_port = 9727;

	install_syslog_handler = FALSE;

	syslog_number = -1;
	syslog_options = 0;

	init_hw_flags = 0;

#if defined( OS_WIN32 )
	display_stack_traceback = FALSE;
#else
	display_stack_traceback = TRUE;
#endif

#if USE_GETOPT
        /* Process command line arguments, if any. */

        error_flag = FALSE;

        while ((c = getopt(argc, argv, "C:d:f:l:L:p:P:sSt")) != -1 ) {
                switch (c) {
		case 'C':
			mx_strncpy( mx_connection_acl_filename,
					optarg, MXU_FILENAME_LENGTH );
			break;
                case 'd':
                        debug_level = atoi( optarg );
                        break;
                case 'f':
                        mx_strncpy( mx_database_filename,
					optarg, MXU_FILENAME_LENGTH );
                        break;
		case 'l':
			install_syslog_handler = TRUE;
			syslog_options = 0;

			syslog_number = atoi( optarg );
			break;
		case 'L':
			install_syslog_handler = TRUE;
			syslog_options = MXF_SYSLOG_USE_STDERR;

			syslog_number = atoi( optarg );
			break;
                case 'p':
                        server_port = atoi( optarg );
                        break;
		case 'P':
			default_display_precision = atoi( optarg );
			break;
		case 's':
			display_stack_traceback = TRUE;
			break;
		case 'S':
			display_stack_traceback = FALSE;
			break;
		case 't':
			init_hw_flags |= MXF_INITHW_TRACE_OPENS;
			break;
                case '?':
                        error_flag = TRUE;
                        break;
                }

                if ( error_flag == TRUE ) {
                        fprintf( stderr,
"Usage: mxserver [-d debug_level] [-f mx_database_file] [-l log_number]\n"
"  [-L log_number ] [-p server_port] [-P display_precision] \n"
"  [-C connection_acl_filename]\n" );
                        exit(1);
                }
        }

#endif /* USE_GETOPT */

	if ( display_stack_traceback ) {
		mxsrv_install_signal_and_exit_handlers();
	}

	mx_set_debug_level( debug_level );

	if ( install_syslog_handler ) {
		sprintf( ident_string, "mxserver@%d", server_port );

		status = mx_install_syslog_handler( ident_string,
					syslog_number, syslog_options );

		if ( status.code != MXE_SUCCESS )
			exit(1);
	}

	mx_info( "*** MX server %s started at %s ***",
			mx_get_version_string(),
			mx_current_time_string( NULL, 0 ) );

	mxsrv_mx_server_socket_struct.port = server_port;

	for ( i = 0; i < mxsrv_num_event_handlers; i++ ) {
		if ( mxsrv_event_handler_list[i].type
				== MXF_SRV_MX_CLIENT_EVENT_TYPE ) {
			break;   /* Exit the for() loop. */
		}
	}

	if ( i >= mxsrv_num_event_handlers ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The MX client event handler is not found "
			"in the mxsrv_event_handler_list array." );

		exit( MXE_CORRUPT_DATA_STRUCTURE );
	}

	mxsrv_mx_server_socket_struct.client_event_handler
		= &mxsrv_event_handler_list[i];

	/* Initialize the subsecond sleep functions (if they need it). */

	mx_msleep(1);

	/* Initialize the MX time keeping functions. */

	mx_initialize_clock_ticks();

	/* Initialize the list of socket handlers. */

	max_sockets = mx_get_max_file_descriptors();

	if ( max_sockets <= 1 ) {
		(void) mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"Not enough file descriptors for the MX server to run." );

		exit( MXE_WOULD_EXCEED_LIMIT );
	}

	handler_array_size = max_sockets;

	socket_handler_list.max_sockets = max_sockets;
	socket_handler_list.num_sockets_in_use = 0;
	socket_handler_list.handler_array_size = handler_array_size;

	socket_handler_list.array = (MXSRV_SOCKET_HANDLER **)
		malloc( handler_array_size * sizeof(MXSRV_SOCKET_HANDLER *) );

	if ( socket_handler_list.array == (MXSRV_SOCKET_HANDLER **) NULL ) {
		(void) mx_error( MXE_OUT_OF_MEMORY, fname,
		"Couldn't allocate socket handler array for %d handlers.",
			handler_array_size );

		exit( MXE_OUT_OF_MEMORY );
	}

	for ( i = 0; i < handler_array_size; i++ ) {
		socket_handler_list.array[i] = NULL;
	}

	/* Initialize the MX device drivers. */

	status = mx_initialize_drivers();

	if ( status.code != MXE_SUCCESS )
		exit( status.code );

	/* Initialize the MX database. */

	mx_record_list = mx_initialize_record_list();

	if ( mx_record_list == (MX_RECORD *) NULL ) {
		(void) mx_error( MXE_FUNCTION_FAILED, fname,
			"Attempt to initialize the record list failed." );

		exit( MXE_FUNCTION_FAILED );
	}

	/* Mark this record list as belonging to a server. */

	list_head_struct = (MX_LIST_HEAD *)
				(mx_record_list->record_superclass_struct);

	if ( list_head_struct == (MX_LIST_HEAD *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_LIST_HEAD pointer for list head record is NULL." );

		exit( MXE_CORRUPT_DATA_STRUCTURE );
	}

	list_head_struct->is_server = TRUE;

	/* Set the default floating point display precision. */

	list_head_struct->default_precision = default_display_precision;

	/* Set server_protocols_active to reflect the protocols that will
	 * be used.
	 */

	list_head_struct->server_protocols_active = MXPROT_MX;

	/* Read the database description file and add the records therein
	 * to the record list.
	 */

	status = mx_read_database_file( mx_record_list,
					mx_database_filename, 0 );

	if ( status.code != MXE_SUCCESS )
		exit( status.code );

	/* Perform database initialization steps that cannot be done until
	 * all the records have been defined.
	 */

	status = mx_finish_database_initialization( mx_record_list );

	if ( status.code != MXE_SUCCESS )
		exit( status.code );

	/* Initialize the MX log file if the log file control record
	 * "mx_log" has been defined.
	 */

	status = mx_log_open( mx_record_list );

	if ( status.code != MXE_SUCCESS )
		exit( status.code );

	/* Initialize all the hardware described by the record list. */

	status = mx_initialize_hardware( mx_record_list, init_hw_flags );

	if ( status.code != MXE_SUCCESS )
		exit( status.code );

#if 0
	/* List all of the records that have been defined. */

	{
		MX_RECORD *current_record;

		current_record = mx_record_list;

		do {
			MX_DEBUG(-2,(
	"%s: Record '%s' defined.  Superclass = %ld, class = %ld, type = %ld",
				fname,
				current_record->name,
				current_record->superclass,
				current_record->class,
				current_record->type ));

			current_record = current_record->next_record;

		} while ( current_record != mx_record_list );
	}
#endif

	/* Initialize all the event handler interfaces. */

	for ( i = 0; i < mxsrv_num_event_handlers; i++ ) {
		init_fn = mxsrv_event_handler_list[i].init;

		if ( init_fn == NULL ) {
			(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"init function pointer for server type ? is NULL." );

			exit( MXE_CORRUPT_DATA_STRUCTURE );
		}

		status = ( *init_fn ) ( mx_record_list,
					&socket_handler_list,
					&mxsrv_event_handler_list[i] );

		if ( status.code != MXE_SUCCESS )
			exit( status.code );
	}

	/* Initialize the socket connection access control list. */

	if ( strlen( mx_connection_acl_filename ) > 0 ) {

		status = mxsrv_setup_connection_acl( mx_record_list,
						mx_connection_acl_filename );

		if ( status.code != MXE_SUCCESS )
			exit( status.code );
	}

	/************ Primary event loop *************/

	mx_info("mxserver: Ready to accept client connections.");

	while (1) {
		/* Sleep a moment, so that we don't eat up all available
		 * CPU time.
		 */

		mx_usleep(500);

#if 0
		MX_DEBUG(-2,("%s: Top of event loop.", fname));
#endif

		/* Initialize the arguments to select(). */

		FD_ZERO(&readfds);

		max_fd = -1;

		for ( i = 0; i < handler_array_size; i++ ) {
			if ( socket_handler_list.array[i] != NULL ) {
				current_socket
			    = socket_handler_list.array[i]->synchronous_socket;

				if ( current_socket < 0 ) {
					MX_DEBUG(0,
("main #1: socket_handler_list.array[%d] = %p, current_socket = %d",
		i, socket_handler_list.array[i], current_socket));
				}

				FD_SET( current_socket, &readfds );

				if ( current_socket > max_fd ) {
					max_fd = current_socket;
				}
			}
		}

#ifdef OS_WIN32
		max_fd = -1;
#else
		max_fd++;
#endif

		timeout.tv_sec = 0;
		timeout.tv_usec = 1000;     /* Wait one millisecond. */
		
		/* Use select() to look for events. */

		num_fds = select( max_fd, &readfds, NULL, NULL, &timeout );

		saved_errno = errno;

		if ( num_fds < 0 ) {
			(void) mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Error in select() while waiting for events.  "
			"Errno = %d.  Error string = '%s'.",
			saved_errno, strerror( saved_errno ) );

		} else if ( num_fds == 0 ) {

			/* Didn't get any events, so do nothing here. */

		} else {
			MX_DEBUG( 2,("%s: select() returned.  num_fds = %d",
					fname, num_fds));

			/* Figure out which sockets had events and
			 * then process the events.
			 */

			for ( i = 0; i < handler_array_size; i++ ) {
			    if ( socket_handler_list.array[i] != NULL ) {

				current_socket
			    = socket_handler_list.array[i]->synchronous_socket;

				if ( current_socket < 0 ) {
					MX_DEBUG(0,
("main #2: socket_handler_list.array[%d] = %p, current_socket = %d",
		i, socket_handler_list.array[i], current_socket));
				}

				if ( FD_ISSET(current_socket, &readfds) ){

					event_handler =
				socket_handler_list.array[i]->event_handler;

					if ( event_handler == NULL ) {
						(void) mx_error(
						MXE_NETWORK_IO_ERROR, fname,
		"Event handler pointer for socket handler %d is NULL.", i);

						continue;
					}

					process_event_fn
					    = event_handler->process_event;

					if ( process_event_fn == NULL ) {
						(void) mx_error(
						MXE_NETWORK_IO_ERROR, fname,
	"process_event function pointer for socket handler %d is NULL.", i);

						continue;
					}

					(void) ( *process_event_fn )
						( mx_record_list,
						  socket_handler_list.array[i],
						  &socket_handler_list,
						  event_handler );
				}
			    }
			}
		}

		/* Are there any queued events that are ready to process? */

		current_clock_tick = mx_current_clock_tick();

		current_record = mx_record_list;

		do {
			if ( current_record->event_queue != NULL ) {

				status = mxsrv_process_queued_event(
						current_record,
						current_clock_tick );
			}
			current_record = current_record->next_record;

		} while ( current_record != mx_record_list );
	}
}

