/*
 * Name:    ms_process_rs232.c
 *
 * Purpose: Functions used to process MX rs232 record events.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_driver.h"
#include "mx_record.h"
#include "mx_rs232.h"
#include "mx_socket.h"

#include "ms_server.h"
#include "ms_process_rs232.h"

mx_status_type
mxsrv_setup_rs232_process_functions( MX_RECORD *record_list_head )
{
	const char fname[] = "mxsrv_setup_rs232_process_functions()";

	MX_RECORD *current_record;
	MX_RECORD_FIELD *record_field;
	MX_RECORD_FIELD *record_field_array;
	long i;

	MX_DEBUG(2,("%s invoked.", fname));

	/* The following is done just in case we were called with
	 * a record that isn't really the list head.
	 */

	record_list_head = record_list_head->list_head;

	/* Go through the record list looking for rs232 records. */

	current_record = record_list_head->next_record;

	while ( current_record != record_list_head ) {

		if ( (current_record->mx_superclass == MXR_INTERFACE)
		  && (current_record->mx_class == MXI_RS232) ) {

			record_field_array
				= current_record->record_field_array;

			for ( i = 0;
				i < current_record->num_record_fields;
				i++ )
			{
				record_field = &record_field_array[i];

				switch( record_field->label_value ) {
				case MXLV_232_GETCHAR:
				case MXLV_232_PUTCHAR:
				case MXLV_232_INPUT_IS_AVAILABLE:
				case MXLV_232_DISCARD_UNREAD_INPUT:
				case MXLV_232_DISCARD_UNWRITTEN_OUTPUT:
				    record_field->process_function
					= mxsrv_rs232_process_function;
				    break;
				default:
				    break;
				}
			}
		}
		current_record = current_record->next_record;
	}

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_rs232_process_function( void *record_ptr,
			void *record_field_ptr, int operation )
{
	const char fname[] = "mxsrv_rs232_process_function()";

	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MX_RS232 *rs232;
	char operation_string[30];
	long label_value;
	char char_value;
	int int_value;
	mx_status_type status;

	record = (MX_RECORD *) record_ptr;
	record_field = (MX_RECORD_FIELD *) record_field_ptr;

	switch ( operation ) {
	case MX_PROCESS_GET:
		strcpy( operation_string, "(MX_PROCESS_GET)" );
		break;
	case MX_PROCESS_PUT:
		strcpy( operation_string, "(MX_PROCESS_PUT)" );
		break;
	default:
		sprintf( operation_string, "(Operation %d)", operation );
		break;
	}

	label_value = record_field->label_value;

	MX_DEBUG(1,("%s: '%s.%s' %s (%ld)",
		fname, record->name, record_field->name,
		operation_string, label_value));

	/*
	 * A label value < 0 tells us that we do not need to do any
	 * further MX record processing. 
	 */

	if ( label_value < 0 ) {
		return MX_SUCCESSFUL_RESULT;
	}

	rs232 = (MX_RS232 *) (record->record_class_struct);

	status = MX_SUCCESSFUL_RESULT;

	switch( operation ) {
	case MX_PROCESS_GET:
		switch( label_value ) {
		case MXLV_232_GETCHAR:
			status = mx_rs232_getchar( record, &char_value, 0 );

			rs232->getchar_value = (int) char_value;
			break;
		case MXLV_232_INPUT_IS_AVAILABLE:
			status = mx_rs232_input_is_available( record,
								&int_value );

			rs232->input_is_available = int_value;
			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_GET label value = %ld",
				fname, label_value));
			break;
		}

		/* Now invoke the value changed callback. */

		if ( status.code == MXE_SUCCESS ) {
			status = mxsrv_invoke_callback_list(
					MXSF_VALUE_CHANGED_CALLBACK,
					record_field );
		}
		break;
	case MX_PROCESS_PUT:
		switch( label_value ) {
		case MXLV_232_PUTCHAR:
			char_value = (char) rs232->putchar_value;

			status = mx_rs232_putchar( record, char_value, 0 );
			break;
		case MXLV_232_DISCARD_UNREAD_INPUT:
			status = mx_rs232_discard_unread_input( record, 0 );
			break;
		case MXLV_232_DISCARD_UNWRITTEN_OUTPUT:
			status = mx_rs232_discard_unwritten_output( record, 0 );
			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_PUT label value = %ld",
				fname, label_value));
			break;
		}

		/* Now invoke the value changed callback. */

		if ( status.code == MXE_SUCCESS ) {
			status = mxsrv_invoke_callback_list(
					MXSF_VALUE_CHANGED_CALLBACK,
					record_field );
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown operation code = %d", operation );
		break;
	}

	return status;
}

